// IBM Confidential OCO Source Material
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2005-2008
// The source code for this program is not published or otherwise divested
// of its trade secrets, irrespective of what has been deposited with the
// U.S. Copyright Office.
//
// $Id$
//
// Module  :  BackingMap.java
//
// Source File Description:
//
//    Need a file description.
//
// Change Activity:
//
// Reason       Version   Date         Userid    Change Description
// ------------ --------- ------------ --------- -----------------------------------------
// XD60_M1_B.1  WASX.XD   02/23/2005   sutter    new code drop
// XD60_M2_B.2  WASX.XD   03/21/2005   sutter    M2 Deliverable
// XD60_M3_B.3  WASX.XD   04/05/2005   kjlaw     M3 add setCopyMode method.
// XD60_M3_B.10 WASX.XD   04/07/2005   sutter    Re-factor datagrid/datagrid.impl
// XD60_M3_B.10 WASX.XD   04/21/2005   kjlaw     add MapEventListener methods.
// 271911       WASX      May 16, 2005 sutter    Partitioning support updates
// 277357       WASX.XD   May 21, 2005 kjlaw     added missing get methods.
// 277186       WASX.XD   06/06/2005   kjlaw     add new BackingMap methods for LockManager.
// 276305       WASX.XD   06/15/2005   kjlaw     add setCopyKey method.
// 284358       WASX.XD   06/21/2005   kjlaw     make TTLEvictor part of core.
// XD601_M1_B.5 WASX.XD   08/26/2005   cheng1    add dynamic indexing support
// XD601_M2_B.2 WASX.XD   09/28/2005   sutter    enable partitioning
// XD601_M1_B.5 WASX.XD   08/26/2005   cheng1    Index operations need to declare throw UnsupportedOperationException
// 335078.1     WASX.XD   01/11/2006   sutter    remove setPartitionId from public API
// b998         6.1       05/22/2006   cdjohnson added getEntityMetadata()
// b11208       WASX.XD   03/24/2008   jian.tang Write-behind loader support
// b18298       7.1       01/22/2009   schmittm  add getMapSetName method
// b22228       7.1.1     02/14/2011   kulack    Event Listener remodel in support of byte array plugin enhancements
// ------------ --------- ------------ --------- -----------------------------------------

package com.ibm.websphere.objectgrid;

import java.util.List;

import com.ibm.websphere.objectgrid.plugins.BackingMapLifecycleListener;
import com.ibm.websphere.objectgrid.plugins.BackingMapPlugin;
import com.ibm.websphere.objectgrid.plugins.EventListener;
import com.ibm.websphere.objectgrid.plugins.Evictor;
import com.ibm.websphere.objectgrid.plugins.Loader;
import com.ibm.websphere.objectgrid.plugins.LogSequence;
import com.ibm.websphere.objectgrid.plugins.MapEventListener;
import com.ibm.websphere.objectgrid.plugins.ObjectTransformer;
import com.ibm.websphere.objectgrid.plugins.OptimisticCallback;
import com.ibm.websphere.objectgrid.plugins.builtins.NoVersioningOptimisticCallback;
import com.ibm.websphere.objectgrid.plugins.index.DynamicIndexCallback;
import com.ibm.websphere.objectgrid.plugins.index.MapIndex;
import com.ibm.websphere.objectgrid.plugins.index.MapIndexPlugin;
import com.ibm.websphere.objectgrid.plugins.index.MapRangeIndex;
import com.ibm.websphere.objectgrid.plugins.io.SerializerAccessor;
import com.ibm.websphere.objectgrid.security.ServerMapPermission;
import com.ibm.websphere.objectgrid.writebehind.WriteBehindLoaderConstants;
import com.ibm.websphere.projector.md.EntityMetadata;

/**
 * This is the public interface to the BackingMap. It is returned when a new Map
 * is defined on the ObjectGrid. It allows the Map to be customized with various
 * plug-ins or by setting properties. The defaults are:
 * <ul>
 * <li>No external <code>Evictor</code>, but an internal time-based evictor is provided by default</li>
 * <li>No <code>Loader</code></li>
 * <li>No <code>EventListener</code>s</li>
 * <li>No <code>MapIndexPlugin</code>s</li>
 * <li>An internal <code>ObjectTransformer</code></li>
 * <li>An internal <code>OptimisticCallback</code></li>
 * <li>Key is not copied</li>
 * <li>A value <code>CopyMode</code> of <code>CopyMode.COPY_ON_READ_AND_COMMIT</code></li>
 * <li>A <code>LockStrategy</code> of <code>LockStrategy.OPTIMISTIC</code></li>
 * <li>A default lock timeout</li>
 * <li><code>null</code> values are supported</li>
 * <li>A default number of buckets</li>
 * <li>A default number of lock buckets</li>
 * <li>Synchronous preload</li>
 * <li>Read/write map by default</li>
 * <li>A TimeToLive of 0 (indicating unlimited time)</li>
 * <li>A TtlEvictor type of <code>TTLType.NONE</code></li>
 * <li>Write-behind updates is disabled</li>
 * <li>Time-based database updates are disabled</li>
 * <li>Eviction triggers are not set</li>
 * </ul>
 * @since WAS XD 6.0, XC10
 */
public interface BackingMap extends ClientReplicableMap {
    /**
     * Default lock timeout used if <code>setLockTimeout(int)</code>
     * is not invoked.
     */
    final static int DEFAULT_LOCK_TIMEOUT = 15; // 15 seconds.

    /**
     * Default number of lock buckets used if <code>setNumberOfBuckets(int)</code>
     * is not invoked.
     * @deprecated Deprecated in 8.6.  Buckets are no longer required.  Use the {@link #isNearCacheEnabled()} flag to disable
     * the near cache in the ObjectGrid configuration XML file.
     */
    final static int DEFAULT_NUMBER_OF_BUCKETS = 10007; // 100649; // prime number.

    /**
     * Default number of lock buckets used if <code>setNumberOfLockBuckets(int)</code>
     * is not invoked.
     */
    final static int DEFAULT_NUMBER_OF_LOCK_BUCKETS = 4007;// 12073; // prime number.

    /**
     * The eviction trigger string constant to enable memory based eviction using memory usage
     * threshold provided by the <tt>java.lang.management.MemoryPoolMXBean</tt>.
     * @see #setEvictionTriggers(String)
     * @since WAS XD 6.1.0.3
     */
    public final static String EVICTIONTRIGGER_MEMORY_USAGE_THRESHOLD = "MEMORY_USAGE_THRESHOLD";

    /**
     * Gets the name of the <code>BackingMap</code>.
     *
     * @return value specified when <code>BackingMap</code> was created.
     */
    public abstract String getName();

    //    /**
    //     *
    //     *           This support is removed for this release, moved to a future release
    //     *
    //     * This allows a new CacheEntryFactory to be plugged in.
    //     * @param map The factory instance to use for this Map.
    //     * <p>
    //     * Note, to avoid an <code>IllegalStateException</code>, this method must be called
    //     * prior to <code>ObjectGrid.initialize()</code> method.  Also, keep in mind that the
    //     * <code>ObjectGrid.getSession()</code> method implicitly calls the
    //     * <code>ObjectGrid.initialize()</code> method if it has yet to be called by the application.
    //     *
    //     * @throws IllegalStateException if this method is called after the
    //     *         <code>ObjectGrid.initialize()</code> method is called.
    //
    //    public abstract void setCacheEntryFactory(CacheEntryFactory map);
    //
    //    /**
    //     *
    //     *           This support is remove for this release, moved to XD6.01
    //     *
    //     * Get the CacheEntryFactory being used by this BackingMap.
    //     *
    //     * @return   the argument that was passed to the <code>setCacheEntryFactory(CacheEntryFactory)</code>
    //     *           method of this interface or the default CacheEntryFactory object if the
    //     *           the setCacheEntryFactory was not previously called for this BackingMap object.
    //     *
    //     * @see #setCacheEntryFactory(CacheEntryFactory)
    //
    //    This support is remove for this release, moved to a future release
    //    public abstract CacheEntryFactory getCacheEntryFactory();
    //
    //     */

    /**
     * Associates an <code>Evictor</code> with this <code>BackingMap</code>.
     * <p>
     * An <code>Evictor</code> aids with cleaning up the cache based on whatever
     * algorithm is desired (LRU, LFU, etc).  Passing <code>null</code> to this
     * method removes a previously set <code>Evictor</code> object from an earlier
     * invocation of this method.
     * <p>
     * Note, to avoid an <code>IllegalStateException</code>, this method must be called
     * prior to the <code>ObjectGrid.initialize()</code> method.  Also, keep in mind that the
     * <code>ObjectGrid.getSession()</code> method implicitly calls the
     * <code>ObjectGrid.initialize()</code> method if it has yet to be called by the application.
     * <p>
     * An Evictor that implements the {@link BackingMapLifecycleListener} is automatically added
     * as if the {@link #addMapEventListener(EventListener)} method was called. Any previous
     * evictor which implements {@link BackingMapLifecycleListener} is removed as if
     * the {@link #removeMapEventListener(EventListener)} method was called.
     * <p>
     * An <code>Evictor</code> may also implement the {@link BackingMapPlugin} interface in order
     * to receive enhanced BackingMap plug-in lifecycle method calls. The plug-in is then
     * also required to correctly implement each of the bean methods related to
     * introspection of its state (for example <code>isInitialized()</code>,
     * <code>isDestroyed()</code>, etc).
     *
     * @param e Evictor instance
     *
     * @throws IllegalStateException if this method is called after the
     *         <code>ObjectGrid.initialize()</code> method is called.
     *
     * @see Evictor
     * @see ObjectGrid#initialize()
     * @see ObjectGrid#getSession()
     */
    public abstract void setEvictor(Evictor e);

    /**
     * Gets the <code>Evictor</code> being used by this BackingMap.
     *
     * @return   the argument that was passed to the <code>setEvictor(Evictor)</code> method of
     *           this interface or <code>null</code> if setEvictor was not previously called
     *           for this BackingMap object.
     * @see Evictor
     * @see #setEvictor(Evictor)
     */
    public abstract Evictor getEvictor();

    /**
     * Sets the ObjectTransformer object for use by this <code>BackingMap</code> and/or
     * <code>Loader</code>.
     * <p>
     * An <code>ObjectTransformer</code> aids with the "serialization" of non-Serializable
     * objects.  It allows a custom copy function to be installed for more
     * efficient object copy operations.
     * <p>
     * Note, to avoid an <code>IllegalStateException</code>, this method must be called
     * prior to the <code>ObjectGrid.initialize()</code> method.  Also, keep in mind that the
     * <code>ObjectGrid.getSession()</code> method implicitly calls the
     * <code>ObjectGrid.initialize()</code> method if it has yet to be called by the application.
     * <p>
     * An ObjectTransformer that implements the {@link BackingMapLifecycleListener} is automatically added
     * as if the {@link #addMapEventListener(EventListener)} method was called. Any previous
     * transformer which implements {@link BackingMapLifecycleListener} is removed as if
     * the {@link #removeMapEventListener(EventListener)} method was called.
     *
     * @param t ObjectTransformer instance
     *
     * @throws IllegalArgumentException if the passed in <code>ObjectTransformer</code>
     *         is <code>null</code>
     * @throws IllegalStateException if this method is called after the
     *         <code>ObjectGrid.initialize()</code> method is called.
     *
     * @see ObjectTransformer
     * @see ObjectGrid#initialize()
     * @see ObjectGrid#getSession()
     */
    public abstract void setObjectTransformer(ObjectTransformer t);

    /**
     * Gets the <code>ObjectTransformer</code> object being used by this
     * <code>BackingMap</code> and/or <code>Loader</code>.
     *
     * @return   the argument that was passed to the <code>setObjectTransformer(ObjectTransformer)</code>
     *           method of this interface or the default <code>ObjectTransformer</code> object if the
     *           <code>setObjectTransformer</code> method was not previously called for this object.
     *
     * @see ObjectTransformer
     * @see #setObjectTransformer(ObjectTransformer)
     */
    public abstract ObjectTransformer getObjectTransformer();

    /**
     * Sets the <code>OptimisticCallback</code>.
     * <p>
     * The OptimisticCallback will be used to check the versions of cache entries during
     * the commit phase.  If no OptimisticCallback was previously set, a default OptimisticCallback
     * will be used. For Entities, the default OptimisticCallback will use a version field that was
     * specified in the entity metadata.  For POJO objects or Entities that do not have a version field
     * specified, the default OptimisticCallback uses the entire object as the version value.
     * In order for it to work for POJO objects, the application's value object needs to have a useful
     * equals(Object) method.  If your application does not require versioning, but is using Optimistic
     * locking, the NoVersioningOptimistCallback should be used.
     * <p>
     * Note, to avoid an <code>IllegalStateException</code>, this method must be called
     * prior to the <code>ObjectGrid.initialize()</code> method.  Also, keep in mind that the
     * <code>ObjectGrid.getSession()</code> method implicitly calls the
     * <code>ObjectGrid.initialize()</code> method if it has yet to be called by the application.
     * <p>
     * An OptimisticCallback that implements the {@link BackingMapLifecycleListener} is automatically added
     * as if the {@link #addMapEventListener(EventListener)} method was called. Any previous
     * optimistic callback which implements {@link BackingMapLifecycleListener} is removed as if
     * the {@link #removeMapEventListener(EventListener)} method was called.
     *
     * @param checker OptimisticCallback instance
     *
     * @throws IllegalArgumentException if the passed in <code>OptimisticCallback</code>
     *         is <code>null</code>
     * @throws IllegalStateException if this method is called after the
     *         <code>ObjectGrid.initialize()</code> method is called.
     *
     * @see OptimisticCallback
     * @see NoVersioningOptimisticCallback
     * @see LockStrategy#OPTIMISTIC
     * @see ObjectGrid#initialize()
     * @see ObjectGrid#getSession()
     */
    public abstract void setOptimisticCallback(OptimisticCallback checker);

    /**
     * Gets the <code>OptimisticCallback</code> being used by this <code>BackingMap</code>
     * and/or <code>Loader</code> or <code>null</code> if the LockStrategy is not optimistic.
     * <p>
     * If no OptimisticCallback was previously set, a default OptimisticCallback will be used.
     * For Entities, the default OptimisticCallback will use a version field that was specified
     * in the entity metadata.  For POJO objects or Entities that do not have a version field
     * specified, the default OptimisticCallback uses the entire object as the version value.
     * In order for it to work for POJO objects, the application's value object needs to have a useful
     * equals(Object) method.  If your application does not require versioning, but is using Optimistic
     * locking, the NoVersioningOptimistCallback should be used.
     *
     * @return   the argument that was passed to the <code>setOptimisticCallback(OptimisticCallback)</code>
     *           method of this interface or the default <code>OptimisticCallback</code> object if the
     *           <code>setOptimisticCallback</code> method was not previously called for this object.
     *           If Optimistic locking is not being used, this method will return null after ObjectGrid.initialize()
     *           has been invoked.
     *
     * @see NoVersioningOptimisticCallback
     * @see OptimisticCallback
     * @see LockStrategy#OPTIMISTIC
     * @see #setOptimisticCallback(OptimisticCallback)
     */
    public abstract OptimisticCallback getOptimisticCallback();

    /**
     * Associates a <code>Loader</code> with this <code>BackingMap</code>.
     * <p>
     * Only one Loader can be associated with a given <code>BackingMap</code>.
     * Passing <code>null</code> to this method removes a previously set
     * <code>Loader</code> object from an earlier invocation of this method
     * and indicates that this <code>BackingMap</code> is not associated with a
     * <code>Loader</code>.
     * <p>
     * Note, to avoid an <code>IllegalStateException</code>, this method must be called
     * prior to the <code>ObjectGrid.initialize()</code> method.  Also, keep in mind that the
     * <code>ObjectGrid.getSession()</code> method implicitly calls the
     * <code>ObjectGrid.initialize()</code> method if it has yet to be called by the application.
     * <p>
     * A loader that implements the {@link BackingMapLifecycleListener} is automatically added
     * as if the {@link #addMapEventListener(EventListener)} method was called. Any previous
     * loader which implements {@link BackingMapLifecycleListener} is removed as if
     * the {@link #removeMapEventListener(EventListener)} method was called.
     * <p>
     * A <code>Loader</code> may also implement the {@link BackingMapPlugin} interface in order
     * to receive enhanced BackingMap plug-in lifecycle method calls. The plug-in is then
     * also required to correctly implement each of the bean methods related to
     * introspection of its state (for example <code>isInitialized()</code>,
     * <code>isDestroyed()</code>, etc).
     *
     * @param loader Loader instance
     *
     * @throws IllegalStateException if this method is called after the
     *         <code>ObjectGrid.initialize()</code> method is called.
     *
     * @see Loader
     * @see ObjectGrid#initialize()
     * @see ObjectGrid#getSession()
     */
    public abstract void setLoader(Loader loader);

    /**
     * Gets the Loader being used by this <code>BackingMap</code>.
     *
     * @return   the argument that was passed to the <code>setLoader(Loader)</code> method of
     *           this interface or <code>null</code> if <code>setLoader</code> was not previously called for this
     *           object.
     *
     * @see Loader
     * @see #setLoader(Loader)
     */
    public abstract Loader getLoader();

    /**
     * Sets the preload mode if a <code>Loader</code> is set for this <code>BackingMap</code>.
     * <p>
     * If the parameter is true then the <code>Loader.preloadMap(Session, BackingMap)</code> is
     * invoked asynchronously; otherwise it blocks the execution when loading data
     * so the cache is unavailable until preload completes. Preloading occurs during
     * <code>ObjectGrid</code> initialization.
     * <p>
     * Note, to avoid an <code>IllegalStateException</code>, this method must be called
     * prior to the <code>ObjectGrid.initialize()</code> method.  Also, keep in mind that the
     * <code>ObjectGrid.getSession()</code> method implicitly calls the
     * <code>ObjectGrid.initialize()</code> method if it has yet to be called by the application.
     *
     * @param async If this is true then the cache is loaded asynchronously otherwise it blocks
     *              and the cache is unavailable until preload completes.
     *
     * @throws IllegalStateException if this method is called after the
     *         <code>ObjectGrid.initialize()</code> method is called.
     *
     * @see Loader#preloadMap(Session, BackingMap)
     */
    public abstract void setPreloadMode(boolean async);

    /**
     * Returns whether this <code>BackingMap</code> will be asynchronously preloaded or not
     * if a <code>Loader</code> is set.
     * <p>
     * If true is returned then the <code>Loader.preloadMap(Session, BackingMap)</code> method is
     * invoked asynchronously; otherwise it blocks the execution when loading data
     * so the cache is unavailable until preload completes.  Preloading occurs during
     * <code>ObjectGrid</code> initialization.
     *
     * @return   the argument that was passed to the <code>setPreloadMode(boolean)</code> method of
     *           this interface or <code>false</code> if setPreloadeMode was not previously
     *           called for this object.
     *
     * @see Loader#preloadMap(Session, BackingMap)
     * @see #setPreloadMode(boolean)
     */
    public abstract boolean getPreLoadMode();

    /**
     * Adds an <code>MapIndexPlugin</code> to this Map. This method assumes the index implementation was constructed
     * with the name of the attribute to index. The name of the index is specified when
     * the index is constructed.
     * <p>
     * Note, to avoid an <code>IllegalStateException</code>, this method must be called
     * prior to <code>ObjectGrid.initialize()</code> method.  Also, keep in mind that the
     * <code>ObjectGrid.getSession()</code> method implicitly calls the
     * <code>ObjectGrid.initialize()</code> method if it has yet to be called by the application.
     * <p>
     * A MapIndexPlugin that implements the {@link BackingMapLifecycleListener} is automatically added
     * as if the {@link #addMapEventListener(EventListener)} method was called. Any previous
     * index which implements {@link BackingMapLifecycleListener} is removed as if
     * the {@link #removeMapEventListener(EventListener)} method was called.
     *
     * @param index The index implementation.
     *
     * @throws IndexAlreadyDefinedException if this index already exists.
     * @throws IllegalStateException if this method is called after the
     *         <code>ObjectGrid.initialize()</code> method is called.
     *
     * @see MapIndexPlugin
     * @see ObjectGrid#initialize()
     * @see ObjectGrid#getSession()
     */
    public void addMapIndexPlugin(MapIndexPlugin index) throws IndexAlreadyDefinedException;

    /**
     * Returns the current list of <code>MapIndexPlugin</code> objects for this BackingMap.
     *
     * @return The current list of <code>MapIndexPlugin</code>s for this BackingMap.
     *         The list is empty if the <code>addMapIndexPlugin(MapIndexPlugin)</code>
     *         or <code>setMapIndexPlugins(List)</code> method was
     *         not previously called for this BackingMap.
     *
     * @see #addMapIndexPlugin(MapIndexPlugin)
     * @see #setMapIndexPlugins(List)
     */
    public abstract List /* MapIndexPlugin */getMapIndexPlugins();

    /**
     * Sets the list of <code>MapIndexPlugin</code> objects for this BackingMap.
     * If the BackingMap already has a <code>List</code> of <code>MapIndexPlugin</code> objects,
     * that list is replaced by the <code>List</code> passed as
     * an argument to the current invocation of this method.
     * <p>
     * Note, to avoid an <code>IllegalStateException</code>, this method must be called
     * prior to <code>ObjectGrid.initialize()</code> method.  Also, keep in mind that the
     * <code>ObjectGrid.getSession()</code> method implicitly calls the
     * <code>ObjectGrid.initialize()</code> method if it has yet to be called by the application.
     *
     * @param indexList  A non-null reference to a <code>List</code> of <code>MapIndexPlugin</code> objects.
     *
     * @throws IllegalArgumentException is thrown if indexList is null
     *         or the indexList contains either a <code>null</code> reference
     *         or an object that is not an instance of <code>MapIndexPlugin</code>.
     *
     * @see MapIndexPlugin
     * @see ObjectGrid#initialize()
     * @see ObjectGrid#getSession()
     */
    public abstract void setMapIndexPlugins(List /* MapIndexPlugin */indexList);

    /**
     * Sets the <code>CopyMode</code>.
     * <p>
     * The <code>CopyMode</code> determines whether a get operation of an entry
     * in the <code>BackingMap</code> returns the actual value, a copy of the value,
     * or a proxy for the value.  In the case of a proxy, the copy of the value does not occur
     * unless a set method of the application provided value interface is invoked.
     * It also determines that when a transaction is committed, whether a copy
     * of the value object of an entry that was marked as dirty by the transaction
     * is put into the <code>BackingMap</code> at commit time.  The <code>CopyMode</code>
     * does not specify if the object is copied when being read or written to a
     * <code>Loader</code>. It is the responsibility of the implementor of a
     * <code>Loader</code> to make copies as appropriate.  The default
     * <code>CopyMode</code> is <code>CopyMode.COPY_ON_READ_AND_COMMIT</code>.
     * <p>
     * Note, to avoid an <code>IllegalStateException</code>, this method must be called
     * prior to the <code>ObjectGrid.initialize()</code> method.  Also, keep in mind that the
     * <code>ObjectGrid.getSession()</code> method implicitly calls the
     * <code>ObjectGrid.initialize()</code> method if it has yet to be called by the application.
     *
     * @param mode must be one of the final static variables defined in <code>CopyMode</code>.
     *        See <code>CopyMode</code> class for an explanation of each mode and how the valueInterface
     *        is used for <code>CopyMode.COPY_ON_WRITE</code> .
     *
     * @param valueInterface the value interface Class object.  Specify null in version 7.1 and later.
     *
     * @throws IllegalArgumentException if mode is CopyMode.COPY_ON_WRITE and
     *         valueInterface parameter is null and CGLIB isn't in the classpath.
     *
     * @throws IllegalStateException if this method is called after the
     *         <code>ObjectGrid.initialize()</code> method is called.
     *
     * @see CopyMode
     * @see ObjectGrid#initialize()
     * @see ObjectGrid#getSession()
     */
    public void setCopyMode(CopyMode mode, Class valueInterface);

    /**
     * Gets the <code>CopyMode</code> being used by this <code>BackingMap</code>.
     *
     * @return   the argument that was passed to the <code>setCopyMode(CopyMode, Class)</code> method of
     *           this interface or the default <code>CopyMode</code> object if <code>setCopyMode</code> was not previously
     *           called for this object.
     *
     * @see CopyMode
     * @see #setCopyMode(CopyMode, Class)
     */
    public abstract CopyMode getCopyMode();

    /**
     * Sets the <code>LockStrategy</code>.
     * <p>
     * The locking strategy represented by the <code>LockStrategy</code> object
     * determines if the internal ObjectGrid lock manager is
     * used whenever a map entry is accessed by a transaction.  The default
     * strategy is <code>LockStrategy.OPTIMISTIC</code>.
     * <p>
     * Note, to avoid an <code>IllegalStateException</code>, this method must be called
     * prior to the <code>ObjectGrid.initialize()</code> method.  Also, keep in mind that the
     * <code>ObjectGrid.getSession()</code> method implicitly calls the
     * <code>ObjectGrid.initialize()</code> method if it has yet to be called by the application.
     *
     * @param lockStrategy must be one of the final static variables defined in
     *        <code>LockStrategy</code>.  See <code>LockStrategy</code> class for an explanation of each
     *        locking strategy.
     *
     * @throws IllegalStateException if this method is called after the
     *         <code>ObjectGrid.initialize()</code> method is called.
     *
     * @see LockStrategy
     * @see ObjectGrid#initialize()
     * @see ObjectGrid#getSession()
     */
    public void setLockStrategy(LockStrategy lockStrategy);

    /**
     * Gets the <code>LockStrategy</code> object being used by this <code>BackingMap</code>.
     *
     * @return   the argument that was passed to the <code>setLockStrategy(LockStrategy)</code> method of
     *           this interface or the default <code>LockStrategy</code> object if <code>setLockStrategy</code> was not previously
     *           called for this object.
     *
     * @see LockStrategy
     * @see #setLockStrategy(LockStrategy)
     */
    public abstract LockStrategy getLockStrategy();

    /**
     * Sets the list of <code>EventListener</code> objects.
     * <p>
     * If this <code>BackingMap</code> already has a <code>List</code> of
     * <code>EventListener</code>s, that list is replaced by the
     * <code>List</code> passed as an argument to the current invocation
     * of this method. This method can be called before and
     * after the <code>ObjectGrid.initialize()</code> method.
     *
     * @param eventListenerList   A non-null reference to a List of
     *        <code>EventListener</code> objects that are instances of
     *        <code>BackingMapLifecycleListener</code> or <code>MapEventListener</code>
     *
     * @throws IllegalArgumentException is thrown if eventListenerList is null,
     *         the eventListenerList contains either a <code>null</code>
     *         reference or an object that is not an instance of
     *         <code>BackingMapLifecycleListener</code> or <code>MapEventListener</code>
     *
     * @deprecated This method is deprecated in version 7.1.1.
     * Use the {@link #addMapEventListener(EventListener)} or {@link #removeMapEventListener(EventListener)} methods.
     * Plugins that implement the ObjectGridLifecycleListener interface are automatically
     * registered with the grid. Using this method will remove those automatically
     * added listeners.
     *
     * @see EventListener
     * @see MapEventListener
     * @see BackingMapLifecycleListener
     * @see #addMapEventListener(EventListener)
     * @see #removeMapEventListener(EventListener)
     */
    @Deprecated
    public void setMapEventListeners(List /*EventListener*/eventListenerList);

    /**
     * Gets the current list of <code>EventListener</code>s.
     *
     * @return the current list of <code>EventListener</code> objects for this <code>BackingMap</code>.
     *
     * @see EventListener
     * @see MapEventListener
     * @see BackingMapLifecycleListener
     */
    public abstract List /*MapEventListener*/getMapEventListeners();

    /**
     * Adds an <code>EventListener</code> to this <code>BackingMap</code>.
     * <p>
     * Note, this method is allowed to be invoked before and after the
     * <code>ObjectGrid.initialize()</code> method.
     *
     * Backing map plug-ins (<code>Loader</code>, <code>Evictor</code>, <code>MapIndexPlugin</code>,
     * <code>ObjectTransformer</code>, <code>OptimisticCallback</code>)
     * that implement the <code>ObjectGridLifecycleListener</code> are automatically
     * added as listeners when added to the <code>BackingMap</code>.
     *
     * @param eventListener A non-null reference to a <code>EventListener</code> to add to the list.
     *     The listener must be an instance of <code>BackingMapLifecycleListener</code> or <code>MapEventListener</code>
     *
     * @throws IllegalArgumentException if eventListener is null or not an instance of
     * <code>BackingMapLifecycleListener</code> or <code>MapEventListener</code>
     *
     * @see EventListener
     * @see MapEventListener
     * @see BackingMapLifecycleListener
     */
    public void addMapEventListener(EventListener eventListener);

    /**
     * Provided for compatibility with old releases, use the
     * {@link #addMapEventListener(EventListener)} method.
     *
     * @deprecated This method is deprecated in version 7.1.1, use the {@link #addMapEventListener(EventListener)} method.
     *
     * @param eventListener A non-null reference to a <code>EventListener</code> to add to the list.
     *     The listener must be an instance of <code>BackingMapLifecycleListener</code> or <code>MapEventListener</code>
     */
    public void addMapEventListener(MapEventListener eventListener);

    /**
     * Removes an <code>EventListener</code> from this <code>BackingMap</code>.
     * <p>
     * Note, this method is allowed to be invoked before and after the
     * <code>ObjectGrid.initialize()</code> method.
     *
     * Backing map plug-ins (<code>Loader</code>, <code>Evictor</code>, <code>MapIndexPlugin</code>,
     * <code>ObjectTransformer</code>, <code>OptimisticCallback</code>)
     * that implement the <code>ObjectGridLifecycleListener</code> are automatically
     * removed as listeners when removed from the <code>ObjectGrid</code>.
     *
     * @param eventListener A non-null reference to an event listener that was
     *                      previously added by invoking either the <code>addMapEventListener(EventListener)</code>
     *                      or <code>setMapEventListeners(List)</code> method of this interface.
     *
     * @throws IllegalArgumentException if eventListener is null or not an instance of
     * <code>BackingMapLifecycleListener</code> or <code>MapEventListener</code>
     *
     * @see EventListener
     * @see MapEventListener
     * @see BackingMapLifecycleListener
     * @see #addMapEventListener(EventListener)
     */
    public void removeMapEventListener(EventListener eventListener);

    /**
     * Provided for compatibility with old releases, use the
     * {@link #removeMapEventListener(EventListener)} method.
     *
     * @param eventListener A non-null reference to an event listener that was
     *                      previously added by invoking either the <code>addMapEventListener(EventListener)</code>
     *                      or <code>setMapEventListeners(List)</code> method of this interface.
     */
    public void removeMapEventListener(MapEventListener eventListener);

    /**
     * Gets the partition identifier being used by this <code>BackingMap</code>.
     *
     * @return   The 0-based index for the partition represented by this <code>BackingMap</code>
     *           instance.  If there is only a single partition defined
     *           for this <code>BackingMap</code> object, a 0 will be returned (default).
     * @since WAS XD 6.0.1
     */
    public abstract int getPartitionId();

    /**
     * Sets the map type of this <code>BackingMap</code>.
     * <p>
     * A map can be a read only map or a read/write map. Passing true
     * as the parameter value will make this map a read only map; passing
     * false as the parameter value will make this map a read/write map.
     * <p>
     * Note, to avoid an <code>IllegalStateException</code>, this method must be called
     * prior to the <code>ObjectGrid.initialize()</code> method.  Also, keep in mind that the
     * <code>ObjectGrid.getSession()</code> method implicitly calls the
     * <code>ObjectGrid.initialize()</code> method if it has yet to be called by the application.
     *
     * @param readOnlyEnabled If set to true, this <code>BackingMap</code> will be a read only map.  If false, the
     *                        map will be a read/write map.
     *
     * @throws IllegalStateException if this method is called after the
     *         <code>ObjectGrid.initialize()</code> method is called.
     *
     */
    public void setReadOnly(boolean readOnlyEnabled);

    /**
     * Retrieves the map type.
     *
     * @return  the argument that was passed to <code>setReadOnly(boolean)</code> method of this interface.
     *          True is returned if this a read only map.  A return value of false implies that this is
     *          a read/write map.  If <code>setReadOnly</code> was never called, the default return
     *          value is false.
     *
     * @see #setReadOnly(boolean)
     */
    public boolean getReadOnly();

    /**
     * Gets the <code>ObjectGrid</code> that owns this <code>BackingMap</code>.
     *
     * @return the ObjectGrid instance that owns this BackingMap.
     *
     * @see ObjectGrid
     */
    public ObjectGrid getObjectGrid();

    /**
     * Sets the number of buckets used by this <code>BackingMap</code>.
     * <p>
     * The <code>BackingMap</code> implementation uses a hash map for its
     * implementation.  If there are a lot of entries in the <code>BackingMap</code>
     * then more buckets means better performance because the risk of collisions
     * is lower as the number of buckets grows. More buckets also means more
     * concurrency.  If number of buckets is 0, no entries will be stored in the
     * map, but the appropriate ObjectGrid and BackingMap plug-ins will still be called.
     * <p>
     * Once the <code>ObjectGrid</code> is initialized this parameter cannot be changed.
     * Therefore, to avoid an <code>IllegalStateException</code>, this method must be called
     * prior to the <code>ObjectGrid.initialize()</code> method.  Also, keep in mind that the
     * <code>ObjectGrid.getSession()</code> method implicitly calls the
     * <code>ObjectGrid.initialize()</code> method if it has yet to be called by the application.
     *
     * @param numBuckets The number of buckets to use.
     *
     * @throws IllegalArgumentException if numBuckets is less than 0.
     * @throws IllegalStateException if this method is called after the
     *         <code>ObjectGrid.initialize()</code> method is called.
     *
     * @see ObjectGrid#initialize()
     * @see ObjectGrid#getSession()
     * @deprecated Deprecated in 8.6.  Buckets are no longer required.  Use the {@link #isNearCacheEnabled()} flag to disable
     * the near cache in the ObjectGrid configuration XML file.
     */
    public void setNumberOfBuckets(int numBuckets);

    /**
     * Gets the number of buckets defined for this <code>BackingMap</code>.
     *
     * @return the same value passed to the <code>setNumberOfBuckets(int)</code> method or
     *         <code>DEFAULT_NUMBER_OF_BUCKETS</code> if <code>setNumberOfBuckets</code> was never called.
     *
     * @see #setNumberOfBuckets(int)
     * @see #DEFAULT_NUMBER_OF_BUCKETS
     * @deprecated Deprecated in 8.6.  Buckets are no longer required.  Use the {@link #isNearCacheEnabled()} flag to disable
     * the near cache in the ObjectGrid configuration XML file.
     */
    public int getNumberOfBuckets();

    /**
     * Sets the number of lock buckets used by the lock manager for this
     * <code>BackingMap</code>.
     * <p>
     * When <code>LockStrategy.OPTIMISTIC</code> or
     * <code>LockStrategy.PESSIMISTIC</code> is used for
     * this <code>BackingMap</code>, a lock manager is created for the
     * <code>BackingMap</code>.  The lock manager uses a hash map to keep track of
     * entries that are locked by 1 or more transactions.  If there are a
     * lot of entries in the hash map, then more lock buckets means better
     * performance as the risk of collisions is lower as the number of buckets
     * grows. More lock buckets also means more concurrency. When the lock strategy
     * is <code>LockStrategy.NONE</code>, no lock manager
     * is used by this <code>BackingMap</code>.  In this case, a call to this
     * method does nothing.
     * <p>
     * Once the <code>ObjectGrid</code> is initialized, the number of lock buckets
     * cannot be changed. Therefore, to avoid an <code>IllegalStateException</code>, this
     * method must be called prior to the <code>ObjectGrid.initialize()</code> method.
     * Also, keep in mind that the <code>ObjectGrid.getSession()</code> method
     * implicitly calls the <code>ObjectGrid.initialize()</code> method if it
     * has yet to be called by the application.
     *
     * @param numBuckets The number of lock buckets to use.
     *
     * @throws IllegalArgumentException if numBuckets is less than 1.
     * @throws IllegalStateException if this method is called after the
     *         <code>ObjectGrid.initialize()</code> method is called.
     *
     * @see LockStrategy
     * @see ObjectGrid#initialize()
     * @see ObjectGrid#getSession()
     */
    public void setNumberOfLockBuckets(int numBuckets);

    /**
     * Gets the number of lock buckets defined for the hash map used by lock
     * manager for this backing map.
     *
     * @return the same value passed to the <code>setNumberOfLockBuckets(int)</code> method or
     *         <code>DEFAULT_NUMBER_OF_LOCK_BUCKETS</code> if <code>setNumberOfLockBuckets</code> was never called.
     *
     * @see #setNumberOfLockBuckets(int)
     * @see #DEFAULT_NUMBER_OF_LOCK_BUCKETS
     */
    public int getNumberOfLockBuckets();

    /**
     * Sets the lock timeout used by the lock manager for this
     * <code>BackingMap</code>.
     * <p>
     * When <code>LockStrategy.OPTIMISTIC</code> or
     * <code>LockStrategy.PESSIMISTIC</code> is used
     * for this <code>BackingMap</code>, a lock manager is created for the
     * <code>BackingMap</code>.  To prevent deadlocks from occurring, the lock
     * manager has a default timeout value for waiting for a lock to be granted.
     * If this timeout limit is exceeded, a <code>LockTimeoutException</code>
     * is thrown.  The default value of <code>DEFAULT_LOCK_TIMEOUT</code> should be sufficient for most
     * applications, but on a heavily loaded system, a timeout may occur when no
     * deadlock exists. In that case, this method can be used to increase the lock
     * timeout value from the default to whatever is desired to prevent false
     * timeout exceptions from occurring.  When the lock strategy is
     * <code>LockStrategy.NONE</code>, no lock manager
     * is used by this <code>BackingMap</code>.  In this case, a call to this
     * method does nothing.  A lock timeout value of zero indicates to not wait for the
     * lock if it is not immediately available.
     * <p>
     * Once the lock manager is initialized, the lock timeout value cannot be changed.
     * Therefore, to avoid an <code>IllegalStateException</code>, this method must be called
     * prior to <code>ObjectGrid.initialize()</code> method.  Also, keep in mind that the
     * <code>ObjectGrid.getSession()</code> method implicitly calls the
     * <code>ObjectGrid.initialize()</code> method if it has yet to be called by the application.
     * When an entry is fetched the lock timeout can be changed for a given transaction
     * using <code>ObjectMap.setLockTimeout(int)</code>
     *
     * @param seconds is the lock timeout value to use in seconds.
     *
     * @throws IllegalArgumentException if seconds is less than 0.
     * @throws IllegalStateException if this method is called after the
     *         <code>ObjectGrid.initialize()</code> method is called.
     *
     * @see #DEFAULT_LOCK_TIMEOUT
     * @see LockStrategy
     * @see LockTimeoutException
     * @see ObjectGrid#initialize()
     * @see ObjectGrid#getSession()
     * @see ObjectMap#setLockTimeout(int)
     */
    public void setLockTimeout(int seconds);

    /**
     * Gets the lock timeout value used by the lock manager for this
     * <code>BackingMap</code>.
     *
     * @return the same value passed to the <code>setLockTimeout(int)</code> method or
     *         <code>DEFAULT_LOCK_TIMEOUT</code> if <code>setLockTimeout</code> was never called.
     *
     * @see #DEFAULT_LOCK_TIMEOUT
     * @see #setLockTimeout(int)
     */
    public int getLockTimeout();

    /**
     * Sets whether this <code>BackingMap</code> supports <code>null</code> values.
     * <p>
     * If <code>null</code> values are supported, users need to be careful when a
     * <code>get</code> operation returns a <code>null</code> reference. It could
     * be due to the fact that the key is not found in the <code>BackingMap</code>,
     * or that the value in the <code>BackingMap</code> is <code>null</code>. To
     * determine if a key was not found, or the value is <code>null</code>, the
     * <code>containsKey</code> method can be used.</p>
     * <p>
     * Note, to avoid an <code>IllegalStateException</code>, this method must be called
     * prior to the <code>ObjectGrid.initialize()</code> method.  Also, keep in mind that the
     * <code>ObjectGrid.getSession()</code> method implicitly calls the
     * <code>ObjectGrid.initialize()</code> method if it has yet to be called by the application.
     *
     * @param nullValuesSupported If set to true, <code>null</code> values are
     *                            supported; otherwise <code>null</code> values
     *                            are not supported.
     *
     * @throws IllegalStateException if this method is called after the
     *         <code>ObjectGrid.initialize()</code> method is called.
     *
     * @see ObjectGrid#initialize()
     * @see ObjectGrid#getSession()
     * @see ObjectMap#containsKey(Object)
     */
    public void setNullValuesSupported(boolean nullValuesSupported);

    /**
     * Gets whether this <code>BackingMap</code> supports <code>null</code> values
     * or not.
     * <p>
     * @return the same value passed to the <code>setNullValuesSupported(boolean)</code>
     *         method or the default value of true if <code>setNullValuesSupported</code>
     *         was never called.
     *
     * @see #setNullValuesSupported(boolean)
     */
    public boolean getNullValuesSupported();

    /**
     * Sets whether or not the key needs to be copied when a map entry
     * is created.
     * <p>
     * Copying the key object allows the application to use the same key
     * object for each ObjectMap operation.  The application changes the key
     * object state prior to each <code>ObjectMap</code> operation so that it
     * can work with different entries using the same key object.  If a separate
     * key object is used for each entry, then there is no reason to copy the key
     * object. This attribute allows an application to make the tradeoff of copying
     * key object versus using more memory as a result of separate key object used
     * by the application for each entry.  If this method is not called, then the
     * default of <code>false</code> is used (e.g. the key is NOT copied).
     * <p>
     * Note, to avoid an <code>IllegalStateException</code>, this method must be called
     * prior to the <code>ObjectGrid.initialize()</code> method.  Also, keep in mind that the
     * <code>ObjectGrid.getSession()</code> method implicitly calls the
     * <code>ObjectGrid.initialize()</code> method if it has yet to be called by the application.
     *
     * @param copy   If <code>true</code> is specified, then this <code>BackingMap</code>
     *               uses the <code>ObjectTransformer.copyKey(Object)</code> method
     *               to copy the key object when necessary.
     *
     * @throws IllegalStateException if this method is called after the
     *         <code>ObjectGrid.initialize()</code> method is called.
     *
     * @see ObjectGrid#initialize()
     * @see ObjectGrid#getSession()
     * @see ObjectTransformer#copyKey(Object)
     */
    public abstract void setCopyKey(boolean copy);

    /**
     * Gets whether keys are copied for this <code>BackingMap</code>.
     *
     * @return the same value passed to the <code>setCopyKey(boolean)</code>
     *         method or the default value of false if <code>setCopyKey</code>
     *         was never called.
     *
     * @see #setCopyKey(boolean)
     */
    public abstract boolean getCopyKey();

    /**
     * Sets "time to live" of each map entry in seconds.
     * <p>
     * If this method is not called, the lifetime of an entry is forever (or until
     * the application explicitly removes or invalidates the entry, or a user
     * defined <code>Evictor</code> evicts the entry).
     *
     * Note, to avoid an <code>IllegalStateException</code>, this method must be called
     * prior to the <code>ObjectGrid.initialize()</code> method.  Also, keep in mind that the
     * <code>ObjectGrid.getSession()</code> method implicitly calls the
     * <code>ObjectGrid.initialize()</code> method if it has yet to be called by the application.
     *
     * @param seconds  the number of seconds a map entry is allowed to live in map
     *                 before being evicted.
     *
     * @throws IllegalArgumentException if seconds is less than 0.
     * @throws IllegalStateException if this method is called after the
     *         <code>ObjectGrid.initialize()</code> method is called.
     *
     * @see #setTtlEvictorType(TTLType)
     * @see ObjectMap#setTimeToLive
     * @see ObjectGrid#initialize()
     * @see ObjectGrid#getSession()
     */
    public void setTimeToLive(int seconds);

    /**
     * Gets the number of seconds for an entry to live.
     * <p>
     * This value returned is in seconds and 0 indicates forever.
     *
     * @return the same value passed to the <code>setTimeToLive(int)</code> method or
     *         0 if <code>setLockTimeout</code> was never called.
     *
     * @see #setTimeToLive(int)
     */
    public int getTimeToLive();

    /**
     * Sets how expiration time of a <code>BackingMap</code> entry is computed.
     * <p>
     * If this method is not called, <code>TTLType.NONE</code> is used
     * to indicate the map entry has no expiration time (e.g. is allowed to live
     * until explicitly removed or invalidated by the application, or evicted
     * by a user defined <code>Evictor</code>).
     * <p>
     * Note, to avoid an <code>IllegalStateException</code>, this method must be called
     * prior to the <code>ObjectGrid.initialize()</code> method.  Also, keep in mind that the
     * <code>ObjectGrid.getSession()</code> method implicitly calls the
     * <code>ObjectGrid.initialize()</code> method if it has yet to be called by the application.
     *
     * @param type must be one of the public constants declared in the
     *        <code>TTLType</code> class.
     *
     * @throws IllegalStateException if this method is called after the
     *         <code>ObjectGrid.initialize()</code> method is called.
     *
     * @see TTLType
     * @see ObjectGrid#initialize()
     * @see ObjectGrid#getSession()
     */
    public void setTtlEvictorType(TTLType type);

    /**
     * Gets how expiration time of a <code>BackingMap</code> entry is computed.
     * <p>
     * @return the <code>TTLType</code> that was passed to the <code>setTtlEvictorType(TTLType)</code> or
     * <code>TTLType.NONE</code> if <code>setTtlEvictorType</code> was never called.
     *
     * @see #setTtlEvictorType(TTLType)
     * @see TTLType
     */
    public TTLType getTtlEvictorType();

    /**
     * Creates a dynamic index on the <code>BackingMap</code>.
     * <p>
     *
     * Required Client Permission: {@link ServerMapPermission#REPLICATE}
     *
     * @param name the name of the index. The name can not be null or a zero length string.
     *
     * @param isRangeIndex Indicate whether to create a <code>MapRangeIndex</code> or a <code>MapIndex</code>.
     *                     If set to true, the index will be a type of <code>MapRangeIndex</code>.
     *
     * @param attributeName The name of the attribute to be indexed. The attributeName can not be null or a zero length string.
     *
     * @param dynamicIndexCallback The callback that will invoke upon dynamic index events.  The dynamicIndexCallback
     *        is optional and can be null.
     *
     * @throws IllegalArgumentException if name or attributeName is null or a zero length string.
     *
     * @throws IndexAlreadyDefinedException if a <code>MapIndexPlugin</code> with the specified name already exists.
     *
     * @see MapIndex
     * @see MapIndexPlugin
     * @see MapRangeIndex
     * @see ObjectMap#getIndex(String)
     * @since WAS XD 6.0.1
     */
    public void createDynamicIndex(String name, boolean isRangeIndex, String attributeName,
            DynamicIndexCallback dynamicIndexCallback) throws IndexAlreadyDefinedException, IllegalArgumentException;

    /**
     * Creates a dynamic index on the <code>BackingMap</code>.
     * <p>
     *
     * Required Client Permission: {@link ServerMapPermission#DYNAMIC_INDEX}
     *
     * <p>
     * A MapIndexPlugin that implements the {@link BackingMapLifecycleListener} is automatically added
     * as if the {@link #addMapEventListener(EventListener)} method was called. Any previous
     * index which implements {@link BackingMapLifecycleListener} is removed as if
     * the {@link #removeMapEventListener(EventListener)} method was called.
     *
     * @param index The index implementation. The index can not be <code>null</code>.
     *
     * @param dynamicIndexCallback The callback that will invoke upon dynamic index events.  The dynamicIndexCallback
     *        is optional and can be null.
     *
     * @throws IllegalArgumentException if index is <code>null</code> or
     *                                  <code>index.getName()</code> returns
     *                                  <code>null</code> or a zero length
     *                                  string.
     *
     * @throws IndexAlreadyDefinedException if a <code>MapIndexPlugin</code> with the specified name already exists.
     *
     * @see MapIndexPlugin
     * @see ObjectMap#getIndex(String)
     * @since WAS XD 6.0.1
     */
    public void createDynamicIndex(MapIndexPlugin index, DynamicIndexCallback dynamicIndexCallback)
            throws IndexAlreadyDefinedException, IllegalArgumentException;

    /**
     * Removes a dynamic index on the <code>BackingMap</code>.
     * <p>
     *
     * Required Client Permission: {@link ServerMapPermission#DYNAMIC_INDEX}
     *
     * @param name the name of the index. The name can not be <code>null</code>.
     *
     * @throws IllegalArgumentException if name is <code>null</code>.
     *
     * @throws IndexUndefinedException if a <code>MapIndexPlugin</code> with the specified name does not exists.
     *
     * @see #createDynamicIndex(MapIndexPlugin, DynamicIndexCallback)
     * @see #createDynamicIndex(String, boolean, String, DynamicIndexCallback)
     * @since WAS XD 6.0.1
     */
    public void removeDynamicIndex(String name) throws IndexUndefinedException, IllegalArgumentException;

    /**
     * Allows access to the <code>PartitionManager</code> that is defined for this
     * <code>BackingMap</code>.  This access may be useful for <code>Loader<code>s during
     * <code>Loader.preloadMap(Session, BackingMap)</code> processing (to properly partition
     * the data to be loaded).
     *
     * @return <code>PartitionManager</code> associated with this <code>BackingMap</code>.
     *
     * @see PartitionManager
     * @see Loader#preloadMap(Session, BackingMap)
     * @since WAS XD 6.0.1
     */
    public PartitionManager getPartitionManager();

    /**
     * Retreive the metadata for the entity associated with this backing map.
     * @return the EntityMetadata if an entity is associated with this backing map or null if there is no
     * entity associated with this backing map.
     * @since WAS XD 6.1
     */
    public EntityMetadata getEntityMetadata();

    /**
     * Enable write-behind updates for this map.
     * <P>
     * If a map is configured with write-behind loader update, the updates (could be insert type,
     * remove type, or update type) to the backend are not instantly updated to the back end by
     * calling the {@link com.ibm.websphere.objectgrid.plugins.Loader#batchUpdate(TxID, LogSequence)}
     * method. Instead, they are queued in a write-behind queue map and updated to
     * the back end periodically.
     * <p>
     * A write-behind update is pushed to the backend periodically within a different transaction
     * from the one the update is made to ObjectGrid. When the write-behind update to the
     * backend fails, for example, due to data integrity problem, it is too late to roll back
     * the original ObjectGrid transaction. ObjectGrid will invalidate the entry and create
     * an entry in a failed database update map. The name of this failed database update map is
     * {@link WriteBehindLoaderConstants#WRITE_BEHIND_FAILED_UPDATES_MAP_PREFIX}+<tt>baseMapName</tt>.
     * The key of the entry in this map is an auto-increment Integer, and the value is a
     * {@link com.ibm.websphere.objectgrid.plugins.LogElement}. The logElement can be used to
     * compensate the failure.
     * <p>
     * Note, to avoid an <code>IllegalStateException</code>, this method must be called
     * prior to the <code>ObjectGrid.initialize()</code> method.  Also, keep in mind that the
     * <code>ObjectGrid.getSession()</code> method implicitly calls the
     * <code>ObjectGrid.initialize()</code> method if it has yet to be called by the application.
     *
     * @param writeBehindParam a write-behind parameter consisting of a maximum update time and/or
     * a maximum key update count. The format of the write-behind parameter is "T[time];C[count]",
     * for example, "T100;C2000". "T100;C2000" means the loader will write to the back end when there
     * are 2000 pending keys to be updated or when 100 seconds have passed since the last
     * update. The default update time is 300 seconds and the default update key count
     * is 1000. You can configure the update time only, the update key count only, or an
     * empty string. The default value(s) will then be used in either of the above three cases.
     * The default value is <tt>null</tt> to disable write-behind updates.
     *
     * @throws IllegalArgumentException if the write behind parameters are unknown or improperly formatted.
     * @throws IllegalStateException if this method is called after the
     *         <code>ObjectGrid.initialize()</code> method is called.
     *
     * @see WriteBehindLoaderConstants
     *
     * @since WAS XD 6.1.0.3
     */
    public void setWriteBehind(String writeBehindParam);

    /**
     * Get the write-behind parameter. A write-behind parameter consists of a maximum update time and/or
     * a maximum key update count. The format of the write-behind parameter is "T[time];C[count]".
     *
     * @return the write-behind parameter. If the write-behind parameter is not set, <tt>null</tt> will be returned.
     *
     * @see #setWriteBehind(String)
     *
     * @since WAS XD 6.1.0.3
     */
    public String getWriteBehind();

    /**
     * Set the time-based database update configuration object.
     * <p>
     * When a time-based database update configuration object is set, a thread will
     * be started automatically to update or invalidate the ObjectGrid maps with the latest updates
     * (inserts and updates) from the database.
     * <p>
     * For a local ObjectGrid map, the thread will be launched in the same JVM.
     * For a distributed ObjectGrid map in an ObjectGrid container, the thread
     * will be automatically launched in partition 0. No database update thread will be started in a
     * client side near cache.
     * <p>
     * Note, to avoid an <code>IllegalStateException</code>, this method must be called
     * prior to the <code>ObjectGrid.initialize()</code> method.  Also, keep in mind that the
     * <code>ObjectGrid.getSession()</code> method implicitly calls the
     * <code>ObjectGrid.initialize()</code> method if it has yet to be called by the application.
     *
     * @param dbUpdateConfig the time-based database update configuration object or <tt>null</tt>.
     *
     * @throws IllegalStateException if this method is called after the
     *         <code>ObjectGrid.initialize()</code> method is called.
     *
     * @since WAS XD 6.1.0.3
     */
    public void setTimeBasedDBUpdateConfig(TimeBasedDBUpdateConfig dbUpdateConfig);

    /**
     * Get the time-based database update configuration object.
     * @return the time-based database update configuration object or <tt>null</tt> if not set.
     *
     * @see #setTimeBasedDBUpdateConfig(TimeBasedDBUpdateConfig)
     *
     * @since WAS XD 6.1.0.3
     */
    public TimeBasedDBUpdateConfig getTimeBasedDBUpdateConfig();

    /**
     * Constant used to indicate this map is not a distributed map.
     * @see #getMapType()
     * @since WAS XD 6.1
     */
    public static final int LOCAL = 0;

    /**
     * Constant used to indicate this map is a server map.
     * @see #getMapType()
     * @since WAS XD 6.1
     */
    public static final int SERVER = 1;

    /**
     * Constant used to indicate this map is a client to a server map
     * @see #getMapType()
     * @since WAS XD 6.1
     */
    public static final int CLIENT = 2;

    /**
     * Returns the type of BackingMap.
     * <p>
     * The return value is equivalent to one of the constants declared on this interface,
     * {@link #LOCAL}, {@link #SERVER}, or {@link #CLIENT}.
     *
     * @return the map type
     *
     * @since WAS XD 6.1
     */
    public int getMapType();

    /**
     * Returns the types of additional eviction triggers.<p>
     *
     * The available eviction trigger strings are a described
     * in the String constants in this interface that begin with the name: <tt>EVICTIONTRIGGER</tt>.
     *
     * @return a semicolon separated list of eviction triggers
     *
     * @since WAS XD 6.1.0.3
     */
    public String getEvictionTriggers();

    /**
     * Sets the types of additional eviction triggers, all evictors for the backing map will use
     * the provided set of triggers.<p>
     *
     * The available eviction trigger strings are a described
     * in the String constants in this interface that begin with the name: <tt>EVICTIONTRIGGER</tt>.<p>
     *
     * Note, to avoid an <code>IllegalStateException</code>, this method must be called
     * prior to the <code>ObjectGrid.initialize()</code> method.  Also, keep in mind that the
     * <code>ObjectGrid.getSession()</code> method implicitly calls the
     * <code>ObjectGrid.initialize()</code> method if it has yet to be called by the application.
     *
     * @param evictionTriggers a semicolon separated list of eviction triggers
     * @throws IllegalStateException if this method is called after the
     *         <code>ObjectGrid.initialize()</code> method is called.
     * @throws IllegalArgumentException if the eviction triggers are unknown or improperly formatted.
     * @since WAS XD 6.1.0.3
     */
    public void setEvictionTriggers(String evictionTriggers);

    /**
     * Retrieves the name of the MapSet that this BackingMap is currently associated with. A "null" return value indicates
       * it is currently not associated with a MapSet.  This method will only return a non null value for a client or server map.
        *
        * @return name of associated MapSet
     * @see #getMapType()
        * @since 7.1
        */
    public String getMapSetName();

    /**
     * Retrieve the SerializerAccessor for this map.
     * @return the SerializerAccessor
     * @since 7.1.1
     */
    public SerializerAccessor getSerializerAccessor();


    /**
     * Retrieve the current life cycle state of this map.
     * @return the current state.
     * @since 7.1.1
     */
    public BackingMapLifecycleListener.State getState();

    /**
     * If true, clients with local caches are automatically invalidated when the data grid map
     * is updated.
     * @return true if client near cache invalidation is enabled.
     * @since 8.6
     */
    public boolean isNearCacheInvalidationEnabled();

    /**
     * If true, clients automatically send time-to-live access information
     * to the remote data grid when accessed and the {@link TTLType#LAST_ACCESS_TIME}
     * TTL evictor type is configured.
     * @return true if last-access time-to-live information is sent to the remote data grid.
     * @since 8.6
     */
    public boolean isNearCacheLastAccessTTLSyncEnabled();

    /**
     * If true, the client near cache is enabled for supported configurations.
     * The client near cache can only be enabled when using optimistic locking or
     * when locking is disabled.
     * @return true if the client near cache is enabled.
     */
    public boolean isNearCacheEnabled();
    /**
     * Retrieves the data format for all data access APIs that return cache keys.
     * @return the data format or null if the default should be used.
     * @since 8.6
     */
    public OutputFormat getKeyOutputFormat();
    
    /**
     * Sets the data format for all data access APIs that return cache keys.
     * @param dataFormat the data format to use or null to use the default.
     * @since 8.6
     */
    public void setKeyOutputFormat(OutputFormat dataFormat);

    /**
     * Retrieves the data format for all data access APIs that return cache values.
     * @return the data format.
     * @since 8.6
     */
    public OutputFormat getValueOutputFormat();

    /**
     * Sets the data format for all data access APIs that return cache values.
     * @param dataFormat the data format to use or null to use the default.
     * @since 8.6
     */
    public void setValueOutputFormat(OutputFormat dataFormat); 

}