// IBM Confidential OCO Source Material
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2005,2007
// The source code for this program is not published or otherwise divested
// of its trade secrets, irrespective of what has been deposited with the
// U.S. Copyright Office.
//
// $Id$
//
// Module  :  Session.java
//
// Source File Description: Reference JavaDoc
//
// Change Activity:
//
// Reason       Version   Date         Userid    Change Description
// ------------ --------- ------------ --------- -----------------------------------------
// XD60_M1_B.1  WASX.XD   01/14/2005   sutter    new code drop
// XD60_M3_B.10 WASX.XD   04/07/2005   sutter    Re-factor datagrid/datagrid.impl
// XD60_M3_B.8  WASX.XD   04/13/2005   sutter    M3 Client/Server updates
// 279166       WASX.XD   06/15/2005   sutter    Moved processLogSequence to interface
// M3           WASX.XD   10/12/2005   pdykes    Add support for isFlushing, isCommiting.
// XD601_M3_B.1 WASX.XD   10/19/2005   kjlaw     added markRollbackOnly method.
// DOW          WASX.XD   11/18/2005   kjlaw     add transaction timeout feature.
// d344949      WXD601.XD 02/06/2006   jhanders  throw NoActiveTransactionException on flush()
// b834         6.1       04/26/2006   cdjohnson Added createObjectQuery(String)
// bug 10175    6.1 ifix3 10/12/2007   jhanders  Add isTransactionActive to Session interface
// b10197       6.1.0.1   01/03/2007   jhanders  Add get/setTransactionIsolation support to Session interface.
// b18296       7.1       02/05/2010   cheng1    SessionHandle support for fixed_partitions
// b18296       7.1.1.1   12/06/2011   cdjohnson add getTransactionType()
// ------------ --------- ------------ --------- -----------------------------------------

package com.ibm.websphere.objectgrid;

import javax.security.auth.Subject;

import com.ibm.websphere.objectgrid.client.ClientProperties;
import com.ibm.websphere.objectgrid.em.EntityManager;
import com.ibm.websphere.objectgrid.plugins.LogSequence;
import com.ibm.websphere.objectgrid.plugins.ObjectGridEventListener;
import com.ibm.websphere.objectgrid.query.ObjectQuery;
import com.ibm.websphere.objectgrid.query.ObjectQueryException;
import com.ibm.websphere.objectgrid.security.ObjectGridPermission;
import com.ibm.websphere.objectgrid.security.plugins.CredentialGenerator;

/**
 * This interface represents a session container for <code>ObjectMap</code>s. A thread must get its own Session
 * object to interact with ObjectGrid. You can think of this interface as a session that can only be used by a single
 * thread at a time. A Session itself is shareable across threads so long as only one thread uses it at a time. However,
 * if a J2EE connection/transaction infrastructure is being used, that won't be shareable across threads and will
 * prevent the Session object from being shared across threads. A good analogy for this object is a JDBC connection to a
 * database.
 *
 * For best performance, use the {@link #close()} method to close the session once it is no longer required.
 *
 * @since WAS XD 6.0, XC10
 *
 * @see ObjectGrid#getSession()
 * @see ObjectGrid#getSession(Subject)
 * @see ObjectGrid#getSession(CredentialGenerator)
 */
public interface Session {
    /** A string indicating the default transaction type */
    public final static String TRANSACTION_TYPE_DEFAULT = "DEFAULT";

    /**
     * A special value for the timeout parameter of the <code>setTransactionTimeout(int)</code> method. This special
     * value is used to indicate that the next transaction started by this Session is allowed unlimited amount of time.
     *
     * @see #setTransactionTimeout(int)
     */
    public final int TRANSACTION_NO_TIMEOUT = 0;
    public static final long DEFAULT_RETRY_TIMEOUT = 30000;

    /**
     * A transaction isolation level constant indicating that dirty reads and
     * non-repeatable reads are prevented; phantom reads can occur.
     *
     * This level prohibits a transaction from reading an uncommitted cache entry, and it also prohibits
     * the situation where one transaction reads an entry, a second transaction alters the entry, and the first transaction
     * rereads the entry, getting different values the second time (a "non-repeatable read").
     *
     * @see #setTransactionIsolation(int)
     * @since WAS XD 6.1.0.1
     */
    public static final int TRANSACTION_REPEATABLE_READ = 4;

    /**
     * A transaction isolation level constant indicating that dirty reads are prevented;
     * non-repeatable reads and phantom reads can occur.
     * This level only prohibits a transaction from reading a cache entry with uncommitted changes in it.
     *
     * @see #setTransactionIsolation(int)
     * @since WAS XD 6.1.0.1
     */
    public static final int TRANSACTION_READ_COMMITTED = 2;

    /**
     * A transaction isolation level constant indicating that dirty reads,
     * non-repeatable reads and phantom reads can occur.
     * This level allows a cache entry changed by one transaction to be read by another transaction before
     * any changes in that entry have been committed (a "dirty read").
     *
     * If any of the changes are rolled back, the second transaction will have retrieved an invalid entry.
     *
     * @see #setTransactionIsolation(int)
     * @since WAS XD 6.1.0.1
     */
    public static final int TRANSACTION_READ_UNCOMMITTED = 1;

    /**
     * The commit protocols that can be used to commit the Session's transaction 
     */
    public enum TxCommitProtocol {
    	
    	/**
         * A commit protocol constant indicating that the Session transaction can read from multiple partitions 
         * but can only write to a single partition. A TransactionException is thrown if the transaction writes to 
         * multiple partitions. The transaction is committed using the one-phase commit protocol.          
         */
    	ONEPHASE,
    	
    	 /**
    	  * A commit protocol constant indicating that the Session transaction can read and write from multiple 
    	  * partitions. The transaction is committed using the two-phase commit protocol. If the transaction 
    	  * only writes to a single partition then the transaction is committed using the one-phase commit protocol.
    	  */         
    	TWOPHASE    	
    }
    
   
    /**
     * Starts a new transaction that does not write changes through to a <code>Loader</code> or ObjectGrid server.
     * <p>
     * Changes made by the session transaction started by this method are only applied to the <code>BackingMap</code>
     * and not given to the <code>Loader</code>. This method can be used to apply changes made in a peer cache to the
     * local <code>BackingMap</code> only. In addition, with a distributed map, this method can be used to start a
     * session transaction which changes will only be applied to the client <code>BackingMap</code>, but not the
     * <code>BackingMap</code> on the server side.
     *
     * @throws TransactionAlreadyActiveException
     *             if there is already an active transaction
     * @throws TransactionException
     *             a <code>TransactionCallbackException</code> occurred or some other error occurred starting a new
     *             transaction
     */
    void beginNoWriteThrough() throws TransactionAlreadyActiveException, TransactionException;

    /**
     * Returns the ObjectMap for the specified name.
     * <p>
     * The ObjectMap is used to retrieve and modify values in the BackingMap. Multiple invocations of this method on the
     * same Session object will always return the same object.
     * <p>
     * This method can also be used to create a BackingMap and its associated ObjectGrid after ObjectGrid initialization.
     * If cacheName does not match the name of a previously created map, a name comparison will be executed against template
     * maps that have been configured.  The ObjectMap and BackingMap will be created if the name matches the regular expression
     * of a template.<p>
     *
     * Required Client Permission: {@link ObjectGridPermission#DYNAMIC_MAP} (when creating a new map from a template)
     *
     * @param cacheName
     *            name of desired map
     * @return ObjectMap the transactional interface to modify values in the map
     *
     * @throws UndefinedMapException
     *             if the map is not defined.
     *
     * @see ObjectGrid#defineMap(String)
     * @see ObjectMap
     */
    ObjectMap getMap(String cacheName) throws UndefinedMapException;

    /**
     * Begins a new transaction.
     *
     * @throws TransactionAlreadyActiveException
     *             if this method is invoked with an active transaction
     * @throws TransactionException
     *             a <code>TransactionCallbackException</code> occurred or some other error occurred starting a new
     *             transaction
     */
    void begin() throws TransactionAlreadyActiveException, TransactionException;

    /**
     * Commits a transaction.
     *
     * @throws NoActiveTransactionException
     *             if this method is invoked with no active transaction
     * @throws TransactionException
     *             if an error occurred during commit processing, see the caused by to determine the root error
     *
     * @see #markRollbackOnly(Throwable)
     */
    public void commit() throws NoActiveTransactionException, TransactionException;

    /**
     * Rolls back a transaction.
     *
     * @throws NoActiveTransactionException
     *             if this method is invoked with no active transaction
     * @throws TransactionException
     *             if an error occurred during rollback processing, see the caused by to determine the root error
     */
    public void rollback() throws NoActiveTransactionException, TransactionException;

    /**
     * Forces the current changes in the Session to the <code>Loader</code> or ObjectGrid server. This method does not
     * commit the changes, it just applies the changes.
     *
     * @throws NoActiveTransactionException
     *             if this method is invoked with no active transaction
     * @throws TransactionException
     *             if an error occurred during flush processing, see the caused by to determine the root error
     */
    public void flush() throws NoActiveTransactionException, TransactionException;

    /**
     * Returns the ObjectGrid that owns this session.
     *
     * @return the owning ObjectGrid instance.
     */
    ObjectGrid getObjectGrid();

    /**
     * Determines if a transaction is currently active.
     *
     * @return true if a transaction is currently active for this session.
     * @since WAS XD 6.1 FIX3
     */

    boolean isTransactionActive();

    /**
     * Gets the <code>TxID</code> transaction identifier, if a transaction is active.
     *
     * @return The current TxID object.
     *
     * @throws NoActiveTransactionException
     *             if this method is invoked with no active transaction
     */
    TxID getTxID() throws NoActiveTransactionException;

    /**
     * Returns whether the current session transaction is writing through to the back end <code>Loader</code> or
     * ObjectGrid server(<code>true</code>), or if the changes are only applying to the <code>BackingMap</code> (<code>false</code>)
     * or client respectively.
     *
     * @return <code>true</code>, if write through is enabled
     *
     * @see #begin()
     * @see #beginNoWriteThrough()
     */
    boolean isWriteThroughEnabled();

    /**
     * Sets the transaction type for future transactions.
     * <p>
     * After this method is called, all future transactions will have the same type until another transaction type is
     * set. If no transaction type is set, the default transaction type <code>TRANSACTION_TYPE_DEFAULT</code> will be
     * used.
     * </p>
     *
     * <p>
     * Transaction types are used mainly for statistical data tracking purpose. Users can predefine types of
     * transactions that will be executed in an application. The idea is to categorize transactions with the same
     * characteristics to one category (type), so one transaction response time statistics can be used to track each
     * transaction type. This approach is useful when your application has different types of transactions. Some types
     * of transactions, such as update transactions, process longer than others transactions, such as read-only
     * transactions. By using the transaction type, different transactions are tracked by different statistics, so the
     * statistics can be more useful.
     * </p>
     *
     * @param tranType
     *            the transaction type for future transactions.
     *
     * @see #TRANSACTION_TYPE_DEFAULT
     */
    void setTransactionType(String tranType);

    /**
     * Retrieves the transaction type that is set with the {@linkplain #setTransactionType(String)} method.
     *
     * @return the transaction type for the session.
     * @since 7.1.1.1
     */
    String getTransactionType();

    /**
     * Processes a <code>LogSequence</code>.
     * <p>
     * Each <code>LogElement</code> within the <code>LogSequence</code> will be examined and the appropriate
     * operation (insert, update, invalidate, etc) will be performed against the <code>BackingMap</code> identified by
     * the <code>LogSequence</code>'s map name. An ObjectGrid Session must be active before this method is invoked.
     * The caller is responsible for issuing the appropriate commit or rollback invocation to complete the Session.
     * Autocommit processing is not available for this method invocation.
     * <p>
     * The main use of this method is for processing a <code>LogSequence</code> that was received by a remote JVM. For
     * example, using the Distributed Commit support, the <code>LogSequence</code>s associated with a given committed
     * Session are distributed to other listening <code>ObjectGrid</code>s in other JVMs. After receiving the
     * <code>LogSequence</code>s at the remote JVM, the listener could start a Session using
     * <code>beginNoWriteThrough()</code>, invoke this method, and commit the Session transaction.
     *
     * @param logSequence
     *            LogSequence of changes to be applied to an active transaction
     *
     * @throws NoActiveTransactionException
     *             if this method is invoked with no active transaction
     * @throws UndefinedMapException
     *             if the map referenced by the LogSequence cannot be found
     * @throws ObjectGridException
     *             if the LogSequence elements cannot be processed
     *
     * @see #beginNoWriteThrough()
     * @see LogSequence
     * @see ObjectGridEventListener
     */
    public void processLogSequence(LogSequence logSequence) throws NoActiveTransactionException, UndefinedMapException,
            ObjectGridException;

    /**
     * Returns whether the current session transaction is performing a session <code>flush()</code> operation. It is
     * helpful to know if a session <code>flush()</code> is active (<code>true</code>), or if only an
     * <code>ObjectMap.flush()</code> is in progress (returns <code>false</code> in this case).
     *
     * @return true, if the session is executing a session <code>flush()</code> call.
     *
     * @see #flush()
     * @see ObjectMap#flush()
     * @since WAS XD 6.0.1
     */
    boolean isFlushing();

    /**
     * Returns whether the current session transaction is performing a session <code>commit()</code> operation. It is
     * helpful to know if a session commit is active (<code>true</code>), or if an <code>ObjectMap.flush()</code>
     * or session <code>flush()</code> is in progress (returns <code>false</code> in these cases).
     *
     * @return true, if session is executing a session <code>commit()</code> call.
     *
     * @see #commit()
     * @see #flush()
     * @see ObjectMap#flush()
     * @since WAS XD 6.0.1
     */
    boolean isCommitting();

    /**
     * Marks the current transaction as being rollback only.
     * <p>
     * Marking a transaction rollback only ensures that even if the <code>commit()</code> method is called for this
     * session transaction, the transaction is rolled back. A rollback is typically done when either ObjectGrid itself
     * or the application knows that data corruption could occur if the <code>commit()</code> method was allowed to
     * commit the transaction. Once this method is called, the Throwable object that is passed to it is chained to the
     * <code>TransactionException</code> that is thrown if the commit method is ever called. Any subsequent calls to
     * this method for the current active transaction is ignored (e.g. only the first call that passes a non null
     * Throwable reference is used). Once the transaction is completed, the rollback only mark is removed so that the
     * next transaction that is started using this session can be committed.
     *
     * @param t
     *            the Throwable that caused this method to be called.
     *
     * @throws NoActiveTransactionException
     *             if there is no active transaction for this Session.
     *
     * @see #commit()
     * @see TransactionException
     * @since WAS XD 6.0.1
     */
    public void markRollbackOnly(Throwable t) throws NoActiveTransactionException;

    /**
     * Returns whether or not the current active session transaction is marked as being rollback only as a result of a
     * prior call to the <code>markRollbackOnly(Throwable)</code> method on this Session.
     *
     * @return true if and only if current session transaction is marked rollback only.
     *
     * @see #markRollbackOnly(Throwable)
     * @since WAS XD 6.0.1
     */
    public boolean isMarkedRollbackOnly();

    /**
     * Sets the transaction timeout for the next transaction started by this Session object to a specified number of
     * seconds.
     * <p>
     * This method does not affect the transaction timeout of any transactions previously started by this Session. It
     * only affects transactions that are started after this method is called. If this method is never called, the
     * ObjectGrid configured transaction timeout value is used.
     *
     * @param timeout
     *            is the transaction timeout value in seconds. Use the special value <code>TRANSACTION_NO_TIMEOUT</code>
     *            if transaction is allowed unlimited amount of time and no transaction timeout should occur.
     *
     * @see #TRANSACTION_NO_TIMEOUT
     * @see ObjectGrid#setTxTimeout(int)
     * @see TransactionTimeoutException
     * @since WAS XD 6.0.1
     */
    public void setTransactionTimeout(int timeout);

    /**
     * Gets current transaction timeout for this Session.
     * <p>
     * The transaction timeout value returned is the value that was configured for the <code>ObjectGrid</code> using
     * <code>ObjectGrid.setTxTimeout(int)</code> or the value passed to <code>setTransactionTimeout(int)</code> to
     * override the value configured on <code>ObjectGrid</code>. The return value is in seconds.
     *
     * @return timeout value in seconds.
     *
     * @see #setTransactionTimeout(int)
     * @see ObjectGrid#setTxTimeout(int)
     * @since WAS XD 6.0.1
     */
    public int getTransactionTimeout();

    /**
     * Retrieves the current transaction isolation level for this session.
     *
     * @return one of the following <tt>Session</tt> constants:
     * {@link #TRANSACTION_READ_UNCOMMITTED}, {@link #TRANSACTION_READ_COMMITTED} or {@link #TRANSACTION_REPEATABLE_READ}
     * @since WAS XD 6.1.0.1
     */
    public int getTransactionIsolation();

    /**
     * Determines whether the current session transaction has timed out.
     *
     * @return true if and only if transaction has timed out.
     *
     * @see #setTransactionTimeout(int)
     * @since WAS XD 6.0.1
     */
    public boolean transactionTimedOut();

    /**
     * Creates an instance of an object query for executing a query over the
     * ObjectMaps visible to this session.
     * <p>
     * When ObjectGrid security is enabled, this method requires an com.ibm.websphere.objectgrid.security.ObjectGridPermission
     * with action "query".<p>
     *
     * Required Client Permission: {@link ObjectGridPermission#QUERY}
     *
     * @param qlString
     *            a query string
     *
     * @return the new query instance.
     *
     * @throws ObjectQueryException
     *             if an error occurs creating the object query.
     * @since WAS XD 6.1
     */
    public ObjectQuery createObjectQuery(String qlString) throws ObjectQueryException;

    /**
     * Retrieve the EntityManager associated with this Session.  Each session is associated
     * with a single EntityManager instance.  Repeated calls to this method on the same Session instance
     * will result in the same EntityManager instance.
     * @return this session's EntityManager instance.
     * @since WAS XD 6.1
     */
    public EntityManager getEntityManager();


    /**
     * Attempts to change the transaction isolation level for this session.   The constants defined in the <tt>Session</tt>
     * interface are the possible transaction isolation levels.<p>
     *
     * This method should normally be invoked prior to beginning a transaction.  Invoking after a transaction
     * has started may result in an exception.
     *
     * @param level one of the following <tt>Session</tt> constants:
     * {@link #TRANSACTION_READ_UNCOMMITTED}, {@link #TRANSACTION_READ_COMMITTED} or {@link #TRANSACTION_REPEATABLE_READ}
     * @since WAS XD 6.1.0.1
     */
    void setTransactionIsolation(int level);

//    /**
//     * setPreferZones - client uses to control where router should send the request.
//     *
//     * @param exclusive - only send the request to targets on the listed zones if set the exclusive to be true.
//     * @param failoverOnly - select targets in another zones only when targets on the current zone is not available if set failoverOly to be true
//     * @param zones - array of zone names
//     * @throws TargetNotAvailableInZonesException
//     *
//     * @since WAS XD 6.1.0.3
//     */
//    void setPreferZones(boolean exclusive, boolean failoverOnly, String[] zones)throws TargetNotAvailableException;
//
//
//    /**
//     * setPreferZones - client uses to control where router should send the request.
//     *
//     * @param exclusive - only send the request to targets on the listed zones if set the exclusive to be true.
//     * @param zones - array of zone names
//     * @throws TargetNotAvailableInZonesException
//     *
//     * @since WAS XD 6.1.0.3
//     */
//    void setPreferZones(boolean exclusive, String[] zones)throws TargetNotAvailableException;
//
//    /**
//     * setPreferZones - client uses to control where router should send the request.
//     *
//     * @param zones - array of zone names
//     * @throws TargetNotAvailableInZonesException
//     *
//     * @since WAS XD 6.1.0.3
//     */
//    void setPreferZones(String[] zones)throws TargetNotAvailableException;
//
//
//    /**
//     *  set Prefer Local JVM
//     *
//     *  If there is shard in local JVM that is appropriate to client request,
//     *  this local shard will be used first this method is set to true.
//     *
//     * @param localJVM
//     *
//     * @since WAS XD 6.1.0.3
//     */
//     void setPreferLocalJVM(boolean localJVM);
//    /**
//     * set Prefer Local Host
//     *
//     * If there are shards in local host machine,
//     *  this local shard will spray among these shards
//     *  on local host machines.
//     *
//     * @param localHost
//     *
//     * @since WAS XD 6.1.0.3
//     */
//    void setPreferLocalHost(boolean localHost);

    /**
     * Retrieves a handle for this session.
     * <p>
     * A SessionHandle contains partition information for the current session and can be re-applied to a new session
     * using the {@link #setSessionHandle(SessionHandle)} method.  A SessionHandle is
     * only applicable for ObjectGrids using per-container partition placement.
     * If {@link #setSessionHandle(SessionHandle)} is not called before invoking this
     * method, a Session Handle is selected using the properties configured in the
     * {@link com.ibm.websphere.objectgrid.client.ClientProperties}.  If there are no per-container partition placement mapsets
     * or more than one in the ObjectGrid, an <code>IllegalStateException</code> is thrown.
     *
     * @return the SessionHandle for this session
     * @throws IllegalStateException if this method is called in an invalid environment.
     * @since WAS XD 6.1.0.3
     */
    public SessionHandle getSessionHandle();

    /**
     * Apply a SessionHandle to this session.
     *
     * @param target the SessionHandle to apply or null to disassociate a SessionHandle from
     * this session.
     * @throws TargetNotAvailableException when the target is no longer available.
     * @throws IllegalStateException if the Session has modified some maps already and the SessionHandle has already been set
     *                               or if this method is called in an invalidate environment.
     * @since WAS XD 6.1.0.3
     */

    public void setSessionHandle(SessionHandle target) throws TargetNotAvailableException;

//
//	/**
//	 * Create a SessionHandle for a specific partition.
//	 *
//	 * @param partitionID the partition id number to associate with this handle.
//	 * @return the SessionHandle for the partition.
//     * @since WAS XD 6.1.0.3
//	 */
//	public SessionHandle createSessionHandle(int partitionID);

    /**
     * Set the request retry timeout to indicate how long to retry a request
     * (in milliseconds) when recoverable failures occur, such as fail-over
     * exceptions.  A request will timeout when either the request timeout expires
     * or the transaction timeout expires, whichever expires first.<p>
     *
     * A value of 0 indicates that all requests should fail immediately
     * and avoid any retry logic. Exceptions that
     * cannot succeed even if tried again such as DuplicateKeyException exceptions
     * will be thrown immediately.
     * <p>
     *
     * A value of -1 indicates that the request retry timeout is not set, meaning that
     * the request duration is governed by the request retry timeout set on the
     * ClientProperties.  If the ClientProperties is also set to -1, then the
     * request retry timeout is governed by the transaction timeout.<p>
     *
     * @param requestRetryTimeout
     *            the duration in milliseconds retry a client request, 0 if the request should
     *            fail immediately or -1 if the request timeout is not set.
     * @see ClientProperties#setRequestRetryTimeout(long)
     * @see #setTransactionTimeout(int)
     * @since 7.0
     */
    public void setRequestRetryTimeout(long requestRetryTimeout);

    /**
     * Retrieves the current request retry timeout for this session. Returns -1 if it was not set.
     *
     * @return
     *            the duration in milliseconds retry a client request, 0 if the request should
     *            fail immediately or -1 if the request timeout is not set.
     * @since 7.0
     */
    public long getRequestRetryTimeout();

    /**
     * Determines if a SessionHandle is currently set on this Session.
     *
     * @return true if a SessionHandle is currently set on this session.
     *
     * @since 7.1
     */

    public boolean isSessionHandleSet();

    /**
     * Closes this session, freeing all resources that are held. Once closed,
     * this session must be discarded. Use one of the
     * {@link ObjectGrid#getSession()} methods to retrieve a new session.
     *
     * If the session has an active transaction, the transaction will be rolled
     * back and the session resources are not freed.
     *
     * @throws ObjectGridRuntimeException
     *             thrown if there is a problem releasing resources held by this
     *             session.
     * @since 7.1.1
     */
    public void close();
    
    /**
     * Set the commit protocol to be used when committing this Session's transaction. The constants defined in the <tt>TxCommitProtocol</tt>
     * enum are the possible commit protocols.<p>
     *
     * This method should normally be invoked prior to beginning a transaction. Invoking after a transaction
     * has started will result in an exception.
     *
     * @param protocol one of the following constants <tt>TxCommitProtocol.ONEPHASE</tt> or <tt>TxCommitProtocol.TWOPHASE</tt>
     * @since WAS XD 8.6
     */
    public void setTxCommitProtocol(TxCommitProtocol protocol);
    
    /**
     * Retrieve the current commit protocol for this Session.
     * 
     * @return one of the following constants <tt>TxCommitProtocol.ONEPHASE</tt> or <tt>TxCommitProtocol.TWOPHASE</tt> 
     * @since WAS XD 8.6
     */
    public TxCommitProtocol getTxCommitProtocol();
}
