// IBM Confidential OCO Source Material
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2012
// The source code for this program is not published or otherwise divested
// of its trade secrets, irrespective of what has been deposited with the
// U.S. Copyright Office.
//
// $Id$
//
// Module  :  ContinuousQueryManager.java
//
// Source File Description:
//
// See javadoc for class.
//
// Change Activity:
//
// Reason       Version   Date         Userid    Change Description
// ------------ --------- ------------ --------- -----------------------------------------
// ------------ --------- ------------ --------- -----------------------------------------

package com.ibm.websphere.objectgrid.continuousquery;

import java.util.Collection;
import java.util.List;

import com.ibm.websphere.objectgrid.UndefinedMapException;
import com.ibm.websphere.objectgrid.continuousquery.exception.ContinuousQueryIncompatibleDuplicateException;
import com.ibm.websphere.objectgrid.continuousquery.filter.AbstractCQFilter;
import com.ibm.websphere.objectgrid.plugins.io.DataSerializer;
import com.ibm.websphere.objectgrid.plugins.io.dataobject.SerializedEntry;

/**
 * The management interface used to define new continuous queries, retrieve an existing query, or destroy a query. An
 * instance can be retrieved via
 * {@link ContinuousQueryManagerFactory#getManager(com.ibm.websphere.objectgrid.ObjectGrid)}
 * 
 * @since 8.6
 */
public interface ContinuousQueryManager {

    /**
     * Define a new continuous query. This query will run in the grid container processes. When cache entries that match
     * the filter criteria are inserted or deleted, the client's {@link ContinuousQueryCache} will be updated and any
     * {@link ContinuousQueryListener} implementations will be invoked. This will also occur if a cache entry is updated
     * such that it now matches, or no longer matches, the filter criteria. Optionally, notifications can be sent for
     * update operations to cache entries that previously matched the criteria, and still match the criteria after the
     * update.
     * 
     * @param mapName
     *            - the name of the map the continuous query will be defined on.
     * @param filter
     *            - the filter that the will determine the contents of the continuous query. Custom filter logic can be
     *            implemented by extending {@link AbstractCQFilter}.
     * @param keysOnly
     *            - true if the query should only return the keys of items that satisfy the query; false to also send
     *            the values as well as the keys. If false, the {@link ContinuousQueryCache} and
     *            {@link ContinuousQueryNotificationEvent} objects will include the value associated with cache entries
     *            that match the query.
     * @param returnInitialResultSet
     *            - true to run the filter on all existing matches in the map at the time of the query definition and
     *            return the items to the cache. If false, the client will only be notified of matches which occur after
     *            the query is defined. Returning the initial result set will have a high computational requirement for
     *            large maps.
     * @param notifyOnUpdated
     *            - true causes this topic to get called every time an existing item that satisfies this query gets
     *            updated and continues to satisfy this query. This will drive {@link ContinuousQueryListener} call
     *            backs as well as updated values in the query cache if keysOnly is false.
     * @param continuousQueryListeners
     *            - used to register listeners before the query is defined. These listeners will be guaranteed to be
     *            invoked for any entries in the initial result set, as well as changes to the result set which might
     *            occur before {@link ContinuousQueryTopic#addListener(ContinuousQueryListener)} can be invoked. Can be
     *            null.
     * @param noCache
     *            - true to indicate that no caching of keys or values should occur on the client. Continuous Query
     *            listeners will still be notified.
     * @param useRawFormat
     * 			  - when true, keys and values returned from the {@link ContinuousQueryCache} and {@link ContinuousQueryListener}
     * 				will be in {@link SerializedEntry} format when a {@link DataSerializer} is defined on the map.
     * 			  - when false, or when no {@link DataSerializer} is defined, keys and values will be returned as
     * 				Objects.
     * @param partitionSubset
     *            - An optional parameter used to indicate that this continuous query should only be defined on the
     *            specified subset of partitions. Passing no arguments indicates this continuous query should be defined
     *            on all existing and future partitions containing this map. An IllegalArgumentException may be thrown
     *            if an invalid partition ID is passed.
     * @return The {@link ContinuousQueryTopic} for this continuous query. If the call to this method would result in an
     *         topic which is identical to an existing topic being created, the existing instance is returned.
     * @throws ContinuousQueryIncompatibleDuplicateException
     * @throws UndefinedMapException
     * @throws IllegalArgumentException
     */

    public <KeyType, ValueType> ContinuousQueryTopic<KeyType, ValueType> defineContinuousQuery(String mapName,
            ContinuousQueryFilter filter, boolean keysOnly, boolean returnInitialResultSet, boolean notifyOnUpdated,
            Collection<ContinuousQueryListener<KeyType, ValueType>> continuousQueryListeners, boolean noCache,
            boolean useRawFormat,Integer... partitionSubset) throws ContinuousQueryIncompatibleDuplicateException, UndefinedMapException;

    /**
     * Define a new continuous query. This query will run in the grid container processes. When cache entries that match
     * the filter criteria are inserted or deleted, the client's {@link ContinuousQueryCache} will be updated and any
     * {@link ContinuousQueryListener} implementations will be invoked. This will also occur if a cache entry is updated
     * such that it now matches, or no longer matches, the filter criteria. Optionally, notifications can be sent for
     * update operations to cache entries that previously matched the criteria, and still match the criteria after the
     * update.
     * 
     * @param mapName
     *            - the name of the map the continuous query will be defined on.
     * @param filter
     *            - the filter that the will determine the contents of the continuous query. Custom filter logic can be
     *            implemented by extending {@link AbstractCQFilter}.
     * @param keysOnly
     *            - true if the query should only return the keys of items that satisfy the query; false to also send
     *            the values as well as the keys. If false, the {@link ContinuousQueryCache} and
     *            {@link ContinuousQueryNotificationEvent} objects will include the value associated with cache entries
     *            that match the query.
     * @param returnInitialResultSet
     *            - true to run the filter on all existing matches in the map at the time of the query definition and
     *            return the items to the cache. If false, the client will only be notified of matches which occur after
     *            the query is defined. Returning the initial result set will have a high computational requirement for
     *            large maps.
     * @param notifyOnUpdated
     *            - true causes this topic to get called every time an existing item that satisfies this query gets
     *            updated and continues to satisfy this query. This will drive {@link ContinuousQueryListener} call
     *            backs as well as updated values in the query cache if keysOnly is false.
     * @param partitionSubset
     *            - An optional parameter used to indicate that this continuous query should only be defined on the
     *            specified subset of partitions. Passing no arguments indicates this continuous query should be defined
     *            on all existing and future partitions containing this map. An IllegalArgumentException may be thrown
     *            if an invalid partition ID is passed.
     * @return The {@link ContinuousQueryTopic} for this continuous query. If the call to this method would result in an
     *         topic which is identical to an existing topic being created, the existing instance is returned.
     * @throws ContinuousQueryIncompatibleDuplicateException
     * @throws UndefinedMapException
     * @throws IllegalArgumentException
     */

    public <KeyType, ValueType> ContinuousQueryTopic<KeyType, ValueType> defineContinuousQuery(String mapName,
            ContinuousQueryFilter filter, boolean keysOnly, boolean returnInitialResultSet, boolean notifyOnUpdated,
            Integer... partitionSubset) throws ContinuousQueryIncompatibleDuplicateException, UndefinedMapException;

    /**
     * Remove a defined continuous query. Does the opposite of
     * {@link #defineContinuousQuery(String, ContinuousQueryFilter, boolean, boolean, boolean, int...)}. If the number
     * of calls to this method for a given topic equals the number of calls to defineContinuousQuery that return the
     * specified topic, the topic will cease receiving updates from the server, all listeners of this topic will be
     * removed and the reference to this topic will be invalid.
     * 
     * @param topic
     *            - the topic to remove
     * @param partitionSubset
     *            - If the continuous query was defined on a subset of partitions, that same set of partitions must be
     *            passed for the remove to succeed. If no subset was specified when the continuous query was defined,
     *            then no subset should be specified for the remove to succeed. An IllegalArgumentException may be
     *            thrown if an invalid partition ID is passed.
     * @return true if the topic was removed, false otherwise
     * @throws IllegalArgumentException
     */
    public boolean removeContinuousQuery(ContinuousQueryTopic topic, Integer... partitionSubset);

    /**
     * Returns a list view of the currently defined Continuous Query Topics.
     * 
     * @return a list of Continuous Query Topics
     */
    public List<ContinuousQueryTopic> getDefinedContinuousQueries();

}
