// IBM Confidential OCO Source Material
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2012
// The source code for this program is not published or otherwise divested
// of its trade secrets, irrespective of what has been deposited with the
// U.S. Copyright Office.
//
// $Id$
//
// Module  :  AbstractCQFilter.java
//
// Source File Description:
//
// An abstract filter for testing if a specified object passed the filter.
package com.ibm.websphere.objectgrid.continuousquery.filter;

import java.io.IOException;
import java.lang.reflect.AccessibleObject;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.util.StringTokenizer;

import com.ibm.ejs.ras.Tr;
import com.ibm.ejs.ras.TraceComponent;
import com.ibm.websphere.objectgrid.continuousquery.ContinuousQueryFilter;
import com.ibm.websphere.objectgrid.continuousquery.exception.ContinuousQueryAttributePathException;
import com.ibm.websphere.objectgrid.continuousquery.exception.ContinuousQueryException;
import com.ibm.websphere.objectgrid.continuousquery.exception.ContinuousQueryGetValueException;
import com.ibm.websphere.objectgrid.io.XsDataInputStream;
import com.ibm.websphere.objectgrid.plugins.io.DataSerializer.DataAttributeInflatable;
import com.ibm.websphere.objectgrid.plugins.io.DataSerializer;
import com.ibm.websphere.objectgrid.plugins.io.MapSerializerPlugin;
import com.ibm.websphere.objectgrid.plugins.io.datadescriptor.DataDescriptor;
import com.ibm.websphere.objectgrid.plugins.io.datadescriptor.MapDataDescriptor;
import com.ibm.websphere.objectgrid.plugins.io.dataobject.DataObjectContext;
import com.ibm.websphere.objectgrid.plugins.io.dataobject.SerializedEntry;
import com.ibm.websphere.projector.Tuple;
import com.ibm.ws.objectgrid.Constants;
import com.ibm.ws.xs.NLSConstants;

/**
 * This is an abstract class which all continuous query filters must extend. 
 * It defines how cached entries should be compared and filtered.
 * 
 * All subclasses must provide a no-argument constructor for serialization purposes.
 * 
 * @param <KeyType>  Type of the key object for the map being queried
 * @param <ValueType> Type of the value object for the map being queried
 * @param <AttributeType> Type of the attribute referenced by the attribute path
 * @param <MatchType> Type of the object being compared to
 * @since 8.6
 */
public abstract class AbstractCQFilter<KeyType,ValueType,AttributeType,MatchType> implements ContinuousQueryFilter<KeyType,ValueType,AttributeType,MatchType> {

    private static final long serialVersionUID = -8449526738355405243L;

    private static final TraceComponent tc = Tr.register(AbstractCQFilter.class, 
            Constants.TR_CONTINUOUSQUERY_GROUP_NAME, Constants.TR_RESOURCE_BUNDLE_NAME);
    
    // Cache information for less reflection and serializer calls
    private Boolean initLock = false;
    private transient boolean initialized = false;
    private transient DataAttributeInflatable serializer = null;
    private transient String[] attributeArray = null; 
    private transient AccessibleObject[] cachedAttributes = null;
    private transient String trimedAttributePath = null;
    private transient boolean keyFilter = false; 
    private transient String addressibleKeyName;
    private transient String pathSeparator;
    
    /**
     * Version of the filter implementation, which can be used
     * internally for backward compatibility.
     * 
     * The version can be used to, for example, determine what filtering
     * logic should be used, how serialization/inflation should be done, etc.
     * 
     * Version 1 filters are shipped with WebSphere eXtreme Scale 8.6.0.0.
     */
    protected int version = 1;
        
    public boolean isKeyFilter() {
        return keyFilter;
    }

    /**
     * Creates a string representation of the continuous query filter, this method is
     * used by {@linkplain #toString()}.
     * <p>
     * The string will be used for identifying and comparing filters,
     * i.e. it will be used in {@linkplain #hashCode()} and {@linkplain #equals(Object)},
     * so the returned value must be in unique format among all filters
     * in the system.
     * 
     * @return an unique string representation of the filter
     */
    public abstract String createString();
    
    public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null || obj.getClass() != this.getClass()) {
            return false;
        }
        if (this.toString().equals(obj.toString())) {
            return true;
        }
        return false;
    }
    
    public int hashCode() {
        return this.toString().hashCode();
    }
    
    public String toString() {
        return "[" + version + "]" + createString();
    }
    
    /**
     * Ensure that attribute path used to retrieve an attribute from a cache entry is valid.
     * 
     * @param attributePath the path to the attribute
     * @throws ContinuousQueryAttributePathException
     */
    protected void validateAttributePath(String attributePath) throws ContinuousQueryAttributePathException {
        if (attributePath == null || attributePath.isEmpty()) {
            // it means compare the full value object
            return;
        }
        if (attributePath.trim().matches(".*\\s.*")) {
            throw new ContinuousQueryAttributePathException(NLSConstants.CANNOT_USE_SPACE_IN_ATTRIBUTE_PATH_CWOBJ7751, attributePath);
        }
        // make sure we have matching parentheses
        for (int i = 0; i < attributePath.length(); i++) {
            switch (attributePath.charAt(i)) {
            case '(': 
                if (i+1 == attributePath.length() || attributePath.charAt(i+1) != ')') {
                    // ( isn't followed by )
                    throw new ContinuousQueryAttributePathException(NLSConstants.UNMATCHED_PARENTHESES_IN_ATTRIBUTE_PATH_CWOBJ7755, attributePath);
                }
                break;
            case ')':
                if (i == 0 || attributePath.charAt(i-1) != '(') {
                    // ( isn't followed by )
                    throw new ContinuousQueryAttributePathException(NLSConstants.UNMATCHED_PARENTHESES_IN_ATTRIBUTE_PATH_CWOBJ7755, attributePath);
                }
            }
        }
    }
    
    /**
     * Retrieves an attribute from the content that is to be filtered based on the specified attribute path.<p>
     * 
     * @param content The cache entry (key and value) to be processed by this filter
     * @param attributePath the path specifying the attribute of this cache entry to be examined.  
     * 						For a description of attribute paths, @see CompareFilter#CompareFilter(String, Object)
     * @return the attribute specified by the attributePath, extracted from the cache entry represented by the FilterContent instance.
     */
    protected Object getAttribute(FilterContent content, String attributePath) throws ContinuousQueryException {
        if (content.getKey() instanceof Tuple) {
            throw new ContinuousQueryGetValueException(NLSConstants.ENTITY_MAP_NOT_SUPPORTED_CWOBJ7750);
        }
        
        if (attributePath == null || attributePath.trim().isEmpty()) {
        	if(content.getValue() instanceof SerializedEntry) {
        		// pass back the inflated NATIVE value object, not the RAW value representation
        		// since that is the only thing that makes sense to compare to.
        		return ((SerializedEntry)content.getValue()).getObject();
        	}
        	else {
        		return content.getValue();
        	}
        } 

        trimedAttributePath = attributePath.trim();
        synchronized (initLock) {
            if (!initialized) {
                validateAttributePath(trimedAttributePath);
                if (content.getKey() instanceof SerializedEntry) {
                    MapSerializerPlugin mapSerializerPlugin = ((SerializedEntry) content.getKey()).getContext().getSerializerAccessor().getMapSerializerPlugin();
                    addressibleKeyName = mapSerializerPlugin.getMapDataDescriptor().getAddressableKeyName();
                    pathSeparator = mapSerializerPlugin.getKeySerializerPlugin().getKeyDataDescriptor().getPathSeparator();
                } else {
                    addressibleKeyName = ContinuousQueryFilter.POJO_ADDRESSABLEKEYNAME;
                    pathSeparator = ContinuousQueryFilter.POJO_PATHSEPARATOR;
                }
                
                String keyPrefix = addressibleKeyName + pathSeparator;
                if (attributePath.startsWith(keyPrefix)) {
                    trimedAttributePath = attributePath.substring(keyPrefix.length());
                    keyFilter = true;
                }
                
                Object object = content.getValue();
                if (keyFilter) {
                    object = content.getKey();
                }
                
                if (object instanceof SerializedEntry) {
                    // get the data serializer, addressible key name and path separator
                    SerializedEntry serEntry = (SerializedEntry) object;
                    
                    if (serializer == null) {
                        MapSerializerPlugin mapSerializerPlugin = serEntry.getContext().getSerializerAccessor().getMapSerializerPlugin();
                        
                        if (keyFilter) {
                            serializer = mapSerializerPlugin.getKeySerializerPlugin();
                            pathSeparator = mapSerializerPlugin.getKeySerializerPlugin().getKeyDataDescriptor().getPathSeparator();
                        } else {
                            serializer = mapSerializerPlugin.getValueSerializerPlugin();
                            pathSeparator = mapSerializerPlugin.getValueSerializerPlugin().getValueDataDescriptor().getPathSeparator();
                        }
                        if (serializer == null) {
                            throw new ContinuousQueryGetValueException(NLSConstants.NO_SERIALIZER_PROVIDED_CWOBJ7760);
                        }
                        
                    }
                }
                initialized = true;
            }
        }
        
        if (trimedAttributePath.equals(addressibleKeyName)) {
            return content.getKey();
        }
        
        if (keyFilter) {
            return getValueFromObject(content.getKey());
        } else {
            return getValueFromObject(content.getValue());
        }
    }
    
    private Object getValueFromObject(Object object) throws ContinuousQueryGetValueException {
        
    	// XDF or custom serializer
        if (object instanceof SerializedEntry) {
            SerializedEntry serEntry = (SerializedEntry) object;
            Object attrCtx = serializer.getAttributeContexts(trimedAttributePath);
            try {
                Object[] result = serializer.inflateDataObjectAttributes(serEntry.getContext(),
                        serEntry.getInputStream(), attrCtx);
                if (result != null && result.length > 0) {
                    if(tc.isDebugEnabled()){
                    	if(DataSerializer.SpecialValue.NOT_FOUND.equals(result[0]))
                    		Tr.debug(tc,"Attribute not found: "+trimedAttributePath+" result="+result[0]);
                    }
                	/*
                	if(DataSerializer.SpecialValue.NOT_FOUND.equals(result[0]))
                	{
                		if(tc.isDebugEnabled()) Tr.debug(tc,"Invalid attribute, throwing error");
                		throw new ContinuousQueryGetValueException(NLSConstants.CANNOT_GET_SERIALIZED_ATTRIBUTE_CWOBJ7753, 
                                new Object[]{trimedAttributePath, serializer.getClass().getName()});
                	}
                	*/
                    return result[0];
                }
                throw new ContinuousQueryGetValueException(NLSConstants.CANNOT_GET_SERIALIZED_ATTRIBUTE_CWOBJ7753, 
                        new Object[]{trimedAttributePath, serializer.getClass().getName()});
                
            } catch (IOException e) {
                throw new ContinuousQueryGetValueException(e, NLSConstants.CANNOT_GET_SERIALIZED_ATTRIBUTE_CWOBJ7753, 
                        new Object[]{trimedAttributePath, serializer.getClass().getName()});
            }
        }
        
        // Handle POJO
        return getValueFromPOJO(object);
    }
    
    private Object getValueFromPOJO(Object object) throws ContinuousQueryGetValueException {

        synchronized (initLock) {
            if (cachedAttributes == null) {
                StringTokenizer tokenizer = new StringTokenizer(trimedAttributePath, pathSeparator);
                attributeArray = new String[tokenizer.countTokens()];
                int index = 0;
                while (tokenizer.hasMoreElements()) {
                    attributeArray[index++] = tokenizer.nextToken();
                }
                cachedAttributes = new AccessibleObject[attributeArray.length];
            }
        }
        Object nextObj = object;
        for (int i = 0; i < cachedAttributes.length; i++) {
            if (nextObj == null) {
            	/*
                throw new ContinuousQueryGetValueException(NLSConstants.CANNOT_GET_ATTRIBUTE_CWOBJ7752, 
                        new Object[]{attributeArray[i], null});
                */
                if (tc.isDebugEnabled()) Tr.debug(tc, "Object is null, cannot process the next attribute '"+attributeArray[i]+"'");
            	return DataSerializer.SpecialValue.NOT_FOUND;
            }
            boolean isMethod = false;
            if (cachedAttributes[i] == null) {
                // Handle POJO field
                int methodIndex = attributeArray[i].indexOf('(');
                if (methodIndex > 0) {
                    // It's a method call
                    String methodName = attributeArray[i].substring(0, methodIndex);
                    try {
                        cachedAttributes[i] = getMethod(nextObj.getClass(), methodName);
                        isMethod = true;
                    } catch (Exception e) {
                        throw new ContinuousQueryGetValueException(e, NLSConstants.CANNOT_INVOKE_METHOD_CWOBJ7754, 
                                new Object[]{methodName, nextObj.getClass().getName()});
                    }
                } else {
                    String methodName = "get" + Character.toUpperCase(attributeArray[i].charAt(0)) + attributeArray[i].substring(1);
                    try {
                        // It's a field or property
                        Field field = getField(nextObj.getClass(), attributeArray[i]);
                        
                        // Check if public field
                        if (Modifier.isPublic(field.getModifiers())) {
                            cachedAttributes[i] = field;
                        } else {
                            // Need to access property through getter method 
                            if (field.getType().equals(boolean.class) || field.getType().equals(Boolean.class)) {
                                methodName = "is" + Character.toUpperCase(attributeArray[i].charAt(0)) + attributeArray[i].substring(1);
                            }
                            cachedAttributes[i] = getMethod(nextObj.getClass(), methodName);
                            isMethod = true;
                        }
                    } catch (NoSuchFieldException e) {
                        if (tc.isDebugEnabled()) Tr.debug(tc, "Field '"+attributeArray[i]+"' is not defined for class "+nextObj.getClass());
                        throw new ContinuousQueryGetValueException(e, NLSConstants.CANNOT_GET_ATTRIBUTE_CWOBJ7752, 
                                new Object[]{attributeArray[i], nextObj.getClass().getName()});
                        
                    } catch (NoSuchMethodException e) {
                        if (tc.isDebugEnabled()) Tr.debug(tc, "Public member method '"+methodName+"' not defined for class "+nextObj.getClass());
                        throw new ContinuousQueryGetValueException(e, NLSConstants.CANNOT_GET_ATTRIBUTE_CWOBJ7752, 
                                new Object[]{attributeArray[i], nextObj.getClass().getName()});
                        
                    } catch (Exception e) {
                        if (tc.isDebugEnabled()) Tr.debug(tc, "Failed to access field '"+attributeArray[i]+"', exception: "+e.getMessage());
                        throw new ContinuousQueryGetValueException(e, NLSConstants.CANNOT_GET_ATTRIBUTE_CWOBJ7752, 
                                new Object[]{attributeArray[i], nextObj.getClass().getName()});
                        
                    }
                }
            }
            if (isMethod || cachedAttributes[i] instanceof Method) {
                Method method = (Method) cachedAttributes[i];
                try {
                    nextObj = method.invoke(nextObj);
                } catch (Exception e) {
                    if (tc.isDebugEnabled()) Tr.debug(tc, "Failed to invoke method '"+method.getName()+"', exception: "+e.getMessage());
                    throw new ContinuousQueryGetValueException(e, NLSConstants.CANNOT_INVOKE_METHOD_CWOBJ7754, 
                            new Object[]{method.getName(), nextObj.getClass().getName()});
                }
            } else {
                Field field = (Field) cachedAttributes[i];
                try {
                    nextObj = field.get(nextObj);
                } catch (Exception e) {
                    if (tc.isDebugEnabled()) Tr.debug(tc, "Failed to access field '"+field.getName()+"', exception: "+e.getMessage());
                    throw new ContinuousQueryGetValueException(e, NLSConstants.CANNOT_GET_ATTRIBUTE_CWOBJ7752, 
                            new Object[]{field.getName(), nextObj.getClass().getName()});
                }
            }
        }
        return nextObj;
    }

    @SuppressWarnings("rawtypes")
    private static Field getField(Class clazz, String fieldName) throws NoSuchFieldException {
        try {
            return clazz.getDeclaredField(fieldName);
        } catch (NoSuchFieldException e) {
            Class superClass = clazz.getSuperclass();
            if (superClass == null) {
                throw e;
            } else {
                return getField(superClass, fieldName);
            }
        }
    }

    @SuppressWarnings({ "rawtypes", "unchecked" })
    private static Method getMethod(Class clazz, String methodName) throws NoSuchMethodException {
        try {
            return clazz.getMethod(methodName);
        } catch (NoSuchMethodException e) {
            Class superClass = clazz.getSuperclass();
            if (superClass == null) {
                throw e;
            } else {
                return getMethod(superClass, methodName);
            }
        }
    }
}
