// IBM Confidential OCO Source Material
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2007
// The source code for this program is not published or otherwise divested
// of its trade secrets, irrespective of what has been deposited with the
// U.S. Copyright Office.
//
// $Id$
//
// Module  :  MapSet.java
//
// Source File Description:
//
//    Need a file description.
//
// Change Activity:
//
// Reason        Version   Date         Userid    Change Description
// ------------- --------- ------------ --------- -----------------------------------------
// b6555         6.1       08/21/2006   vandersn  make public
// b23976        7.1.1.0   06/30/2011   koehler   MMR: Intra-Domain support
// ------------- --------- ------------ --------- -----------------------------------------
package com.ibm.websphere.objectgrid.deployment;

import java.util.Collection;
import java.util.List;

import com.ibm.websphere.objectgrid.ObjectGridRuntimeException;
import com.ibm.websphere.objectgrid.config.BackingMapConfiguration;

/**
 * Represents a set of BackingMaps within an ObjectGrid.  These maps are grouped within the MapSet for the purpose of placing them
 * across the same set of containers.  The MapSet allows for customization of many placement behaviors. 
 *
 * @since WAS XD 6.1 FIX3
 */
public interface MapSet {
    /**
     * Represents a placement strategy where there is a fixed number of partitions that are balanced across the
     * available set of containers. This strategy can be used in either server core or client/server mode.  This is the
     * default placement strategy.
     */
    public static final byte PLACEMENT_STRATEGY_FIXED_PARTITIONS = 0;

    /**
     * Represents a placement strategy where each container gets a fixed number of primary shards and the replicas are
     * then balanced across the other containers. This strategy can be used only in server core mode as clients can not
     * be made to route in this mode.
     * 
     * The numInitialContainers attribute is ignored when this placement type is used.
     */
    public static final byte PLACEMENT_STRATEGY_PER_CONTAINER = 1;

       
//    /**
//     * If no scope is defined or if 'domain' is defined for the scope then the placement behavior will be the traditional behavior of 
//     * placement.  Each container in the grid will be eligible for placement of both primary and replica shards.
//     * 
//     */
//    public static final byte PLACEMENT_SCOPE_PER_DOMAIN = 2;
        
//    /**
//     * If 'zone' scope is defined then primaries will be balanced independently within each zone.  This means if there are two zones 
//     * then there will be two primaries for each partition.
//     * 
//     * PER_ZONE will not be implemented in the initial offering (7.1.1)
//     * 
//     * Container scope combined with per container placement strategy is
//     * an erroneous combination that will result in an error log and failed
//     * container deployment.
//     */
//    public static final byte PLACEMENT_SCOPE_PER_ZONE = 3;
        
//    /**
//     * If 'container' scope is defined then primaries will be placed on every container.  One side effect of this type of configuration 
//     * is that replica configuration is now ignored.  A message will be logged if minSync, maxSync, maxAsync are greater than zero.
//     * 
//     * While any number of partitions will be permitted generally it will only make sense to have one partition since every container 
//     * will have all of the data.  In some cases it may make sense to have more to chunk up the data for revision querying or other reasons.
//     * 
//     * Scope is only applicable with the fixed partition strategy.
//     */
//    public static final byte PLACEMENT_SCOPE_PER_CONTAINER = 4;
 
//    /**
//     * Represents a container placement topology is used to indicate an automatic link topology. If the RING topology is selected, 
//     * then a grid will be selected to be the ring and every grid will connect to two other grids.
//     * This is the default container placement topology.     
//     * 
//     * The PLACEMENT_SCOPE_TOPOLOGY_RING attribute is ignored when the container placement strategy PLACEMENT_SCOPE_PER_DOMAIN is used.
//     */
//    public static final byte PLACEMENT_SCOPE_TOPOLOGY_RING = 0;

//    /**
//     * Represents a container placement topology is used to indicate an automatic link topology. If the HUB topology is selected, 
//     * then a grid will be selected to be the hub and every other grid will connect to it.
//     *  
//     * The PLACEMENT_SCOPE_TOPOLOGY_HUB attribute is ignored when the container placement strategy PLACEMENT_SCOPE_PER_DOMAIN is used.   
//     */
//    public static final byte PLACEMENT_SCOPE_TOPOLOGY_HUB = 1;

    /**
     * Set the name of this MapSet
     * @param mapSetName the name of the MapSet
     */
    public void setName(String mapSetName);

    /**
     * Get the name of this MapSet
     * @return the name of the MapSet
     */
    public String getName();

    /**
     * Add a BackingMapConfiguration to this MapSet.  A MapSet must contain one or more BackingMapConfiguration objects.
     * @param backingMapConfig the BackingMapConfiguration to add.
     * @throws ObjectGridRuntimeException if the BackingMapConfiguration is not found in the containing ObjectGridDeployment object.
     */
    public void addMap(BackingMapConfiguration backingMapConfig);

    /**
     * Remove a BackingMapConfiguration from this MapSet
     * @param mapName name of the map to remove
     * @return the BackingMapConfiguration that was removed. Null is returned if no BackingMapConfiguration was found with a name that matches the mapName.
     */
    public BackingMapConfiguration removeMap(String mapName);

    /**
     * Get a BackinMapConfiguration that has been added to this MapSet.
     * @param mapName - name of the BackingMapConfiguration to retrieve.
     * @return the BackingMapConfiguration with the mapName specified
     */
    public BackingMapConfiguration getMap(String mapName);

    /**
     * Get a Collection of BackingMapConfiguration objects that have been added to this MapSet
     * @return a Collection of BackingMapConfiguration objects
     */
    public Collection getMaps();

    /**
     * Get the maximum number of asynchronous replicas for each partition in the MapSet.
     * @return the maximum number of asynchronous replicas for each partition.
     * @see #setMaxAsyncReplicas(int)
     */
    public int getMaxAsyncReplicas();

    /**
     * Set the maximum number of asynchronous replicas for each partition in the MapSet.  This number
     * must be greater than or equal to 0.
     * @param maxAsyncReplicas the maximum number of asynchronous replicas for each partition
     * @see #getMaxAsyncReplicas()
     */
    public void setMaxAsyncReplicas(int maxAsyncReplicas);

    /**
     * Get the maximum number of synchronous replicas for each partition in the MapSet.
     * @return the maximum number of synchronous replicas for each partition
     * @see #setMaxSyncReplicas(int)
     */
    public int getMaxSyncReplicas();

    /**
     * Set the maximum number of synchronous replicas for each partition in the MapSet.  This number must be
     * greater than or equal to 0.
     * @param maxSyncReplicas the maximum number of synchronous replicas for each partition.
     */
    public void setMaxSyncReplicas(int maxSyncReplicas);

    /**
     * Get the minimum number of synchronous replicas for each partition in the MapSet.
     * @return the minimum number of synchronous replicas for each partition
     * @see #setMinSyncReplicas(int)
     */
    public int getMinSyncReplicas();

    /**
     * Set the minimum number of synchronous replicas for each partition in the MapSet.  This number must be greater
     * than or equal to 0.<p>
     *
     * Placement of this MapSet will not occur until there are containers available to support each partition's primary
     * as well as the number of minimum synchronous replicas specified.
     * @param minSyncReplicas - the minimum number of synchronous replicas for each partition
     */
    public void setMinSyncReplicas(int minSyncReplicas);

    /**
     * Get the number of partitions of this MapSet.  Each map within this MapSet will be divided into this many partitions.
     * @return the number of partitions
     * @see #setNumberOfPartitions(int)
     */
    public int getNumberOfPartitions();

    /**
     * Set the number of partitions for this MapSet.  This number must be greater than or equal to 1.
     * @param numberOfPartitions - each map within this MapSet will be divided into this number of partitions
     */
    public void setNumberOfPartitions(int numberOfPartitions);

    /**
     * Get the placement strategy employed by this MapSet.
     * @return The placement strategy employed by this MapSet.
     * @see #PLACEMENT_STRATEGY_FIXED_PARTITIONS
     * @see #PLACEMENT_STRATEGY_PER_CONTAINER
     */
    public byte getPlacementStrategy();
    
//    /**
//     * Get the placement scope employed by this MapSet.
//     * 
//     * Container scope combined with per container placement strategy is
//     * an erroneous combination that will result in an error log and failed
//     * container deployment.
//     * @see #PLACEMENT_SCOPE_PER_DOMAIN
//     * @see #PLACEMENT_SCOPE_PER_ZONE
//     * @see #PLACEMENT_SCOPE_PER_CONTAINER
//     */
//    public byte getPlacementScope();
     
    /**
     * Set the placement strategy that this MapSet should employ.
     * @param strategy The strategy that this MapSet should employ for placement.
     * @see #PLACEMENT_STRATEGY_FIXED_PARTITIONS
     * @see #PLACEMENT_STRATEGY_PER_CONTAINER
     */
    public void setPlacementStrategy(byte strategy);

//    /**
//     * Set the placement scope that this MapSet should employ.
//     * 
//     * Container scope combined with per container placement strategy is
//     * an erroneous combination that will result in an error log and failed
//     * container deployment.
//     * @see #PLACEMENT_SCOPE_PER_DOMAIN
//     * @see #PLACEMENT_SCOPE_PER_ZONE
//     * @see #PLACEMENT_SCOPE_PER_CONTAINER
//     */
//    public void setPlacementScope(byte scope);  

//    /**
//     * Get the container placement scope topology employed by this MapSet.
//     * @return The container placement scope topology employed by this MapSet..
//     * @see #PLACEMENT_SCOPE_TOPOLOGY_RING
//     * @see #PLACEMENT_SCOPE_TOPOLOGY_HUB
//     */
//    public byte getPlacementScopeTopology();

//    /**
//     * Set the container placement scope topology that this MapSet should employ.
//     * @param topology The topology that this MapSet should employ for container placement scope.
//     * @see #PLACEMENT_SCOPE_PER_DOMAIN
//     * @see #PLACEMENT_SCOPE_PER_ZONE
//     * @see #PLACEMENT_SCOPE_PER_CONTAINER
//     */
//    public void setPlacementScopeTopology(byte topology);
    
    /**
     * Enable or disable reading data from replicas.
     * @param replicaReadEnabled set to true to enable reading from replica.
     * Set to false to force all reads to occur on the primary.
     */
    public void setReplicaReadEnabled(boolean replicaReadEnabled);

    /**
     * Get the replica read enabled setting.
     * @return true if reading from replicas is enabled, false if all reads are routed to the primary.
     * @see #setReplicaReadEnabled(boolean)
     */
    public boolean isReplicaReadEnabled();

    /**
     * Set the number of containers that must be available to host this MapSet before initial placement will occur.
     * Before placement is triggered the available containers must also be able to support the minimum number of synchronous replicas.<p>
     *
     * This number must be greater than or equal to 1.  
     * 
     * The numInitialContainers is ignored if placement strategy is set to PLACEMENT_STRATEGY_PER_CONTAINER.
     * @param numInitialContainers the number of containers (1 or more) that must be available.
     * @see #setMinSyncReplicas(int)
     * @see #setPlacementStrategy(byte)
     */
    public void setNumInitialContainers(int numInitialContainers);

    /**
     * Get the initial number of containers.
     * @return the number of containers that must be available to host this MapSet before initial placement occurs.
     * @see #setNumInitialContainers(int)
     */
    public int getNumInitialContainers();

    /**
     * Setting this to true instructs the placement service to automatically attempt to replace any shards that
     * were lost as the result of container loss.<p>
     *
     * Setting this to false will cause lost shard to remain unplaced.<p>
     *
     * By default, lost shards will be automatically replaced.
     * @param autoReplaceLostShards set to false to prevent lost shards from automatically being replaced
     */
    public void setAutoReplaceLostShards(boolean autoReplaceLostShards);

    /**
     * Get the replace lost shards setting.
     *
     * @return true if lost shards will automatically be replaced upon the loss of a container
     * @see #setAutoReplaceLostShards(boolean)
     */
    public boolean isAutoReplaceLostShards();

    /**
     * When development mode is disabled, each shard from a single partition will be placed onto a container running on a different machine.
     * Enabling development mode allows shards from the same partition to be placed onto containers running on the same machine.
     * @param developmentMode
     */
    public void setDevelopmentMode(boolean developmentMode);

    /**
     * Get the development mode setting.
     * @return true if development mode is enabled
     * @see #setDevelopmentMode(boolean)
     */
    public boolean isDevelopmentMode();


    /**
     * Get a List of ShardMappings that have been added to this MapSet.
     *
     * @return a List of ShardMappings
     * @see #addShardMapping(ShardMapping)
     */
    public List getShardMappings();


    /**
     * Add a ShardMapping to this MapSet.  Only 1 primary ShardMapping can be added.<p>
     *
     * ShardMappings are not required.  If ShardMappings are used, there must be one for each shard in the partition.  One
     * for the primary shard, a ShardMapping for each synchronous replica up to the number of maximum synchronous replicas, and
     * a ShardMapping for each asynchronous replica up to the number of maximum asynchronous replicas.<p>
     *
     * The order in which the ShardMappings are added is significant.  Within each replica type (synchronous, asynchronous), the ShardMapping's
     * position in the List, relative to the same ShardType, will determine which shard the ShardMapping applies to.  For example,
     * the first ShardMapping added for a synchronous replica shard type will determine the placement behavior of the first synchronous
     * replica.<p>
     * @param shardMapping the ShardMapping to add.
     */
    public void addShardMapping(ShardMapping shardMapping);
}
