// IBM Confidential OCO Source Material
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2006-2009
// The source code for this program is not published or otherwise divested
// of its trade secrets, irrespective of what has been deposited with the
// U.S. Copyright Office.
//
// $Id$
//
// Module  :  QueryQueue.java
//
// Source File Description:
//
//    See javadoc.
//
// Change Activity:
//
// Reason       Version   Date         Userid    Change Description
// ------------ --------- ------------ --------- -----------------------------------------
// b9220     XD6.1.0.1 08/03/2007   jian.tang  Use command framework to execute query queue
// b17239    7.0.0.1   09/01/2009   jian.tang  Target: Evict Queryqueues using LRU algorithm 
// ---------------------------------------------------------------------------------------

package com.ibm.websphere.objectgrid.em;

import com.ibm.websphere.objectgrid.ObjectGridException;
import com.ibm.websphere.objectgrid.PartitionManager;
import com.ibm.websphere.projector.Tuple;

/**
 *
 * A query queue allows users to get entities qualified by a query in an iterative manner. A query queue is similar
 * to a query. It contains a query string, and named parameters or positioned parameters. However, it gets
 * the results in an iterative manner. In a client server environment, retrieval of entities to the client
 * is triggered by the {@link QueryQueue#getNextEntity(long)} or
 * {@link QueryQueue#getNextEntities(int, long)} method calls. A timeout value (in milliseconds) can be used in
 * these two methods to indicate the timeout value. Therefore, the user will not wait forever.
 * <p>
 * Setting parameters on the QueryQueue object can only be done before the {@link QueryQueue#getNextEntity(long)} or
 * {@link QueryQueue#getNextEntities(int, long)} method is called. A query queue is uniquely identified by the query string 
 * and the parameters. You cannot dynamic change the parameters for a query queue like you can do with a query. 
 * <p>
 * A partition indicates where the the query must be routed to in a distributed environment. If the partition
 * id is not specified, the QueryQueue will be routed to all partitions, starting from a random partition.
 * <p>
 * For example, the following code get the next Task entity for a specific task type and status.  
 * After the task is read, we assign it
 * to an agent, changing the status to STATUS_ASSIGNED, which removes it from the result set:
 * <pre>
 * QueryQueue queue = em.createQueryQueue("SELECT t FROM Tasks t WHERE t.type=?1 AND t.status=?2", Task.class);
 * queue.setParameter(1, TYPE_SALES_QUESTION);
 * queue.setParameter(2, STATUS_UNASSIGNED);
 *
 * EntityTransaction tran = em.getTransaction();
 * tran.begin();
 * Task nextTask = (Task) queue.getNextEntity(10000);
 * if (nextTask != null) {
 *     // Change the status of the task, assign to an agent.
 *     assignTask(task);
 * }
 * tran.commit();
 *
 * </pre>
 *
 *
 * In this example, the entity returned is of Task type. An entity class is used to specify
 * which class the result(s) will be projected to. The specified entity class must have the same entity
 * name as specified in the query string. If a null value is used as the entity class, then the result
 * will not be projected; the key Tuple(s) will be returned.
 * <p>
 * For the same example, you can use an subset entity type, such as SubTask, to
 * get a subset entity (an entity class with a subset of the attributes of the actual entity),
 * or you can use null to get the key Tuple(s).
 *
 * @see Tuple
 *
 * @since WAS XD 6.1 FIX3
 *
 */
public interface QueryQueue {

    /**
     * Used as a parameter on the {@link #getNextEntity(long)} or {@link #getNextEntity(long)} method.
     * It specifies the method should not block.
     */
    public static final long TIMEOUT_NONE = 0;

    /**
     * Used as a parameter on the {@link #getNextEntity(long)} or {@link #getNextEntity(long)} method.
     * It specifies the method should block until an entity becomes available.
     */
    public static final long TIMEOUT_INFINITE = Long.MAX_VALUE;

    /**
     * Bind an argument to a named parameter.
     *
     * @param name
     *            the parameter name
     * @param value the value of the parameter.
     * @return the same QueryQueue instance
     * @throws IllegalArgumentException
     *             if parameter name does not correspond to parameter in query string or argument is of incorrect type.
     *             This exception may be deferred until the QueryQueue is run.
     * @throws IllegalStateException 
     *             if this method is called after the {@link #getNextEntity(long)} or {@link #getNextEntities(int, long)} method is called.
     */
    QueryQueue setParameter(String name, Object value);

    /**
     * Bind an argument to a positional parameter.
     *
     * @param position the position of the parameter.  The first parameter is 1.
     * @param value the value of the parameter.
     * @return the same QueryQueue instance
     * @throws IllegalArgumentException
     *             if position does not correspond to positional parameter of query or argument is of incorrect type.
     *             This exception may be deferred until the QueryQueue is run.
     * @throws IllegalStateException 
     *             if this method is called after the {@link #getNextEntity(long)} or {@link #getNextEntities(int, long)} method is called.
     */
    QueryQueue setParameter(int position, Object value);

    /**
     * Set the partition to route the QueryQueue to.<p>
     *
     * Required if the maps in the QueryQueue are partitioned and the EntityManager
     * does not have affinity to a single schema root entity's partition.<p>
     *
     * Use the {@link PartitionManager} to determine the number of partitions
     * for a given entity's backing map.
     *<p>
     * The default partition ID for a QueryQueue is -1, which means the query is sent to
     * all partitions.
     * @param partitionId the partition to route the QueryQueue to.
     * @return the same QueryQueue instance
     */
    QueryQueue setPartition(int partitionId);

    /**
     * Get the next entity.
     * @param timeout the timeout value for this method call. The method will return after the specified time.
     * If the value is set to {@link #TIMEOUT_INFINITE}. it will wait until an entity is retrieved. If
     * the value is set to {@link #TIMEOUT_NONE}, it will not wait.
     * @see #TIMEOUT_NONE
     * @see #TIMEOUT_INFINITE
     * @return the next entity. null if there is no entity matched with the query queue.
     */
    Object getNextEntity(long timeout) throws ObjectGridException;

    /**
     * Get the next batch of entities. The number in this batch is specified by
     * the parameter numEntities.
     * <p>
     * The parameter numEntities is just a hint. It does not guarantee that the
     * actual number of returned entities. For example, if numEntities is 10, and there
     * are only 5 matched entities left in one partition, the number of returned
     * entities will be 5.  However, it does guarantee that the number of returned
     * entities will never exceed the specified number. If there is no matched entity,
     * an empty array will be returned.
     *
     * @param numEntities the number of entities to return
     * @param timeout the timeout value for this method call. The method will return after the specified time.
     * If the value is set to {@link #TIMEOUT_INFINITE}. it will wait until an entity is retrieved. If
     * the value is set to {@link #TIMEOUT_NONE}, it will not wait.
     * @see #TIMEOUT_NONE
     * @see #TIMEOUT_INFINITE
     * @return an array of entities. An 0-length array will be returned if there is no entity matched
     * with the QueryQueue query.
     */
    Object[] getNextEntities(int numEntities, long timeout) throws ObjectGridException;
}
