// IBM Confidential OCO Source Material
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2008,2009
// The source code for this program is not published or otherwise divested
// of its trade secrets, irrespective of what has been deposited with the
// U.S. Copyright Office.
//
// $Id$
//
// Source File Description:
//
//  Need a file description.
//
// Change Activity:
//
// Reason       Version   Date         Userid    Change Description
// ------------ --------- ------------ --------- -----------------------------------------
// d11207       WASX.XD   03/04/2008   bnewport  OpenJPA Load and Preload
// d12175       6.1.0.3   05/19/2008   jian.tang add persistencePropMap to ClientLoader.load method
// b12562       6.1.0.3   06/19/2008   jian.tang Support ManyToOne as Id
// b15138       6.1.0.5   02/24/2009   jian.tang  Throw IllegalArgumentException when the persistenceUnitName is not configured
// b14919       6.1.0.5   02/16/2009   jian.tang Throw OptimisticCollisionException at commit time
// b17362       7.0.0.0   09/15/2009   jian.tang Catch Throwable of JPA EM rollback in JPATxCallback
// ------------ --------- ------------ --------- -----------------------------------------
package com.ibm.websphere.objectgrid.jpa;

import java.util.Collection;
import java.util.Map;

import javax.persistence.EntityManager;
import javax.persistence.EntityManagerFactory;

import com.ibm.ejs.ras.Tr;
import com.ibm.ejs.ras.TraceComponent;
import com.ibm.websphere.objectgrid.ObjectGrid;
import com.ibm.websphere.objectgrid.OutputFormat;
import com.ibm.websphere.objectgrid.Session;
import com.ibm.websphere.objectgrid.TxID;
import com.ibm.websphere.objectgrid.plugins.ExceptionMapper;
import com.ibm.websphere.objectgrid.plugins.ObjectGridEventGroup;
import com.ibm.websphere.objectgrid.plugins.ObjectGridEventListener;
import com.ibm.websphere.objectgrid.plugins.OptimisticCollisionException;
import com.ibm.websphere.objectgrid.plugins.TransactionCallback;
import com.ibm.websphere.objectgrid.plugins.TransactionCallbackException;
import com.ibm.websphere.objectgrid.plugins.io.annotations.PluginOutputFormat;
import com.ibm.ws.ffdc.FFDCFilter;
import com.ibm.ws.objectgrid.Constants;
import com.ibm.ws.objectgrid.ObjectGridImpl;
import com.ibm.ws.objectgrid.ObjectGridManagerImpl;
import com.ibm.ws.objectgrid.em.EMFactoryImpl;
import com.ibm.ws.objectgrid.em.EntityManagerImpl;
import com.ibm.ws.objectgrid.jpa.DB2ConnectionHelper;
import com.ibm.ws.objectgrid.jpa.JPAEMFFactory;
import com.ibm.ws.objectgrid.jpa.JPAProviderHelperImpl;
import com.ibm.ws.objectgrid.jpa.SavedClientInfo;
import com.ibm.ws.objectgrid.plugins.ExceptionMapperWrapper;
import com.ibm.ws.objectgrid.runtime.RuntimeInfo;

/**
 * This transaction callback implementation is used for ObjectGrid built-in 
 * Java Persistence API (JPA) loaders.
 * It coordinates ObjectGrid transactions and JPA entity manager transactions.
 * <p>
 * An instance of JPATxCallback works with <code>JPALoader</code> or <code>JPAEntityLoader</code>
 * and place a JPA entity manager object in the slot on the <code>TxID</code> object that
 * <code>Loader</code>s can obtain. JPATxCallback is required to be configured on the ObjectGrid 
 * when <code>JPALoader</code> or <code>JPAEntityLoader</code> is used. Otherwise, a 
 * <ode>LoaderException</code> will be thrown.
 * <p>
 * When the ObjectGrid transaction commits or rollbacks, this transaction callback
 * will receive the events and commit or rollback the corresponding JPA entity manager
 * transactions.
 * <p>
 * A JPATxCallback has to be configured with the persistence unit name and optionally
 * a <code>JPAPropertyFactory</code> to create a JPA <code>EntityManagerFactory</code>. The 
 * <code>JPAPropertyFactory</code>  is used to create a JPA property map 
 * to override the default persistence properties.  
 * <p>
 * This transaction callback can have an inner transaction callback which is used
 * for a different purpose. This transaction callback will pass the transaction callback
 * events, such as initialize, begin, commit, and rollback, to the inner transaction
 * callback instance. When an inner transaction callback is used, users have to use spring
 * to configure the inner transaction callback instance and then set it on the
 * JPATxCallback instance. Here is an example:
 * <pre>
 *   &lt;bean id="innerTxCallback" class="..."&gt;
 *   &lt;/bean&gt
 *   &lt;bean id="jpaPropFactory" class=".." scope="shard"&gt
 *   &lt;/bean&gt
 *     &lt;bean id="jpaTxCallback" class="com.ibm.websphere.objectgrid.jpa.JPATxCallback" scope="shard" &gt
 *       &lt;property name="persistenceUnitName" value="employeePU"/&gt
 *       &lt;property name="innerTransactionCallback" ref="innerTxCallback"/&gt
 *       &lt;property name="JPAPropertyFactory" ref ="jpaPropFactory"/&gt
 *     &lt;/bean&gt
 * </pre>
 * <p>
 * For better mapping a generic <code>javax.persitence.PersistenceException</code> to a more consumable exception, 
 * an implementation of the <code>ExceptionMapper<code> can be configured as an attribute of this <code>JPATxCallback</code> 
 * object. This <code>ExceptionMapper</code> instance will also be used by the <code>JPALoader</code> or 
 * <code>JPAEntityLoader</code> to map JPA exceptions. If there are attributes need to be set on the <code>ExceptionMapper<code>
 * object, spring-style configuration has to be used. 

 * @since WAS XD 6.1.0.3
 * @see JPALoader
 * @see JPAEntityLoader
 * @see JPAPropertyFactory
 * @see ExceptionMapper
 */
@PluginOutputFormat(keyFormat=OutputFormat.RAW, valueFormat=OutputFormat.RAW)
public class JPATxCallback implements TransactionCallback, ObjectGridEventListener, ObjectGridEventGroup.ShardEvents {

    private final static String CLASS_NAME = JPATxCallback.class.getName();

    private final static TraceComponent tc = Tr.register(CLASS_NAME, Constants.TR_JPA_GROUP_NAME,
            Constants.TR_RESOURCE_BUNDLE_NAME);

    private final static TraceComponent tc_clientInfo = Tr.register(CLASS_NAME+"2", "ObjectGridClientInfo",
            Constants.TR_RESOURCE_BUNDLE_NAME);

    private int emSlot;

    private String punitName;

    private volatile EntityManagerFactory emf;

    private String ogName;

    protected TransactionCallback nestedCB;

    JPAPropertyFactory factory;

    // Whether the ObjectGrid contain entity maps. 
    // If the ObjectGrid contain entity maps, we will disable the CGLIB proxy
    private boolean isOGEM = false;

    EMFactoryImpl ogEMF = null;

    JPAProviderHelperImpl jpaProviderHelper = null;

    private Map propMap = null;

    ExceptionMapper exceptionMapper = null;

    private DB2ConnectionHelper db2ConnHelper;
    
    private int savedDB2ClientInfoSlot;
    
    private boolean firstCall = true;
    
    /**     
     * The default constructor
     */
    public JPATxCallback() {
    }

    /**
     * This method is called when the ObjectGrid transaction is started. The inner
     * transaction callback, if any, will be notified with this begin event.
     *
     * @param txid the ObjectGrid transaction context ID
     *
     * @see TransactionCallback#begin(TxID)
     */
    public void begin(TxID txid) throws TransactionCallbackException {
        if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled())
            Tr.debug(tc, "begin " + txid);

        // Do not generate proxies
        if (isOGEM && ogEMF.getState() == EMFactoryImpl.STATE_ACTIVATED) {
            ((EntityManagerImpl) txid.getSession().getEntityManager()).generateProxies(false);
        }
        
        if (nestedCB != null)
            nestedCB.begin(txid);
    }

    /**
     * Commit the entity manager transaction if one is enlisted in this
     * ObjectGrid transaction.
     * <p>
     * The inner transaction callback, if any, will also be notified with
     * this commit event.
     * @param txid the ObjectGrid transaction context ID
     *
     * @see TransactionCallback#commit(TxID)
     */
    public void commit(TxID txid) throws TransactionCallbackException {
        if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled())
            Tr.debug(tc, "commit " + txid);

        if (nestedCB != null)
            nestedCB.commit(txid);

        if (txid.getSlot(emSlot) != null) {
            EntityManager em = null;
            SavedClientInfo savedDB2ClientInfo = null;
            try {
                em = getEntityManager(txid);
                // DB2 client information
                if (ObjectGridManagerImpl.isTraceEnabled && tc_clientInfo.isEventEnabled() && db2ConnHelper.isDB2()) {
                    if (db2ConnHelper.isPooledConnection() || RuntimeInfo.instance().isWASInstall()) {
                        savedDB2ClientInfo = db2ConnHelper.getSavedClientInfo(txid, savedDB2ClientInfoSlot);
                    }
                }
//                if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled())
//                    Tr.debug(tc, "commit " + em);
                em.getTransaction().commit();
            } catch (javax.persistence.OptimisticLockException ole) {
                OptimisticCollisionException oce = jpaProviderHelper.createOptimisticCollisionException(
                        "An OptimisticLockException is thrown at JPATxCallback.commit time for ObjectGrid " + ogName
                                + ": " + ole.getMessage(), this.ogName, null, null, ole, ogEMF);
                if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled())
                    Tr.debug(tc, "commit ", oce);
                throw new TransactionCallbackException(oce);
            } catch (Exception e) {
                // If it is an SQLException, we need to ffdc its next exception, if existent. 
                jpaProviderHelper.ffdcSQLNextException(e);
                
                if (jpaProviderHelper.isOptimisticLockException(e)) {
                    OptimisticCollisionException oce = jpaProviderHelper.createOptimisticCollisionException(
                            "An OptimisticLockException is thrown at JPATxCallback.commit time for ObjectGrid "
                                    + ogName + ": " + e.getMessage(), this.ogName, null, null, e, ogEMF);

                    if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled())
                        Tr.debug(tc, "commit ", oce);
                    throw new TransactionCallbackException(oce);
                } else if (exceptionMapper != null) {
                    e = (Exception) exceptionMapper.map(e);
                    // Here we might throw a LoaderNotAvaiableException
                    // The caller who is interested in LoaderNotAvaiableException should check 
                    // the chained exception.
                    if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled())
                        Tr.debug(tc, "commit ", e);
                    throw new TransactionCallbackException(e);
                } else {
                    if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled())
                        Tr.debug(tc, "commit ", e);
                    throw new TransactionCallbackException(e);
                }
            } finally {
                // If commit is successful, we should make sure not throwing any exception.
                
                try {
                    // reset DB2 Client Info
                    if ((savedDB2ClientInfo != null) && (!savedDB2ClientInfo.getConnectionInfo().isClosed())) {
                        db2ConnHelper.setClientInfoAtTransactionCommit(txid, savedDB2ClientInfo);
                    } 
//                    if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled())
//                        Tr.debug(tc, "clear " + em);
                    em.clear();
                    
                } catch (Throwable t) {
                    FFDCFilter.processException(t, JPATxCallback.class.getName() + ".commit", "220", this);
                }

                // clear(txid) will never throw an exception.
                clear(txid);

            }
        }
    }

    /**
     * reserve a slot for the JPA entity manager during the ObjectGrid initialization.
     * <p>
     * The inner transaction callback, if any, will also be notified with
     * this initialize event.
     *
     * @param objectGrid A reference to the ObjectGrid.
     *
     * @throws TransactionCallbackException if an error occurs during processing
     *
     * @see TransactionCallback#initialize(ObjectGrid)
     */
    public void initialize(ObjectGrid objectGrid) throws TransactionCallbackException {
        if (punitName == null) {
            throw new IllegalArgumentException(
                    "The persistenceUnitName property is not configured for the transaction callback com.ibm.websphere.objectgrid.jpa.JPATxCallback of ObjectGrid "
                            + objectGrid.getName());
        }

        emSlot = objectGrid.reserveSlot(TxID.SLOT_NAME);
        objectGrid.addEventListener(this);

        ogEMF = ((ObjectGridImpl) objectGrid).getEMFactoryImpl();
        isOGEM = ogEMF.hasRegisteredEntities();

        if (nestedCB != null)
            nestedCB.initialize(objectGrid);

        this.ogName = objectGrid.getName();
    }

    /**
     * Check whether there is an active external transaction.
     * <p>
     * This method will call the same method on the inner transaction callback, if any,
     * to check whether there is an active external transaction. If no inner transaction
     * callback exists, a <code>false</code> is returned.
     * @param session the ObjectGrid Session instance
     * @return true if an external transaction is active
     * @see TransactionCallback#isExternalTransactionActive(Session)
     */
    public boolean isExternalTransactionActive(Session session) {
        if (nestedCB != null)
            return nestedCB.isExternalTransactionActive(session);
        return false;

    }

    /**
     * Rollback the JPA entity manager transaction if one is enlisted in this
     * ObjectGrid transaction.
     * <p>
     * The inner transaction callback, if any, will also be notified with
     * this rollback event.
     *
     * @param txid the ObjectGrid transaction context ID
     * @see TransactionCallback#rollback(TxID)
     */
    public void rollback(TxID txid) throws TransactionCallbackException {
        if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled())
            Tr.debug(tc, "rollback " + txid);

        if (nestedCB != null) {
            try {
                nestedCB.rollback(txid);
            } catch (Throwable t) {
                FFDCFilter.processException(t, JPATxCallback.class.getName() + ".rollback", "280", this);
                if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled())
                    Tr.debug(tc, "rollback", t);
            }
        }

        // The clear will rollback the JPA EM transaction if it's still active,
        // and close the JPA EM.
        clear(txid);
    }

    /**
     * Retrieve the JPA Persistence Unit name
     * @return the JPA persistence unit name
     */
    public String getPersistenceUnitName() {
        return punitName;
    }

    /**
     * Set the JPA persistence unit name
     * @param punitName the persistence unit name
     */
    public void setPersistenceUnitName(String punitName) {
        this.punitName = punitName;
        jpaProviderHelper = new JPAProviderHelperImpl(punitName);
    }

    /**
     * Get the JPA entity manager associated with this ObjectGrid transaction.
     * <p>
     * If no JPA entity manager is associated with this ObjectGrid transaction context ID. An
     * instance will be created and associated ith the ObjectGrid transaction context ID.
     * @param txid the ObjectGrid transaction context id
     * @return the JPA entity manager
     */
    public EntityManager getEntityManager(TxID txid) {
        EntityManager em = (EntityManager) txid.getSlot(emSlot);
        if (em == null) {
            if (emf == null) {
                if (propMap == null && factory != null) {
                    ObjectGrid og = txid.getSession().getObjectGrid();
                    int pid = ((ObjectGridImpl) og).getPartitionId();
                    propMap = factory.getPersistenceProperties(og, pid);
                }
                // This synchronized block should only be hit the first time the emf is created. 
                // Although the getEntityManagerFactory method is synchronized, we still 
                // need this block to make sure one JPATxCallback instance only 
                // call getEntityManagerFactory once.
                synchronized (this) {
                    if (emf == null) {
                        emf = JPAEMFFactory.getEntityManagerFactory(punitName, propMap);
                    }
                }

            }
            try {
                em = emf.createEntityManager();
                // DB2 client information
                if (ObjectGridManagerImpl.isTraceEnabled && tc_clientInfo.isEventEnabled()) {
                    if (firstCall) {
                        db2ConnHelper = new DB2ConnectionHelper(txid, em, punitName, ogName, emf);
                        firstCall = false;
                    }                   
                    if (db2ConnHelper.isDB2()) {
                        savedDB2ClientInfoSlot = db2ConnHelper.setClientInfoAtTransactionBegin(txid, db2ConnHelper.getConnection(em));
                    }
                }
                em.getTransaction().begin();
                txid.putSlot(emSlot, em);
//                if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled())
//                    Tr.debug(tc, "Created " + em + ". Added to slot " + emSlot + ". Started transaction. ");
            } catch (RuntimeException e) {
                FFDCFilter.processException(e, JPATxCallback.class.getName() + ".getEntityManager", "350", this);
                if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled())
                    Tr.debug(tc, "getEntityManager", e);
                throw e;
            }
        }

        return em;
    }

    private void clear(TxID tx) {
        if (tx.getSlot(emSlot) != null) {
            EntityManager em = null;
            try {
                em = getEntityManager(tx);
                if (em.getTransaction().isActive()) {
                    if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled())
                        Tr.debug(tc, "Transaction still active in clear, rollback " + em);
                    em.getTransaction().rollback();
                }
            } catch (Throwable t) {
                FFDCFilter.processException(t, JPATxCallback.class.getName() + ".clear", "300", this);
                if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled())
                    Tr.debug(tc, "clear", t);
            } finally {
                if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled())
                    Tr.debug(tc, "clear tx slot " + emSlot);
                tx.putSlot(emSlot, null);
                
                if (em != null) {
                    try {
//                        if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled())
//                            Tr.debug(tc, "close " + em);
                        em.close();
                    } catch (Throwable t) {
                        FFDCFilter.processException(t, JPATxCallback.class.getName() + ".clear", "310", this);
//                        if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled())
//                            Tr.debug(tc, "em close failed:", t);
                    }
                }
            }
        }
    }

    /**
     * Set the inner transaction callback instance.
     * @param inner the inner transaction callback instance.
     */
    public void setInnerTransactionCallback(TransactionCallback inner) {
        this.nestedCB = inner;
    }

    public void destroy() {
        throw new UnsupportedOperationException("ObjectGridEventListener.destroy()");
    }

    public void initialize(Session session) {
        throw new UnsupportedOperationException("ObjectGridEventListener.initialize(Session)");
    }

    public void transactionBegin(String txid, boolean isWriteThroughEnabled) {
        throw new UnsupportedOperationException("ObjectGridEventListener.transactionBegin(String, boolean)");
    }

    public void transactionEnd(String txid, boolean isWriteThroughEnabled, boolean committed, Collection changes) {
        throw new UnsupportedOperationException(
                "ObjectGridEventListener.transactionEnd(String, boolean, boolean, Collection)");
    }

    public void shardActivated(ObjectGrid grid) {
        // no-op
    }

    public void shardDeactivate(ObjectGrid grid) {
        if (emf != null) {
            if (propMap == null && factory != null) {
                int pid = ((ObjectGridImpl) grid).getPartitionId();
                propMap = factory.getPersistenceProperties(grid, pid);
            }

            synchronized (this) {
                // If shardDeactivate cannot be called concurrently, then this synchronized 
                // is not necessary. However, considering this call is rare, sacrificing the 
                // performance for safety here is probably ok.
                if (emf != null) {
                    JPAEMFFactory.returnEntityManagerFactory(punitName, propMap, emf);
                    emf = null;
                }
            }
        }
    }

    /**
     * Set the <code>JPAPropertyFactory</code> instance. The <code>JPAPropertyFactory</code> instance can be 
     * used to plug in specific persistence property map to the {@link JPATxCallback} object.
     * @param jpaPropertyFactory
     */
    public void setJPAPropertyFactory(JPAPropertyFactory jpaPropertyFactory) {
        factory = jpaPropertyFactory;
    }

    /**
     * @return the <code>JPAPropertyFactory</code> instance.
     */
    JPAPropertyFactory getJPAPropertyFactory() {
        return factory;
    }

    /**
     * Set the {@link ExceptionMapper}. The <code>ExceptionMapper</code> instance can be used to plug in 
     * JPA-specific or database-specific exception mapping functionalities.
     * 
     * @param exceptionMapper the ExceptionMapper instance
     * @see ExceptionMapper
     * @since 7.0
     */
    public void setExceptionMapper(ExceptionMapper exceptionMapper) {
        if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled())
            Tr.debug(tc, this + " - set exceptionMapper to " + exceptionMapper);
        this.exceptionMapper = new ExceptionMapperWrapper(exceptionMapper);
    }

    /**
     * @return the <code>ExceptionMapper</code> instance.
     */
    ExceptionMapper getExceptionMapper() {
        return exceptionMapper;
    }
}
