// IBM Confidential OCO Source Material
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2008
// The source code for this program is not published or otherwise divested
// of its trade secrets, irrespective of what has been deposited with the
// U.S. Copyright Office.
//
// $Id$
//
// Source File Description:
//
//  Need a file description.
//
// Change Activity:
//
// Reason       Version   Date         Userid    Change Description
// ------------ --------- ------------ --------- -----------------------------------------
// d11146       WASX.XD   04/08/2008   jian.tang Time-based database update
// ------------ --------- ------------ --------- -----------------------------------------
package com.ibm.websphere.objectgrid.jpa.dbupdate;

import java.util.HashMap;
import java.util.Map;

import com.ibm.ejs.ras.Tr;
import com.ibm.ejs.ras.TraceComponent;
import com.ibm.websphere.objectgrid.ObjectGrid;
import com.ibm.websphere.objectgrid.ObjectGridException;
import com.ibm.websphere.objectgrid.TimeBasedDBUpdateConfig.DBUpdateMode;
import com.ibm.ws.objectgrid.Constants;
import com.ibm.ws.objectgrid.dbupdate.DBUpdaterThread;
import com.ibm.ws.objectgrid.jpa.JPAXMLHelper;
import com.ibm.ws.xs.NLSConstants;

/**
 * A time-based database updater can be used to periodically get the latest inserts and updates to the
 * ObjectGrid backend database. A time-based database updater interacts with Java Persistence
 * API (JPA) to get the latest changes (inserts and updates) from the database using a specific
 * timestamp field. Therefore, a JPA provider is required to use this feature.
 * The database here can be any backend supported by the chosen JPA provider.
 * <P>
 * The time-based database updater periodically query the database using JPA interfaces to
 * get the JPA entities which represent the newly inserted and updated records in the database.
 * In order to do that, every record in the database should have a "timestamp" to identify the time
 * or sequence when the record was last updated or inserted. The "timestamp" here does not have to be
 * in a timestamp format. It can be in a an integer or long format, as long as it generates a unique
 * increasing value.
 * <p>
 * Several commercial databases have provided this capability.
 * <p>
 * For example, in DB2 9.5, you
 * can define a column using the ROW CHANGE TIMESTAMP format as follows:
 * <pre>
 * <b>ROWCHGTS TIMESTAMP NOT NULL
 *      GENERATED ALWAYS
 *      FOR EACH ROW ON UPDATE AS
 *      ROW CHANGE TIMESTAMP</b>
 * </pre>
 *
 * In Oracle, you can use the pseudo-column <b>ora_rowscn</b>, which represents the system change
 * number of the record.
 * <p>
 *
 * Using this timestamp, the time-based database updater utilize the following two queries to get the latest
 * database changes:
 * <ol>
 * <li>The first query selects all the entities whose timestamp is greater than a timestamp:
 * The query syntax is  <b>SELECT o FROM ENTITY o where o.TIMESTAMP > ?1</b>.
 * <li>The second query gets the maximum timestamp from all exsting JPA entities. The query syntax is
 * <b>SELECT MAX(o.TIMSTAMP) FROM ENTITY o</b>. <li>
 * </ol>
 * In both queries, the "ENTITY" is the JPA entity name, and the "TIMESTAMP" is the JPA entity attribute
 * which represents the timestamp column in database.
 * <p>
 * However, record removes to the backend database cannot be detected.
 * <p>
 * The time-based database updater can be launched in the following two ways:
 * <OL>
 * <LI>For a distributed or local ObjectGrid, if a "timeBasedDBUpdate" element is configured for a backing
 * map, a time-based database updater is automatically launched. For the distributed objectgrid, it will
 * only be launched in partition 0.
 * <LI>For a distributed or local ObjectGrid, a user application can launch a time-based database updater
 * using this TimeBasedDBUpdater interface by calling <tt>startDBUpdate</tt>.
 * </OL>
 *
 * @since XD 6.1.0.3
 * @see com.ibm.websphere.objectgrid.jpa.dbupdate.annotation.Timestamp
 * @see com.ibm.websphere.objectgrid.BackingMap
 */
public class TimeBasedDBUpdater {
    private static final String CLASS_NAME = TimeBasedDBUpdater.class.getName();

    private static final TraceComponent tc = Tr.register(CLASS_NAME, Constants.TR_JPA_GROUP_NAME,
            Constants.TR_RESOURCE_BUNDLE_NAME);

    private static TimeBasedDBUpdater instance = new TimeBasedDBUpdater();

    Map<String, DBUpdaterThread> updaters = new HashMap();

    private TimeBasedDBUpdater() {
    }

    /**
     * @return an instance of TimeBasedDBUpdater.
     */
    public static TimeBasedDBUpdater instance() {
        return instance;
    }

    /**
     * Start a time-base database update thread to keep the ObjectGrid data updated from the
     * database back end
     * @param objectGrid the ObjectGrid instance
     * @param mapName the map name to be updated
     * @param punitName the JPA persistence unit name
     * @param entityClass the JPA entity class which represents the database table
     * @param timestampField the timestamp field used by ObjectGrid to query the changes
     * @param mode the database update mode.
     * @throws ObjectGridException
     */
    public synchronized void startDBUpdate(ObjectGrid objectGrid, String mapName, String punitName, Class entityClass,
            String timestampField, DBUpdateMode mode) {
        String fullName = objectGrid.getName() + "." + mapName;
        if (updaters.get(fullName) != null) {
            throw new IllegalStateException("A database update thread is running for this ObjectGrid map: " + fullName);
        }

        if (punitName == null) {
            punitName = JPAXMLHelper.getFirstPU();

            if (punitName == null) {
                throw new IllegalArgumentException("There is no persistence unit defined in the META-INF/persistence.xml");
            }
            Tr.info(tc, NLSConstants.DEFAULT_PERSISTENCE_UNIT_CWOBJ3112I, punitName);
        }

        DBUpdaterThread dbUpdaterThread = new DBUpdaterThread(objectGrid, mapName, punitName, entityClass, timestampField,
                mode);
        updaters.put(fullName, dbUpdaterThread);

        dbUpdaterThread.startUpdateThread();
    }

    /**
     * Stop the time-based database update thread
     * @param objectGrid the ObjectGrid instance
     * @param mapName the map name to be updated
     *
     * @throws IllegalArgumentException If there is no time-based database update thread started for this map.
     */
    public synchronized void stopDBUpdate(ObjectGrid objectGrid, String mapName) {
        String fullName = objectGrid.getName() + "." + mapName;
        if (updaters.get(fullName) == null) {
            throw new IllegalArgumentException("A database update thread is not running for this ObjectGrid map: "
                    + fullName);
        }
        DBUpdaterThread updater = updaters.get(fullName);

        updater.stopUpdateThread();
        updaters.remove(fullName);
    }

}
