package com.ibm.websphere.objectgrid.management;

import javax.management.MBeanException;
import javax.management.openmbean.OpenDataException;
import javax.management.openmbean.TabularData;
import javax.transaction.xa.XAException;
import javax.transaction.xa.Xid;

/**
 * This MBean interface allows a client process to retrieve and heuristically resolve all indoubt transactions 
 * from a partition running in a dynamic environment. <p>
 * 
 * An indoubt transaction is a WebSphere eXtreme Scale multi-partition transaction where one or more of the partitions participating
 * in the transaction have completed the first phase of the two-phase commit protocol but are waiting to begin the second phase of the protocol.
 * During this indoubt period locks are held on all data written to the BackingMaps until the decision to commit or rollback the transaction is determined. <p>  
 *  
 * The object name pattern for this MBean is:
 * <pre>
 * com.ibm.websphere.objectgrid:type=Transaction,objectGridName=&lt;objectgrid&gt;,mapSetName=&lt;mapset&gt;,partition=&lt;partition id&gt;,host=&lt;host&gt;,ogServerName=&lt;server&gt;
 * </pre>
 *  
 * @since 8.6
 */
public interface TransactionMBean {
	
	/**
	 * Indicates that this partition has been assigned as the Transaction Manager (TM) for this transaction.
	 * The responsibility of the TM partition is to record the decision of the transaction and to ensure transactional 
	 * consistency across all partitions involved in the transaction when a failure occurs during the second phase
	 * of the two-phase commit protocol 
	 */
	public final static String TM = "TM";
	
	/**
	 * Indicates that this partition is a participant or Resource manager (RM) in this transaction.
	 */
	public final static String RM = "RM";
	
	/**
	 * Indicates that the transaction at this partition has successful completed the first phase of the two-phase commit protocol and is waiting to begin the second phase of the protocol. 
	 */
	public final static String PREPARED = "PREPARED";
	
	/**
	 * Indicates that the outcome of this transaction is to commit. 
	 * The TM partition is waiting for all the partitions involved in the transaction to commit.
	 */
	public final static String COMMIT = "COMMIT";
	
	/**
	 * Indicates that the outcome of this transaction is rollback. 
	 * The TM partition is waiting for all the partitions involved in the transaction to rollback.
	 */		
	public final static String ROLLBACK = "ROLLBACK";
	
	/**
	 * Indicates that this transaction was heuristically committed at this partition.
	 */	
	public final static String HEUR_COMMIT = "HEUR_COMMIT";
		
	/**
	 * Indicates that this transaction was heuristically rolled back at this partition.
	 */	
	public final static String HEUR_ROLLBACK = "HEUR_ROLLBACK";
	
	/**
	 * Timestamp column name defined in the CompositeData returned by the getIndoubtList method 
	 */
	public final static String COLUMN_TIMESTAMP = "Timestamp";

	/**
	 * Xid column name defined in the CompositeData returned by the getIndoubtList method 
	 */
	public final static String COLUMN_XID = "Xid";
	
	/**
	 * State column name defined in the CompositeData returned by the getIndoubtList method 
	 */
	public final static String COLUMN_STATE = "State";
	
	/**
	 * Role column name defined in the CompositeData returned by the getIndoubtList method 
	 */
	public final static String COLUMN_ROLE = "Role";
	
	/**
	 * ResyncAttempts column name defined in the CompositeData returned by the getIndoubtList method 
	 */
	public final static String COLUMN_RESYNCATTEMPTS = "ResyncAttempts";
		
	/**
     * Retrieve the type of the shard this TransactionMBean is instantiate on.
     * @return The shard type.
     */
    public String getType();
    
	/**
     * Retrieve a list of all transactions that are in the indoubt state at this partition.  
     * If the filter specification argument is specified, then only indoubt transactions that meet the
     * filter criteria are returned. <p>
     * The filter specification should be in the form: 
	 * <code>filterName=value(:&lt;filterName=value&gt;)*</code> 
	 * <table>
	 * <tr>
	 * 		<td><strong><u>filterName</u></strong><td><strong><u>Description</u></strong></td>
	 * </tr> 
	 * </tr>
	 * <tr valign="top">
	 * 		<td>age</td>
	 * 		<td>Retrieve indoubt transactions that have been in the system for more than the specified value. Value specified in minutes</td> 		
	 * </tr>
	 * <tr valign="top">
	 * 		<td>xid</td>
	 * 		<td>Retrieve only the specified indoubt transaction. Value represents the transaction identifier</td> 		
	 * </tr>
	 * </table>
     * <p>
     * This method returns a TabularData object representing the details of each indoubt transaction at this partition.<p>
     *
     * Each CompositeData (row in the TabularData) contains the following items:<p>
     * <table>
     * <tr>
     *     <td><strong><u>Item Name</u></strong><td><strong><u>Type</u></strong><td><strong><u>Description</u></strong>
     * </tr>
     * <tr>
     *     <td>Timestamp</td>
     *     <td>String</td>
     *     <td>Specifies the time the transaction entered the indoubt state at this partition.</td>
     * </tr>
     * <tr>
     *     <td>State</td>
     *     <td>String</td>
     *     <td>The state of the transaction at this partition.</td>
     * </tr>
     * <tr>
     *     <td>Role</td>
     *     <td>String</td>
     *     <td>The role the transaction is using this partition for. Valid roles include: <code>TM</code> or <code>RM</code>.</td>
     * </tr>
     * <tr valign="top">
     *     <td>ResyncAttempts</td>
     *     <td>Integer</td>
     *     <td>If the partition is acting as the <code>RM</code> then this value indicates the  
     *         the number of times this partition has attempted to re-synchronize with the TM partition for this transaction. Otherwise it indicates the number of times this partition 
     *         has attempted to re-synchronize with the RM partitions involved in this transaction.</td>
     * </tr>
     * <tr>
     * 		<td>Xid</td>
     * 		<td>String</td>
     * 		<td>The X/Open Transaction Identifier</td>
     * </tr>
     * </table>
     * @param filterSpec - The filter specification
     * @return TabularData representing all partitions involved in the transaction
     * @since 8.6
     */
    TabularData getIndoubtList(String filterSpec) throws OpenDataException;
    
    /**
     * Heuristically commit the specified transaction at this partition. 
     * The operation is only valid if the state of the transaction 
     * is <code>PREPARED</code><p>
     * Please use extreme caution before invoking this function. The best strategy is to wait for eXtreme scale to automatically 
     * resolve this transaction. You could experience data integrity problems if you manually commit or roll back a transaction in one of the
     * partition, and the opposite action is taken against another partition. 
     * @param xid The transaction identifier
     * @throws MBeanException
     */
    public void commit(String xid) throws MBeanException;
    
    /**
     * Heuristically rollback the specified transaction at this partition. 
     * The operation is only valid if the state of the transaction 
     * is <code>PREPARED</code><p>
     * Please use extreme caution before invoking this function. The best strategy is to wait for eXtreme scale to automatically 
     * resolve this transaction. You could experience data integrity problems if you manually commit or roll back a transaction in one of the
     * partition, and the opposite action is taken against another partition.
     * @param xid The transaction identifier 
     * @throws MBeanException
     */
    public void rollback(String xid) throws MBeanException;
    
    /**
     * Heuristically forget the specified transaction at this partition. 
     * The operation is only valid if the state of the transaction  
     * is <code>HEUR_COMMIT</code> or <code>HEUR_ROLLBACK</code><p>
     * 
     * Please use extreme caution before invoking this function. The best strategy is to wait for eXtreme scale to automatically 
     * resolve this transaction. You could experience data integrity problems if you manually commit or roll back a transaction in one of the
     * partition, and the opposite action is taken against another partition.
     * @param xid The transaction identifier    
     * @throws MBeanException
     */    
    public void forget(String xid) throws MBeanException;
	
}
