package com.ibm.websphere.objectgrid.plugins;

import com.ibm.websphere.objectgrid.BackingMap;

/**
 * Listener for BackingMap life cycle events.
 *
 * The lifecycle listener can indicate a fatal error
 * condition using the <code>LifecycleFailedException</code> for
 * a lifecycle event associated with the <code>INITIALIZING</code>,
 * <code>INITIALIZED</code>, <code>OFFLINE</code>, <code>STARTING</code>,
 * and <code>PRELOAD</code> states. The <code>LifecycleFailedException</code>
 * will cause an abort and destruction of the shard.
 * <p>
 * A BackingMapLifecycleListener implementation may also implement the
 * {@link BackingMapPlugin} interface in order
 * to receive enhanced BackingMap plug-in lifecycle method calls. The plug-in is
 * also required to correctly implement each of the bean methods related to
 * introspection of its state (for example <code>isInitialized()</code>,
 * <code>isDestroyed()</code>, etc).
 *
 * @since 7.1.1
 * @see BackingMap#addMapEventListener(EventListener)
 * @see EventListener
 */
public interface BackingMapLifecycleListener extends EventListener {

    /**
     * The state of a BackingMap's life cycle.
     */
    public enum State {
        /**
         * The BackingMap instance has been created.
         */
        NEW,

        /**
         * The BackingMap instance initialization phase is starting. The BackingMap
         * and BackingMap plug-ins are about to be initialized.
         *
         * A LifecycleFailedException sent during this state transition
         * will result in an abort of the state change and destruction
         * of the BackingMap.
         */
        INITIALIZING,

        /**
         * The BackingMap initialization phase is complete. All BackingMap plug-ins
         * are initialized. The INITIALIZED state may re-occur when shard
         * placement activities (promotion, demotion) occur.
         *
         * A LifecycleFailedException sent during this state transition
         * will result in an abort of the state change and destruction
         * of the BackingMap.
         */
        INITIALIZED,

        /**
         * The BackingMap instance is being activated for use in a local
         * instance, client instance or as an instance in a primary or
         * replica shard on the server.
         * All BackingMap plug-ins have been created an initialized.
         * All ObjectGrid plug-ins in the ObjectGrid instance owning this
         * BackingMap instance have been initialized.
         *
         * A LifecycleFailedException sent during this state transition
         * will result in an abort of the state change and destruction
         * of the BackingMap.
         */
        STARTING,

        /**
         * The BackingMap instance is set to the PRELOAD state by the
         * StateManager API for preload, or is performing preload using
         * the configured Loader.
         * All BackingMap plug-ins have been created an initialized.
         */
        PRELOAD,

        /**
         * The BackingMap instance is ready for work as a
         * local instance, client instance or an instance in a primary or replica
         * shard on the server. This is the typical steady state of the BackingMap.
         * All BackingMap plug-ins have been created an initialized.
         * All ObjectGrid plug-ins in the ObjectGrid instance owning this
         * BackingMap instance have been initialized.
         */
        ONLINE,

        /**
         * Work is stopping on the BackingMap as a result of the StateManager
         * API or other event. No new work will be allowed. Your plug-in should end any
         * existing work as soon as possible.
         */
        QUIESCE,

        /**
         * All work is stopped on the BackingMap as a result of the
         * StateManager API or other event. No new work will be allowed.
         */
        OFFLINE,

        /**
         * The BackingMap instance is starting the destroy phase.
         * BackingMap plug-ins for the instance are about to be destroyed.
         */
        DESTROYING,

        /**
         * The BackingMap instance and all BackingMap plug-ins have been destroyed.
         */
        DESTROYED
    }

    /**
     * The <code>LifecycleEvent</code> describes the situation that
     * caused a change in the lifecycle of the <code>BackingMap</code>.
     */
    static interface LifecycleEvent {
        /**
         * Retrieve the BackingMap instance that the event applies to.
         * @return the BackingMap instance that the event applies to.
         */
        public BackingMap getBackingMap();

        /**
         * Get the lifecycle state occurring for this event.
         *
         * @return State - the new life cycle state
         */
        public State getState();

        /**
         * Answers true when the BackingMap instance is currently readable.  Data can be
         * read from the map.  A map is readable for all local and client grid instances
         * that have been initialized.
         * Server instances are locally readable when they are currently on-line. Replica
         * instances on the server are not necessarily readable depending on the current
         * configuration of the grid.
         */
        public boolean isReadable();

        /**
         * Answers true when the BackingMap instance is currently writable.  Data can be
         * modified in the map.  A map is writable for all local and client grid instances when
         * the {@link BackingMap#getReadOnly()} answers false.
         * Map instances on the primary shard, server instance are locally writable when
         * they are currently on-line and the {@link BackingMap#getReadOnly()} answers false.
         */
        public boolean isWritable();
    };

    /**
     * Invoked when there is a change to the BackingMap's life cycle.
     * When LifecycleFailedExceptions are thrown for the following lifecycle
     * states, the BackingMap instance will abort the state change and
     * be destroyed: INITIALIZING, INITIALIZED, STARTING, and PRELOAD.
     * Exceptions thrown during other phases will be logged as error conditions.
     * @param event contextual data for the event.
     *
     * @throws LifecycleFailedException thrown when there is a fatal exception. For
     * some states, the BackingMap instance will abort the state change and be destroyed.
     */
    public void backingMapStateChanged(LifecycleEvent event) throws LifecycleFailedException;
}
