// IBM Confidential OCO Source Material
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2006-2010
// The source code for this program is not published or otherwise divested
// of its trade secrets, irrespective of what has been deposited with the
// U.S. Copyright Office.
//
// $Id$
//
// Module  :  JMSObjectGridEventListener.java
//
// Source File Description:
//
// Change Activity:
//
// Reason       Version   Date         Userid    Change Description
// ------------ --------- ------------ --------- -----------------------------------------
// b994         6.1       07/28/2006   cheng1    new part
// b14685       6.1       01/09/2009   cheng1    use ObjectGrid.getObjectGridType() to determine the type
// b16883       7.0.x     07/22/2009   cheng1    filter out LogSequence before publishing JMS message
// b17811       7.0.x     10/30/2009   cheng1    publish messages even in beginNoWriteThrough mode.
// b20980       6105 FIX5 08/10/2010   jian.tang Push the security context on thread if running in WAS
// ------------ --------- ------------ --------- -----------------------------------------

package com.ibm.websphere.objectgrid.plugins.builtins;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;
import java.util.Set;
import java.util.StringTokenizer;

import javax.jms.ExceptionListener;
import javax.jms.JMSException;
import javax.jms.ObjectMessage;
import javax.jms.Topic;
import javax.jms.TopicConnection;
import javax.jms.TopicConnectionFactory;
import javax.jms.TopicPublisher;
import javax.jms.TopicSession;
import javax.jms.TopicSubscriber;
import javax.naming.InitialContext;

import com.ibm.ejs.ras.Tr;
import com.ibm.ejs.ras.TraceComponent;
import com.ibm.websphere.objectgrid.BackingMap;
import com.ibm.websphere.objectgrid.LockStrategy;
import com.ibm.websphere.objectgrid.ObjectGrid;
import com.ibm.websphere.objectgrid.ObjectGridException;
import com.ibm.websphere.objectgrid.ObjectGridRuntimeException;
import com.ibm.websphere.objectgrid.OutputFormat;
import com.ibm.websphere.objectgrid.Session;
import com.ibm.websphere.objectgrid.plugins.DistributionMode;
import com.ibm.websphere.objectgrid.plugins.LogElement;
import com.ibm.websphere.objectgrid.plugins.LogSequence;
import com.ibm.websphere.objectgrid.plugins.LogSequenceFilter;
import com.ibm.websphere.objectgrid.plugins.LogSequenceTransformer;
import com.ibm.websphere.objectgrid.plugins.ObjectGridEventListener;
import com.ibm.websphere.objectgrid.plugins.io.annotations.PluginOutputFormat;
import com.ibm.ws.objectgrid.Constants;
import com.ibm.ws.objectgrid.ObjectGridManagerImpl;
import com.ibm.ws.objectgrid.util.ObjectInputStreamResolver;
import com.ibm.ws.objectgrid.util.security.SecurityContextRunnableFactory;

/**
 * This class is a JMS-based implementation of <code>ObjectGridEventListener</code>. It is designed to support client
 * side near cache invalidation and peer-to-peer replication mechanism.
 * <P>
 * The client invalidation mechanism can be used in distributed ObjectGrid environment to ensure client near cache data
 * to be synchronized with servers or other clients. Without this function, client near cache could hold stale data.
 * However, even with this JMS-based client invalidation mechanism, you have to realize the timing window for
 * updating client near cache.  There will be a delay for the ObjectGrid runtime to publish updates.
 * <P>
 * The peer-to-peer replication mechanism can be used in both distributed and local ObjectGrid environment. It is an
 * ObjectGrid core to core replication process and allows data updates flow among local ObjectGrids and distributed
 * ObjectGrids.  For example, with this mechanism you can move data updates from an ObjectGrid cluster to a local
 * ObjectGrid, or from an ObjectGrid cluster to another ObjectGrid cluster in different system domain.
 * <P>
 * This class requires user to configure JMS and JNDI information in order to obtain required JMS resources.
 * Additionally, replication related properties must be set correctly.  In J2EE environment, the JNDI should be
 * available in both web and EJB containers.  In this case, the JNDI property is optional unless you want to obtained
 * external JMS resources.
 * <P>
 * This JMS-based implementation of ObjectGridEventListener has some properties you can configure via xml or
 * programmatic approaches. They could be only for client invalidation mechanism, only for peer-to-peer replication,
 * or common for both. Most properties are optional for customizing the behavior to achieve desired function. Within
 * the value of property, the semicolon ( ; ) can be used as delimiter to separate elements.
 * <P>
 * Client invalidation mechanism related properties:
 * <ul>
 * <li>invalidationModel: This property is used in distributed ObjectGrid environment to enable invalidation mechanism
 * in client near cache. Possible values include:
 *     <ul>
 *     <li>NONE_INVALIDATION_MODEL: default value. This means no client invalidation mechanism.</li>
 *     <li>CLIENT_SERVER_MODEL: In this model, all server processes will be in publisher role that will publish all
 *         transaction changes to designated JMS destination.  All client processes will be in receiver roles and will
 *         receive all transactional changes from designated JMS destination.</li>
 *     <li>CLIENT_AS_DUAL_ROLES_MODEL: Client as dual roles model:  In this model, all server processes have nothing to
 *         do with JMS destination. All client processes will be both JMS publisher and receiver roles.  Transactional
 *         changes occur on client will be published to JMS destination and all clients will receive these
 *         transactional changes.</li>
 *     </ul>
 * <li>invalidationStrategy: This property controls the type of updates to publish or receive. Possible values include:
 * </li>
 *     <ul>
 *     <li>INVALIDATE: default value. See {@link DistributionMode#INVALIDATE}</li>
 *     <li>INVALIDATE_CONDITIONAL: See {@link DistributionMode#INVALIDATE_CONDITIONAL}</li>
 *     <li>PUSH: (see {@link DistributionMode#PUSH})</li>
 *     <li>PUSH_CONDITIONAL: See {@link DistributionMode#PUSH_CONDITIONAL}</li>
 *     <li>PUSH_INCLUDED: This is an extension to PUSH strategy.  With PUSH strategy, all updates including INSERT type
 *     will be published and received.  Because it also includes INSERT type update, the receivers will insert data into
 *     its maps locally that might not be desired.  Besides, the receivers might be overwhelmed by inserting too many
 *     unnecessary data. The PUSH_INCLUDED strategy is like PUSH strategy, but it does not include INSERT type update.
 *     The suffix INCLUDED means the changes only include existing data, and exclude new data that is the result of
 *     INSERT operation.</li>
 *     <li>PUSH_INCLUDED_CONDITIONAL: This is an extension to PUSH_CONDITIONAL strategy. Similar to the PUSH_INCLUDED, the
 *     PUSH_INCLUDED_CONDITIONAL is like PUSH_CONDITIONAL, but it does not include INSERT type update.</li>
 *     </ul>
 * </ul>
 * <P>
 * Peer-to-peer replication related properties:
 * <ul>
 * <li>replicationRole: This property specifies the role of the ObjectGrid from the perspective of JMS. An ObjectGrid
 * core could be a JMS publisher, subscriber, or both. For example, if a ObjectGrid core just wants to receive updates
 * from other ObjectGrid cores, the value of this property should be SUBSCRIBER_ROLE. Possible values include:</li>
 *     <ul>
 *     <li>NONE_ROLE: default value. The NONE_ROLE replication role indicates the current ObjectGrid has nothing to do
 *     with JMS.  This virtually disables this plugin on the current ObjectGrid instance.</li>
 *     <li>DUAL_ROLES: The DUAL_ROLES replication role indicates the current ObjectGrid will act as both JMS publisher
 *     and subscriber role.</li>
 *     <li>PUBLISHER_ROLE: The PUBLISHER_ROLE replication role indicates the current ObjectGrid will only act as JMS
 *     publisher role.</li>
 *     <li>SUBSCRIBER_ROLE: The SUBSCRIBER_ROLE replication role indicates the current ObjectGrid will only act as JMS
 *     subscriber role.</li>
 *     </ul>
 * <li>replicationStrategy: This property controls the type of updates to publish or receive.  It has same possible values
 * as invalidationStrategy.</li>
 * </ul>
 * <P>
 * Required common properties to both client invalidation and Peer-to-peer replication mechanism:
 * <ul>
 * <li>jms_topicConnectionFactoryJndiName</li>
 * <li>jms_topicJndiName</li>
 * <li>jms_topicName</li>
 * </ul>
 * <P>
 * Optional common properties to both client invalidation and Peer-to-peer replication mechanism:
 * <ul>
 * <li>jms_userid</li>
 * <li>jms_password</li>
 * <li>jndi_properties: This property specifies JNDI properties for naming lookup.  It should have the same content
 * as regular jndi.properties file that is used to connect to JNDI naming server.  An example of the value of this
 * property is "java.naming.factory.initial=org.apache.activemq.jndi.ActiveMQInitialContextFactory;java.naming.provider.url=tcp://localhost:61616;connectionFactoryNames=defaultTCF;topic.defaultTopic=defaultTopic"
 * when using external ActiveMQ JMS provider. </li>
 * <li>mapsToPublish: This property controls  which maps to be included during publishing and receiving updates.
 * If not specified, the default is all maps defined in the ObjectGrid configuration. An example of the value of this
 * property is "agent;profile;pessimisticMap"</li>
 * <li>enableOnClientObjectGrid: This property controls  whether or not to enable this JMS-based ObjectGrid plugin in
 * client ObjectGrid. Possible values include:</li>
 *     <ul>
 *     <li>true: default value</li>
 *     <li>false</li>
 *     </ul>
 * <li>enableOnServerObjectGrid: This property controls  whether or not to enable this JMS-based ObjectGrid plugin in
 * server ObjectGrid. Possible values include:</li>
 *     <ul>
 *     <li>true: default value</li>
 *     <li>false</li>
 *     </ul>
 * </ul>
 *
 * <p>
 * There are two models available for client invalidation mechanism in distributed ObjectGrid environment:
 *     <ul>
 *     <li>Client-server model: In this model, all server processes will be in publisher role that will publish all
 *         transaction changes to designated JMS destination.  All client processes will be in receiver roles and will
 *         receive all transactional changes from designated JMS destination.
 *     </li>
 *     <li>Client as dual roles model:  In this model, all server processes have nothing to do with JMS destination.
 *         All client processes will be both JMS publisher and receiver roles.  Transactional changes occur on client
 *         will be published to JMS destination and all clients will receive these transactional changes
 *     </li>
 *     </ul>
 * <p>
 * The following is an xml configuration example to enable client invalidation mechanism with client-server model
 * on an ObjectGrid configuration.
 * <pre>
 *     &lt;bean id="ObjectGridEventListener" className="com.ibm.ws.objectgrid.plugins.JMSObjectGridEventListener"&gt;
 *         &lt;property name="invalidationModel" type="java.lang.String" value="CLIENT_SERVER_MODEL" description="" /&gt;
 *         &lt;property name="invalidationStrategy" type="java.lang.String" value="PUSH" description="" /&gt;
 *         &lt;property name="jms_topicConnectionFactoryJndiName" type="java.lang.String" value="defaultTCF" description="" /&gt;
 *         &lt;property name="jms_topicJndiName" type="java.lang.String" value="defaultTopic" description="" /&gt;
 *         &lt;property name="jms_topicName" type="java.lang.String" value="defaultTopic" description="" /&gt;
 *     &lt;/bean&gt;
 * </pre>
 *
 * The following is an xml configuration example to enable client invalidation mechanism with client as dual roles model
 * on an ObjectGrid configuration.
 * <pre>
 *     &lt;bean id="ObjectGridEventListener" className="com.ibm.ws.objectgrid.plugins.JMSObjectGridEventListener"&gt;
 *         &lt;property name="invalidationModel" type="java.lang.String" value="CLIENT_AS_DUAL_ROLES_MODEL" description="" /&gt;
 *         &lt;property name="invalidationStrategy" type="java.lang.String" value="PUSH" description="" /&gt;
 *         &lt;property name="jms_topicConnectionFactoryJndiName" type="java.lang.String" value="defaultTCF" description="" /&gt;
 *         &lt;property name="jms_topicJndiName" type="java.lang.String" value="defaultTopic" description="" /&gt;
 *         &lt;property name="jms_topicName" type="java.lang.String" value="defaultTopic" description="" /&gt;
 *     &lt;/bean&gt;
 * </pre>
 * <P>
 * The peer-to-peer replication mechanism allows ObjectGrid cores to send transactional changes among each other.
 * This mechanism can be configured on both local and distributed ObjectGrid. The following is an xml
 * configuration example to enable peer-to-peer replication mechanism on an ObjectGrid configuration.
 * <pre>
 *     &lt;bean id="ObjectGridEventListener" className="com.ibm.ws.objectgrid.plugins.JMSObjectGridEventListener"&gt;
 *         &lt;property name="replicationRole" type="java.lang.String" value="DUAL_ROLES" description="" /&gt;
 *         &lt;property name="replicationStrategy" type="java.lang.String" value="PUSH" description="" /&gt;
 *         &lt;property name="jms_topicConnectionFactoryJndiName" type="java.lang.String" value="defaultTCF" description="" /&gt;
 *         &lt;property name="jms_topicJndiName" type="java.lang.String" value="defaultTopic" description="" /&gt;
 *         &lt;property name="jms_topicName" type="java.lang.String" value="defaultTopic" description="" /&gt;
 *         &lt;property name="jms_userid" type="java.lang.String" value="" description="" /&gt;
 *         &lt;property name="jms_password" type="java.lang.String" value="" description="" /&gt;
 *         &lt;property name="jndi_properties" type="java.lang.String" value="java.naming.factory.initial=org.apache.activemq.jndi.ActiveMQInitialContextFactory;java.naming.provider.url=tcp://localhost:61616;connectionFactoryNames=defaultTCF;topic.defaultTopic=defaultTopic" description="jndi properties" /&gt;
 *     &lt;/bean&gt;
 * </pre>
 *
 * The mapsToPublish property in both client invalidation mechanism and peer-to-peer replication mechanism is optional.
 * It allows specifying maps that should publish and receive transactional chnages.  If this property is not specified,
 * All maps defined in the ObjectGrid will publish and receive transactional changes.
 * <P>
 * The jndi_properties property is only required when JMS resource is not available locally.  For example, a Java
 * program uses a local ObjectGrid that has this JMS-based ObjectGridEventListener configured to enable peer-to-peer
 * replication mechanism.  The JMS server is another Java process, so that the jndi_properties property is required
 * for the Java program to perform JNDI lookup on remote JMS server.  Usually, in WebSphere Application Server
 * environment, the application that uses ObjectGrid should not need to set this jndi_properties property, because
 * usually the JMS resources will be configured to be available for applications installed on the WebSphere servers.
 * <P>
 * In distributed ObjectGrid environment, both client invalidation mechanism and peer-to-peer replication mechanism can
 * specify whether this JMS-based ObjectGridEventListener should run on either server or client process.
 * This can be achieved by the setting of the following two properties:
 * <Pre>
 *     &lt;property name="enableOnClientObjectGrid" type="boolean" value="false" description="" /&gt;
 *     &lt;property name="enableOnServerObjectGrid" type="boolean" value="false" description="" /&gt;
 * </Pre>
 * Both enableOnClientObjectGrid and enableOnServerObjectGrid properties are default to true that means by default
 * the configured JMS-based ObjectGridEventListener will run on both server and client process.  If set to false,
 * the configured JMS-based ObjectGridEventListener will be disable on the specified process.  For example, if a
 * distributed ObjectGrid has these two properties set to false as above, that means this configured JMS-based
 * ObjectGridEventListener is virtually disabled.
 * <p>
 * If you would like to set thess properties programmatically, you can follow Java Bean pattern. For example, the
 * setInvalidationModel(String) method is for setting invalidationModel property. The following is a programming
 * example:
 * <pre>
 *      JMSObjectGridEventListener jmsObjectGridEventListener = new JMSObjectGridEventListener();
 *      jmsObjectGridEventListener.setInvalidationModel(JMSObjectGridEventListener.CLIENT_SERVER_MODEL_STRING);
 * </pre>
 * Note: Use constants provided by this class in property setters to avoid typo, e.g.
 * the value of JMSObjectGridEventListener.CLIENT_SERVER_MODEL_STRING is "CLIENT_SERVER_MODEL".
 * <P> 
 *  
 * The JMSObjectGridEventListener provides an extension point, the onMessage
 * method, for application that wants to be notified when receiving JMS message.
 * <P>
 * A normal usage scenario of this extension point is when configuring client invalidation mechanism with 
 * CLIENT_SERVER_MODEL invalidationModel and application wants to know evictions occur in server ObjectGrid from client
 * ObjectGrid. In this usage scenario, application can provide a class that extends JMSObjectGridEventListener and put 
 * handling logic in overridden onMessage method, and configure this extended JMSObjectGridEventListener the same way 
 * as JMSObjectGridEventListener.
 * <P>
 * The extended JMSObjectGridEventListener is child class of JMSObjectGridEventListener and can only override two 
 * methods: initialize and onMessage method. Overriding initialize method is optional. If the child class of 
 * JMSObjectGridEventListener needs to use any ObjectGrid artifacts, such as ObjectGrid, Session, and so on, in 
 * onMessage method, it can get these ObjectGrid artifacts in initialize method and cache them as instance variables. 
 * Within onMessage method, cached ObjectGrid artifacts can be used to process passed collection of LogSequences. 
 * <p>
 * Note: the overridden initialize method has to invoke super.initialize method in order to initialize parent
 * JMSObjectGridEventListener appropriately
 * <P>
 * The onMessage method will be invoked when the JMSObjectGridEventListener instance is in JMS receiver role and receives 
 * JMS messages that wrap a collection of LogSequences. The collection of LogSequences will be passed to the onMessgae 
 * method and the application can process the LogSequence collection within the onMessage method. A LogSequence 
 * represents a series of map operations, such as INSERT, DELETE, EVICT, and so on. With the collection of LogSequences 
 * passed to OnMessage method, application can know all server side ObjectGrid activities from client ObjectGrid.
 * <P>
 * In summary, to use the onMessage extension point, application has to do the following steps:
 * <ul>
 * <li>Extends JMSObjectGridEventListener class and override the onMessage method that is designated to process the 
 * collection of LogSequences, which represent a series of map operations.</li>
 * <li>Configure the extended JMSObjectGridEventListener the same way as JMSObjectGridEventListener as 
 * ObjectGridEventListener for ObjectGrids.</li>
 * </ul>
 *  
 * @since WAS XD 6.1
 *
 * @see ObjectGridEventListener
 * @see LogSequenceFilter
 */

@PluginOutputFormat(keyFormat=OutputFormat.RAW, valueFormat=OutputFormat.RAW)
public class JMSObjectGridEventListener implements ObjectGridEventListener, LogSequenceFilter, Runnable, ExceptionListener {
    private final static TraceComponent tc = Tr.register(com.ibm.websphere.objectgrid.plugins.builtins.JMSObjectGridEventListener.class,
            Constants.TR_GROUP_NAME, Constants.TR_RESOURCE_BUNDLE_NAME);

    /**
     * The INVALIDATE invalidation or replication strategy is same as
     * {@link DistributionMode#INVALIDATE}
     */
    final static int INVALIDATE = 0;
    /**
     * The INVALIDATE invalidation or replication strategy is same as
     * {@link DistributionMode#INVALIDATE_CONDITIONAL}
     */
    final static int INVALIDATE_CONDITIONAL = 1;
    /**
     * The PUSH invalidation or replication strategy is same as
     * {@link DistributionMode#PUSH}
     */
    final static int PUSH = 2;
    /**
     * The PUSH_CONDITIONAL invalidation or replication strategy is same as
     * {@link DistributionMode#PUSH_CONDITIONAL}
     */
    final static int PUSH_CONDITIONAL = 3;
    /**
     * This is an extension to PUSH strategy.  With PUSH strategy, all updates including INSERT type
     * will be published and received.  Becasue it also includes INSERT type update, the receivers will insert data
     * into its maps locally that might not be desired.  Besides, the receivers might be overwhelmed by inserting too
     * many unnecessary data. The PUSH_INCLUDED strategy is like PUSH strategy, but it does not include INSERT type
     * update. The suffix INCLUDED means the changes only include existing data, and exclude new data that is the
     * result of INSERT operation.
     */
    final static int PUSH_INCLUDED = 4;
    /**
     * This is an extension to PUSH_CONDITIONAL strategy. Similar to the PUSH_INCLUDED, the PUSH_INCLUDED_CONDITIONAL
     * is like PUSH_CONDITIONAL, but it does not include INSERT type update.
     */
    final static int PUSH_INCLUDED_CONDITIONAL = 5;

    /**
     * The constant string that can be used in setInvalidationStrategy(String) or setReplicationStrategy(String) method
     * for setting invalidationStrategy or replicationStrategy as INVALIDATE. The INVALIDATE invalidation or
     * replication strategy is same as  <a href="DistributionMode#INVALIDATE">DistributionMode#INVALIDATE</a>
     */
    public final static String INVALIDATE_STRING = "INVALIDATE";
    /**
     * The constant string that can be used in setInvalidationStrategy(String) or setReplicationStrategy(String) method
     * for setting invalidationStrategy or replicationStrategy as INVALIDATE_CONDITIONAL. The INVALIDATE_CONDITIONAL
     * invalidation or replication strategy is same as
     * <a href="DistributionMode#INVALIDATE_CONDITIONAL">DistributionMode#INVALIDATE_CONDITIONAL</a>
     */
    public final static String INVALIDATE_CONDITIONAL_STRING = "INVALIDATE_CONDITIONAL";
    /**
     * The constant string that can be used in setInvalidationStrategy(String) or setReplicationStrategy(String) method
     * for setting invalidationStrategy or replicationStrategy as PUSH. The PUSH invalidation or replication strategy
     * is same as  <a href="DistributionMode#PUSH">DistributionMode#PUSH</a>
     */
    public final static String PUSH_STRING = "PUSH";
    /**
     * The constant string that can be used in setInvalidationStrategy(String) or setReplicationStrategy(String) method
     * for setting invalidationStrategy or replicationStrategy as PUSH_CONDITIONAL. The PUSH_CONDITIONAL invalidation
     * or replication strategy is same as <a href="DistributionMode#PUSH_CONDITIONAL">DistributionMode#PUSH_CONDITIONAL</a>
     */
    public final static String PUSH_CONDITIONAL_STRING = "PUSH_CONDITIONAL";
    /**
     * The constant string that can be used in setInvalidationStrategy(String) or setReplicationStrategy(String) method
     * for setting invalidationStrategy or replicationStrategy as PUSH_INCLUDED. This is an extension to PUSH strategy.
     * With PUSH strategy, all updates including INSERT type will be published and received.  Becasue it also includes
     * INSERT type update, the receivers will insert data into its maps locally that might not be desired.  Besides,
     * the receivers might be overwhelmed by inserting too many unnecessary data. The PUSH_INCLUDED strategy is like
     * PUSH strategy, but it does not include INSERT type update. The suffix INCLUDED means the changes only include
     * existing data, and  exclude new data that is the result of INSERT operation.
     */
    public final static String PUSH_INCLUDED_STRING = "PUSH_INCLUDED";
    /**
     * The constant string that can be used in setInvalidationStrategy(String) or setReplicationStrategy(String) method
     * for setting invalidationStrategy or replicationStrategy as PUSH_INCLUDED_CONDITIONAL. This is an extension to
     * PUSH_CONDITIONAL strategy. Similar to the PUSH_INCLUDED, the PUSH_INCLUDED_CONDITIONAL is like PUSH_CONDITIONAL,
     * but it does not include INSERT type update.
     */
    public final static String PUSH_INCLUDED_CONDITIONAL_STRING = "PUSH_INCLUDED_CONDITIONAL";


    /**
     * Standalone or local ObjectGrid.
     */
    final static int OBJECTGRID_STANDALONE = 0;
    /**
     * Client ObjectGrid in distributed ObjectGrid environment.
     */
    final static int OBJECTGRID_CLIENT = 1;
    /**
     * Server ObjectGrid in distributed ObjectGrid environment.
     */
    final static int OBJECTGRID_SERVER = 2;

    final static String OBJECTGRID_STANDALONE_STRING = "STANDALONE";
    final static String OBJECTGRID_CLIENT_STRING = "CLIENT";
    final static String OBJECTGRID_SERVER_STRING = "SERVER";

    /**
     * The NONE_INVALIDATION_MODEL invalidation model means the client invalidation mechanism is disabled.
     */
    final static int NONE_INVALIDATION_MODEL = 0;
    /**
     * In the CLIENT_SERVER_MODEL invalidation model, the client ObejctGrid will be in JMS subscriber role and the
     * server ObjectGrid will be in JMS publisher role.
     */
    final static int CLIENT_SERVER_MODEL = 1;
    /**
     * In the CLIENT_AS_DUAL_ROLES_MODEL invalidation model, the client ObejctGrid will be in both JMS subscriber and
     *  and publisher role. The server will not publish or receive any updates.
     */
    final static int CLIENT_AS_DUAL_ROLES_MODEL = 2;

    /**
     * The constant string that can be used in setInvalidationModel(String) method for setting invalidationModel
     * as NONE_INVALIDATION_MODEL that implies disabling client invalidation mechanism.
     */
    public final static String NONE_INVALIDATION_MODEL_STRING = "NONE_INVALIDATION_MODEL";
    /**
     * The constant string that can be used in setInvalidationModel(String) method for setting invalidationModel
     * as CLIENT_SERVER_MODEL. In this model, all server processes will be in publisher role that will publish all
     * transaction changes to designated JMS destination.  All client processes will be in receiver roles and will
     * receive all transactional changes from designated JMS destination.
     */
    public final static String CLIENT_SERVER_MODEL_STRING = "CLIENT_SERVER_MODEL";
    /**
     * The constant string that can be used in setInvalidationModel(String) method for setting invalidationModel
     * as CLIENT_AS_DUAL_ROLES_MODEL. In this model, all server processes have nothing to do with JMS destination. All
     * client processes will be both JMS publisher and receiver roles.  Transactional changes occur on client will be
     * published to JMS destination and all clients will receive these transactional changes.
     */
    public final static String CLIENT_AS_DUAL_ROLES_MODEL_STRING = "CLIENT_AS_DUAL_ROLES_MODEL";

    /**
     * The NONE_ROLE replication role indicates the current ObjectGrid has nothing to do with JMS.  This virtually
     * disable this plugin on the current ObjectGrid instance.
     */
    final static int NONE_ROLE = 0;
    /**
     * The DUAL_ROLES replication role indicates the current ObjectGrid will act as both JMS publisher and subscriber
     * role.
     */
    final static int DUAL_ROLES = 1;
    /**
     * The PUBLISHER_ROLE replication role indicates the current ObjectGrid will only act as JMS publisher role.
     */
    final static int PUBLISHER_ROLE = 2;
    /**
     * The SUBSCRIBER_ROLE replication role indicates the current ObjectGrid will only act as JMS subscriber role.
     */
    final static int SUBSCRIBER_ROLE = 3;
    /**
     * The constant string that can be used in setReplicationRole(String) method for setting replicationRole as
     * NONE_ROLE. The NONE_ROLE replication role indicates the current ObjectGrid has nothing to do with JMS.  This
     * virtually disable this plugin on the current ObjectGrid instance.
     */
    public final static String NONE_ROLE_STRING = "NONE_ROLE";
    /**
     * The constant string that can be used in setReplicationRole(String) method for setting replicationRole as
     * DUAL_ROLES. The DUAL_ROLES replication role indicates the current ObjectGrid will act as both JMS publisher and
     * subscriber role.
     */
    public final static String DUAL_ROLES_STRING = "DUAL_ROLES";
    /**
     * The constant string that can be used in setReplicationRole(String) method for setting replicationRole as
     * PUBLISHER_ROLE. The PUBLISHER_ROLE replication role indicates the current ObjectGrid will only act as JMS
     * publisher role.
     */
    public final static String PUBLISHER_ROLE_STRING = "PUBLISHER_ROLE";
    /**
     * The constant string that can be used in setReplicationRole(String) method for setting replicationRole as
     * SUBSCRIBER_ROLE. The SUBSCRIBER_ROLE replication role indicates the current ObjectGrid will only act as JMS
     * subscriber role.
     */
    public final static String SUBSCRIBER_ROLE_STRING = "SUBSCRIBER_ROLE";

    /**
     * If specified this is the user id when obtaining a JMS connection.
     */
    String userid;

    /**
     * This must be specified when JMS user id is specified.
     */
    String password;

    /**
     * This is the JNDI binding information of the topicConnectionFactory.
     */
    String topicConnectionFactoryJndiName = "defaultTCF";

    /**
     * This is the JNDI binding information of the topic.
     */
    String topicJndiName = "defaultTopic";

    /**
     * This is a user specified property with the topic connection factory.
     */
    TopicConnectionFactory topicConnectionFactory;

    /**
     * This is a user specified property for the topic to use for publishing.
     */
    Topic topic;

    /**
     * This allows the topic to be specified using a string.
     */
    String topicName = "defaultTopic";

    /**
     * This is the distribution mode of the cluster.
     */
    DistributionMode mode = DistributionMode.INVALIDATE; //default

    /**
     * This is the extension flag for PUSH and PUSH_CONDITIONAL DistributionMode. If true, only perform on existing
     * CacheEntry in the BaseMap. This means to exclude INSERT LogElement. Set the default to true to exclude INSERT
     * LogElement during serialization/inflate.
     */
    boolean includedMode = false;

    /**
     * This is the topic connection for publishing messages.
     */
    TopicConnection publishConnection;

    /**
     * This is the JMS session used for publishing messages.
     */
    TopicSession publishSession;

    /**
     * This is the message publisher for publishing updates.
     */
    TopicPublisher publisher;

    TopicConnection subscriberConnection;

    TopicSession subscriberSession;

    TopicSubscriber subscriber;

    /**
     * This is the grid associated with this listener.
     */
    ObjectGrid myGrid;

    /**
     * This is the session associated with this listener.
     */
    Session mySession;

    /**
     * This is the thread listening for messages.
     */
    Thread listenerThread;

    boolean listenerRunning;

    Set/* <String> */publishMaps = new HashSet();

    String processId = null;

    Properties jndiProperties;

    /*
     * The invalidationModel flag is to determine the model of client invalidation mechasnism.
     * Three models are defined: NONE_INVALIDATION_MODEL, CLIENT_SERVER_MODEL and CLIENT_AS_DUAL_ROLES_MODEL.
     * The NONE_INVALIDATION_MODEL indicates this is not for client invalidation mechanism, and it might be used for
     * peer-2-peer replication in which situation the replicationRole can be set directly.
     *
     */
    int invalidationModel = NONE_INVALIDATION_MODEL;

    /*
     * The invalidationStrategy flag is to determine the DistributionMode type in replication process.
     * Six strategies are defined: INVALIDATE, INVALIDATE_CONDITIONAL, PUSH, PUSH_CONDITIONAL
     *     PUSH_INCLUDED, PUSH_INCLUDED_CONDITIONAL
     */
    int invalidationStrategy = INVALIDATE;

    /*
     * The replicationRole flag is to determine how to manage JMS publilsher and subscriber.
     * Three roles are defined: DUAL_ROLES, PUBLISHER_ROLE, and SUBSCRIBER_ROLE
     */
    int replicationRole = NONE_ROLE;

    /*
     * The ObjectGrid could be STANDALONE, CLIENT, and SERVER.
     */
    int objectgridType;

    /*
     * The enableOnClientObjectGrid control whether or not to run this plugin on the client ObjectGrid
     * in the distributed ObjectGrid env.  The default is true.  If wish to disable the plugin on client ObjectGrid,
     * set it to false.
     */
    boolean enableOnClientObjectGrid = true;

    /*
     * The enableOnServerObjectGrid control whether or not to run this plugin on the server ObjectGrid
     * in the distributed ObjectGrid env.  The default is true.  If wish to disable the pugin on server ObjectGrid,
     * set it to false.
     */
    boolean enableOnServerObjectGrid = true;
        
    Set currentTxIdSet = new HashSet();
    

	/**
     * This constructs the object and initializes any variables.
     */
    public JMSObjectGridEventListener() {
        super();
    }

    /*
     * The configuredMapsToPublish holds the configured map names that should be included in publishMaps
     */
    List configuredMapsToPublish = new ArrayList();
    
    List validConfiguredMapsToPublish = new ArrayList();
    
    List invalidConfiguredMapsToPublish = new ArrayList();
    
    List mapNamesWithNullBackingMap = new ArrayList();
    
    List pessimisticMaps = new ArrayList();
    
    List actualCompleteMapNames = new ArrayList();
    
    boolean publishAllMaps = false;

    /**
     * This is called when the ObjectGrid is initialized. This method obtains JMS connections and starts them. It then
     * creates a TopicSession to publish all changes in the local ObjectGrid. It creates a thread to listen for messages
     * from peer ObjectGrids.
     *
     * @see com.ibm.websphere.objectgrid.plugins.ObjectGridEventListener#initialize(com.ibm.websphere.objectgrid.Session)
     */
    public void initialize(Session session) {
        if (ObjectGridManagerImpl.isTraceEnabled && tc.isEntryEnabled()) {
            Tr.entry(tc, "initialize");
        }

        mySession = session;
        myGrid = session.getObjectGrid();

        setObjectGridType();

        if(objectgridType==OBJECTGRID_CLIENT && enableOnClientObjectGrid==false){
            if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled()) {
                Tr.debug(tc, "This ObjectGridEventListener instance is disabled becasue enableOnClientObjectGrid is false and the ObjectGrid is client ObjectGrid.");
            }
            myGrid.removeEventListener(this);
            return;
        } else if(objectgridType==OBJECTGRID_SERVER && enableOnServerObjectGrid==false){
            if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled()) {
                Tr.debug(tc, "This ObjectGridEventListener instance is disabled becasue enableOnServerObjectGrid is false and the ObjectGrid is server ObjectGrid.");
            }
            myGrid.removeEventListener(this);
            return;
        }
        
        List completeMapNames = myGrid.getListOfMapNames();
        // verify if ObjectGrid instance initialize each map.
        // maps could be configured , but not necessary to be in the same MapSet
        // An ObjectGrid instacne in a Shard could only host some maps in multiple MapSet per ObjectGrid case.
        
        for(int i=0; i< completeMapNames.size(); i++){
        	String mapName = (String)completeMapNames.get(i);
    		if(myGrid.getMap(mapName) != null) {
    			actualCompleteMapNames.add(mapName);
    		}  
        }  
        
        if(configuredMapsToPublish==null || configuredMapsToPublish.size()==0){
        	// users do not configure this property or it is empty
            //if not configured, implies to include all maps.
            //should exclude PESSIMISTIC maps, replicated maps, and other maps that turn off the invalidation.
            setMapsToPublish(actualCompleteMapNames);           
        } else {
        	// Users specifies maps to publish scenario
        	//need to processConfiguredMapsToPublish first.
            processConfiguredMapsToPublish();
            // if publishMaps.size==0, means no configured maps or even configured but not valid.
            // when users specify maps to publish, these map names could be invalid
            // if all are invalid, the publishMaps is empty, this is similar to incorrect configuration
            // in this case, turn off this listener.
        }
        
        boolean forceDisableJMSPublisher = false;
        
        if(publishMaps.size()==0) {
        	// nothing to publish on this listener, some BackingMaps may be not initialized.
        	forceDisableJMSPublisher = true;
        } else if(publishMaps.size()==actualCompleteMapNames.size()){
            //means all maps are included.
            //empty the publishMaps to tune performance during transactionEnd()
            publishMaps.clear();
            publishAllMaps = true;
        }


        setCurrentReplicationRole();

        displayConfigurationInfo();
        
        // clean temporary storages
        validConfiguredMapsToPublish = null;
        invalidConfiguredMapsToPublish = null;
        mapNamesWithNullBackingMap = null;
        pessimisticMaps = null;
        actualCompleteMapNames = null;


        //at this point, we know the replication role of the current instance.
        //throw exception if the role is NONE_ROLE, to disable this plugin
        if (replicationRole == NONE_ROLE) {
            if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled()) {
                Tr.debug(tc, "This ObjectGridEventListener instance is disabled becasue the replication role is NONE_ROLE");
            }
            myGrid.removeEventListener(this);
            return;
        } else  if(replicationRole == PUBLISHER_ROLE && forceDisableJMSPublisher){
        	if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled()) {
                Tr.debug(tc, "This ObjectGridEventListener instance is disabled becasue the replication role is PUBLISHER_ROLE and there is no actual map to publish changes.");
            }
            myGrid.removeEventListener(this);
            return;
        } 

        try {
            initializeJMS();

            if(replicationRole == PUBLISHER_ROLE){
                startJMSPublisher();
            } else if (replicationRole == SUBSCRIBER_ROLE){
                startJMSSubscriber();
            } else if (replicationRole == DUAL_ROLES){
                startJMSPublisher();
                startJMSSubscriber();
            }

        } catch (Throwable e) {
            e.printStackTrace();
            throw new ObjectGridRuntimeException("Cannot initialize", e);
        }

        if (ObjectGridManagerImpl.isTraceEnabled && tc.isEntryEnabled()) {
            Tr.exit(tc, "initialize");
        }
    }

    private void setObjectGridType(){
        int objectGridTypeInt = myGrid.getObjectGridType();        
        if(objectGridTypeInt == ObjectGrid.LOCAL){
            objectgridType = OBJECTGRID_STANDALONE;
        } else if(objectGridTypeInt == ObjectGrid.CLIENT) {
            objectgridType = OBJECTGRID_CLIENT;
        } else if(objectGridTypeInt == ObjectGrid.SERVER) {
            objectgridType = OBJECTGRID_SERVER;
        }
    }

    /*
     * This method will set the replication role on the current instance.
     * The role is determined by the configured invalidation model
     */
    private void setCurrentReplicationRole(){
        if(invalidationModel==CLIENT_SERVER_MODEL){
            if(objectgridType==OBJECTGRID_SERVER){
                setReplicationRole(PUBLISHER_ROLE);
            } else if(objectgridType==OBJECTGRID_CLIENT){
                setReplicationRole(SUBSCRIBER_ROLE);
            } else if (objectgridType==OBJECTGRID_STANDALONE){
                setReplicationRole(NONE_ROLE);
            }
        } else if(invalidationModel==CLIENT_AS_DUAL_ROLES_MODEL){
            if(objectgridType==OBJECTGRID_SERVER){
                setReplicationRole(NONE_ROLE);
            } else if(objectgridType==OBJECTGRID_CLIENT){
                setReplicationRole(DUAL_ROLES);
            } else if (objectgridType==OBJECTGRID_STANDALONE){
                setReplicationRole(NONE_ROLE);
            }
        }
    }


    /**
     * Obtain JMS objects
     */
    private void initializeJMS() throws Exception {
        if (ObjectGridManagerImpl.isTraceEnabled && tc.isEntryEnabled()) {
            Tr.entry(tc, "initializeJMS");
        }

        javax.naming.Context ic = null;
        if (jndiProperties != null) {
            ic = new InitialContext(jndiProperties);
        } else {
            // use default constructor.
            // this should work in WAS server env.
            ic = new InitialContext();
        }
        
        // in WAS, when the app installed on the AppServer that also is a SIBus messgae engine member
        // the message member may not be ready when initializing ObjectGrid.
        // We can wait up to 60 seconds when trying to find JMS service
        int retryCounter = 0;
        while(topicConnectionFactory == null) {
            try{
                topicConnectionFactory = (TopicConnectionFactory) ic.lookup(topicConnectionFactoryJndiName);
                topic = (Topic) ic.lookup(topicJndiName);
                break;
            } catch (Exception e) {
                retryCounter++;           
                if(retryCounter==10) {
                    throw e;
                } else {
                    Thread.sleep(6000);
                }
            }
        }
     
        if (ObjectGridManagerImpl.isTraceEnabled && tc.isEntryEnabled()) {
            Tr.exit(tc, "initializeJMS");
        }
    }

    /**
     * create JMS publisher
     */
    private void startJMSPublisher() throws Exception {
        if (ObjectGridManagerImpl.isTraceEnabled && tc.isEntryEnabled()) {
            Tr.entry(tc, "startJMSPublisher");
        }
        
        if (userid != null) {
            publishConnection = topicConnectionFactory.createTopicConnection(userid, password);
        } else {
            publishConnection = topicConnectionFactory.createTopicConnection();
        }

        // publishConnection.setExceptionListener(this);

        publishConnection.start();

        // the jms session is not transactional (false).
        publishSession = publishConnection.createTopicSession(false, javax.jms.Session.AUTO_ACKNOWLEDGE);

        if (topic == null) {
            if (topicName == null) {
                throw new ObjectGridRuntimeException("Topic not specified");
            } else {
                topic = publishSession.createTopic(topicName);
            }
        }

        publisher = publishSession.createPublisher(topic);

        if (ObjectGridManagerImpl.isTraceEnabled && tc.isEntryEnabled()) {
            Tr.exit(tc, "startJMSPublisher");
        }
    }

    /**
     * create JMS receiver
     */
    private void startJMSSubscriber() throws Exception {
        if (ObjectGridManagerImpl.isTraceEnabled && tc.isEntryEnabled()) {
            Tr.entry(tc, "startJMSSubscriber");
        }

        listenerRunning = true;
        listenerThread = new Thread(SecurityContextRunnableFactory.getRunnable(this));
        listenerThread.start();

        if (ObjectGridManagerImpl.isTraceEnabled && tc.isEntryEnabled()) {
            Tr.exit(tc, "startJMSSubscriber");
        }
    }

    /*
     * (non-Javadoc)
     *
     * @see com.ibm.websphere.objectgrid.plugins.ObjectGridEventListener#transactionBegin(java.lang.String, boolean)
     */
    public void transactionBegin(String txid, boolean isWriteThroughEnabled) {
    }

    /*
     * (non-Javadoc)
     *
     * @see com.ibm.websphere.objectgrid.plugins.ObjectGridEventListener#transactionEnd(java.lang.String, boolean,
     *      boolean, java.util.Collection)
     */

    /**
     * This is a JMS message property for the txid.
     */
    final static String PROP_TX = "TXID";

    /**
     * This is a JMS message property for the grid name.
     */
    final static String PROP_GRIDNAME = "NAME";

    /**
     * This is a JMS message property for the number of sequences in a message.
     */
    final static String PROP_COUNT = "CNT";

    /**
     * This is a JMS message property for the process id.
     */
    final static String PROP_PID = "PID";
    
    /**
     * This is called whenever a transaction commits for the grid. If its a write through transaction then we will
     * publish a message with the changes. This is synchronised to make sure only one thread uses the session plus it
     * enforces an order on the messages transmitted.
     */
    public synchronized void transactionEnd(String txid, boolean isWriteThroughEnabled, boolean committed,
            Collection changes) {
        // Only PUBLISHER_ROLE and DUAL_ROLES need to publish messages.
        if (replicationRole == SUBSCRIBER_ROLE || replicationRole == NONE_ROLE) {
            return;
        }

        if (processId == null) {
            // use the first txid as the process id.
            processId = txid;
        }
        
        try {
            if (!currentTxIdSet.remove(txid) && committed)  { 
            	Iterator iterator = changes.iterator();
                Collection publishChanges = new ArrayList();
                while (iterator.hasNext()) {
                    LogSequence ls = (LogSequence) iterator.next();
                    boolean needToPublish = false;                        
                    if(publishMaps.isEmpty() || publishMaps.contains(ls.getMapName())) {
                    	 Iterator logElementIter = ls.getAllChanges();
                         for (int i = ls.size() - 1; i >= 0; --i) {
                             LogElement le = (LogElement) logElementIter.next();	 
                             if(accept(le)) {
                             	// as long as at least one log element pass accept() method
                             	// need to publish it.
                             	needToPublish = true;
                             	break;
                             }
                         }
                                                  
                         if (needToPublish) {
                             publishChanges.add(ls);
                         }
                    }                        
                } 

                if (publishChanges.size() > 0) {
                    ByteArrayOutputStream bos = new ByteArrayOutputStream();
                    ObjectOutputStream oos = new ObjectOutputStream(bos);

                    // the LogSequenceTransformer can be used in 6011 and later
                    // use the LogSequenceTransformer to serialize changes
                    // it will use LogSequenceFilter to filter out unwanted LogElements
                    LogSequenceTransformer.serialize(publishChanges, oos, this, mode);

                    if (publishChanges.size() > 0) {
                        Iterator iter = publishChanges.iterator();
                        boolean isPublishChangesEmpty = true;
                        while (iter.hasNext()) {
                            LogSequence ls = (LogSequence) iter.next();
                            if (ls.size() > 0) {
                                isPublishChangesEmpty = false;
                                // -> need to publish message
                                break;
                            }
                        }

                        if (!isPublishChangesEmpty) {
                            // make an object message for the changes
                            oos.flush();
                            ObjectMessage om = publishSession.createObjectMessage(bos.toByteArray());
                            // set properties
                            om.setStringProperty(PROP_TX, txid);
                            om.setStringProperty(PROP_GRIDNAME, myGrid.getName());
                            om.setStringProperty(PROP_PID, processId);

                            if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled()) {
                                Tr.debug(tc, "transactionEnd -> Publishing message ... <"
                                        + om.getStringProperty(PROP_GRIDNAME) + ", " + om.getStringProperty(PROP_TX));
                            }

                            // Should need to check if publishConnection is OK?
                            if (topicConnectionFactory == null || publishConnection == null) {
                                refreshPublishConnection();
                            }

                            publisher.publish(om);

                        } else {

                        }

                    } else {

                    }
                } else {
                	if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled()) {
                        Tr.debug(tc, "transactionEnd -> current transaction is originated by this instance, do not process it, currentTxIdSet.size() = " + currentTxIdSet.size());
                    }
                }
            }
        } catch (Throwable e) {
            // The javax.jms.JMSException: MQJMS3011: Failed to publish message to MQ queue
            // may occur when system is overloaded and the publisher fails to publish message.

            e.printStackTrace();
            refreshPublishConnection();

            // should need to store failed sending messages, and re-send?
        }
    }

    private void refreshPublishConnection() {
        if (ObjectGridManagerImpl.isTraceEnabled && tc.isEntryEnabled()) {
            Tr.entry(tc, "refreshPublishConnection");
        }

        try {
            if (publishConnection != null) {
                publishConnection.close();
            }
        } catch (Throwable t) {
            t.printStackTrace();
        }

        try {
            if (userid != null) {
                publishConnection = topicConnectionFactory.createTopicConnection(userid, password);
            } else {
                publishConnection = topicConnectionFactory.createTopicConnection();
            }

            publishConnection.start();

            // the jms session is not transactional (false).
            publishSession = publishConnection.createTopicSession(false, javax.jms.Session.AUTO_ACKNOWLEDGE);

            if (topic == null) {
                if (topicName == null) {
                    throw new ObjectGridRuntimeException("Topic not specified");
                } else {
                    topic = publishSession.createTopic(topicName);
                }
            }

            publisher = publishSession.createPublisher(topic);

            if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled()) {
                Tr.debug(tc, "refreshPublishConnection -> passed");
            }
        } catch (Throwable t) {
            t.printStackTrace();
            if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled()) {
                Tr.debug(tc, "refreshPublishConnection -> faiiled");
            }
        }

        if (ObjectGridManagerImpl.isTraceEnabled && tc.isEntryEnabled()) {
            Tr.exit(tc, "refreshPublishConnection");
        }
    }

    private void refreshSubscriberConnection() {
        if (ObjectGridManagerImpl.isTraceEnabled && tc.isEntryEnabled()) {
            Tr.entry(tc, "refreshReceiverConnection");
        }

        try {
            if (subscriberConnection != null) {
                subscriberConnection.close();
            }
        } catch (Throwable t) {
            t.printStackTrace();
        }

        try {
            if (userid != null)
                subscriberConnection = topicConnectionFactory.createTopicConnection(userid, password);
            else
                subscriberConnection = topicConnectionFactory.createTopicConnection();

            subscriberConnection.start();
            // get a jms session for receiving the messages. Non transactional.
            subscriberSession = subscriberConnection.createTopicSession(false, javax.jms.Session.AUTO_ACKNOWLEDGE);
            // get a subscriber for the topic, true indicates don't receive messages transmitted using
            // publishers
            // on this connection. Otherwise, we'd receive our own updates.
            subscriber = subscriberSession.createSubscriber(topic, null, true);
            if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled()) {
                Tr.debug(tc, "refreshReceiverConnection -> passed");
            }
        } catch (Throwable t) {
            t.printStackTrace();
            if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled()) {
                Tr.debug(tc, "refreshReceiverConnection -> faiiled");
            }
        }

        if (ObjectGridManagerImpl.isTraceEnabled && tc.isEntryEnabled()) {
            Tr.exit(tc, "refreshReceiverConnection");
        }
    }

    /**
     * This is used to filter the LogElements written to each message. We use the mode to determine the filter.
     *
     * @see com.ibm.websphere.objectgrid.plugins.LogSequenceFilter#accept(com.ibm.websphere.objectgrid.plugins.LogElement)
     */
    public boolean accept(LogElement logElement) {
        boolean toReturn = false;
        LogElement.Type type = logElement.getType();
        if (mode == DistributionMode.INVALIDATE || mode == DistributionMode.INVALIDATE_CONDITIONAL) {
            toReturn = (type == LogElement.EVICT || type == LogElement.UPDATE || type == LogElement.DELETE || type == LogElement.CLEAR);
        } else {
            if (includedMode) {
                // exclude INSERT LogElement.
                toReturn = (type == LogElement.EVICT || type == LogElement.UPDATE || type == LogElement.DELETE || type == LogElement.CLEAR);
            } else {
                toReturn = (type == LogElement.EVICT || type == LogElement.UPDATE || type == LogElement.DELETE || type == LogElement.CLEAR)
                        || (type == LogElement.INSERT);
            }

        }

        if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled()) {
            Tr.debug(tc, "accept() = " + toReturn + ", " + logElement.getType());
        }

        return toReturn;
    }

    private synchronized boolean isListenerRunning() {
        return listenerRunning;
    }

    private synchronized void setIsListenerRunning(boolean isRunning) {
        listenerRunning = isRunning;
    }

    /**
     * This is the thread which receives messages on a topic and then applies them to the local grid.
     *
     * @see java.lang.Runnable#run()
     */
    public void run() {
        try {
            // The 1st call on refreshSubscriberConnection will lauch the subscriber for use.
            // if javax.jms.IllegalStateException from JMS communication, refreshSubscriberConnection again.
            refreshSubscriberConnection();

            while (isListenerRunning()) {
                ObjectMessage om = null;
                try {
                    om = (ObjectMessage) subscriber.receive(2000);
                } catch (javax.jms.IllegalStateException e) {
                    // The JMS broker might not be available temporarily
                    try {
                        Thread.sleep(1000);
                    } catch (Throwable t) {
                        // ...
                    }
                    refreshSubscriberConnection();
                }

                if (om != null) {
                	final boolean debug = (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled());
                	StringBuffer buffer = null;
                	if(debug) {
                        buffer = new StringBuffer();
                        buffer.append("Received JMS message <" + om.getStringProperty(PROP_GRIDNAME) + ", "
                                + om.getStringProperty(PROP_TX));
                	}
                    
                    String msgpid = om.getStringProperty(PROP_PID);
                    boolean isMessageSentByMe = false;
                    if (this.processId != null && this.processId.equals(msgpid)) {
                        isMessageSentByMe = true;
                        if(debug) {
                        	buffer.append(" -> This message is sent by me. Ignore it.\n");
                        }
                    }

                    if (!isMessageSentByMe) {
                    	if(debug) {
                    		buffer.append(" -> This message comes from peers. Process it.");
                    	}
                        boolean committed = false;
                        try {
                            // Use Session that was passed in on the initialize...
                            mySession.beginNoWriteThrough();
                            String currentTxId = mySession.getTxID().toString();
                            currentTxIdSet.add(currentTxId);
                            if(debug) {
                            	buffer.append(" currentTxId = " + currentTxId + "\n");
                            	buffer.append(" currentTxIdSet.size() = " + currentTxIdSet.size() + "\n");
                        	}
                            
                            byte[] raw = (byte[]) om.getObject();
                            ByteArrayInputStream bis = new ByteArrayInputStream(raw);
                            ObjectInputStream ois = new ObjectInputStreamResolver(bis);

                            // using the LogSequenceTransformer.inflate()
                            // the following can only be used on local ObjectGrid instance
                            // or distributed ObjectGrid instance in 6011 or above/
                            Collection collection = LogSequenceTransformer.inflate(ois, myGrid);

                            Iterator iter = collection.iterator();
                            if(debug) {
	                            if (collection.size() == 0) {
	                                buffer.append(" LogSequencesCollectionSize= " + collection.size() + "\n");
	                            } else {
	                                buffer.append(" LogSequencesCollectionSize= " + collection.size());
	                            }
                            }

                            while (iter.hasNext()) {
                                LogSequence seq = (LogSequence) iter.next();
                                
                                // when debug=true, printout the content of each LogSequence.
                                if(debug) {
                                	buffer.append(", currentLogsequenceSize= " + seq.size() + "\n");
                                	buffer.append("\tMap Name=" + seq.getMapName() + "\n");
	                                Iterator logElementIter = seq.getAllChanges();
	                                for (int i = seq.size() - 1; i >= 0; --i) {
	                                    LogElement le = (LogElement) logElementIter.next();	                                    
		                                buffer.append("\tMap Name=" + seq.getMapName() + "\n");
		                                buffer.append("\t" + le.getType() + " -> key=" + le.getKey() + ", ");
	                                }
	                                buffer.append("\n");
                                }
                                
                                // Note: It is important that application have to handle transaction update conflicts
                                // during the processing changes come from peers.
                                // The common way is to use OptimisticCallback on BackingMap to perform version checking
                                // if the LockStrategy is OPTIMISTIC.
                                // The bottom line is the application is responsible for manage object version,
                                // otherwise, lost update or data inconsistency will occur.

                                mySession.processLogSequence(seq);
                            }

                            onMessage(collection);
                            
                            mySession.commit();
                            committed = true;

                            // catch occasioal exception and continue the listener.
                        } catch (IOException e) {
                            e.printStackTrace();
                        } catch (JMSException e) {
                            e.printStackTrace();
                        } catch (ObjectGridException e) {
                            e.printStackTrace();
                        } catch (Throwable t) {
                            // unexpected, stop JMS Listener
                            throw t;
                        } finally {
                            if (!committed) {
                                try {
                                    mySession.rollback();
                                } catch (Exception e) {
                                    // ignoire
                                }
                            }

                        }
                    }

                    if (debug) {
                        Tr.debug(tc, buffer.toString());
                    }
                }
            }

            subscriberConnection.close();
            setIsListenerRunning(false);

        } catch (IllegalStateException e) {
            e.printStackTrace();
            // The IllegalStateException occurs when the ObjectGrid instance is destroyed.
            // this is expected if the sample call destroy method on ObjectGrid instances,
            // which will call the destroy method of this JMSCluster.
        } catch (Throwable e) {
            e.printStackTrace();
        } finally {
            if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled()) {
                Tr.debug(tc, "JMS Listener stopped");
            }

            setIsListenerRunning(false);
            if (subscriberConnection != null)
                try {
                    subscriberConnection.close();
                } catch (JMSException e) {
                }
        }

    }

    /**
     * @return Returns the DistributionMode.
     */
    private DistributionMode getMode() {
        return mode;
    }

    /**
     * @param mode
     *            The DistributionMode to set.
     */
    private void setMode(DistributionMode mode) {
        this.mode = mode;
    }

    /**
     * @return Returns the JMS topic.
     */
    public Topic getTopic() {
        return topic;
    }

    /**
     * @param topic
     *            The JMS topic to set.
     */
    public void setTopic(Topic topic) {
        this.topic = topic;
    }

    /**
     * @return Returns the topicConnectionFactory.
     */
    public TopicConnectionFactory getTopicConnectionFactory() {
        return topicConnectionFactory;
    }

    /**
     * @param topicConnectionFactory
     *            The topicConnectionFactory to set.
     */
    public void setTopicConnectionFactory(TopicConnectionFactory topicConnectionFactory) {
        this.topicConnectionFactory = topicConnectionFactory;
    }

    /**
     * @return Returns the topicName.
     */
    private String getTopicName() {
        return topicName;
    }

    /**
     * @param topicName
     *            The topicName to set.
     */
    private void setTopicName(String topicName) {
        this.topicName = topicName;
    }

    /**
     * @return Returns the String representation of the invalidationStrategy.
     */
    public String getInvalidationModelString(int invalidationModelInt) {
        switch (invalidationModelInt) {
        case NONE_INVALIDATION_MODEL:
            return NONE_INVALIDATION_MODEL_STRING;
        case CLIENT_SERVER_MODEL:
            return CLIENT_SERVER_MODEL_STRING;
        case CLIENT_AS_DUAL_ROLES_MODEL:
            return CLIENT_AS_DUAL_ROLES_MODEL_STRING;
        default:
            return null;
        }
    }

    /**
     * @return Returns the String representation of the invalidationStrategy.
     */
    private String getReplicationRoleString(int replicationRoleInt) {
        switch (replicationRoleInt) {
        case NONE_ROLE:
            return NONE_ROLE_STRING;
        case DUAL_ROLES:
            return DUAL_ROLES_STRING;
        case PUBLISHER_ROLE:
            return PUBLISHER_ROLE_STRING;
        case SUBSCRIBER_ROLE:
            return SUBSCRIBER_ROLE_STRING;
        default:
            return null;
        }
    }


    /**
     * @return Returns the String representation of the invalidationStrategy.
     */
    private String getInvalidationStrategyString(int invalidationStrategyInt) {
        switch (invalidationStrategyInt) {
        case INVALIDATE:
            return INVALIDATE_STRING;
        case INVALIDATE_CONDITIONAL:
            return INVALIDATE_CONDITIONAL_STRING;
        case PUSH:
            return PUSH_STRING;
        case PUSH_CONDITIONAL:
            return PUSH_CONDITIONAL_STRING;
        case PUSH_INCLUDED:
            return PUSH_INCLUDED_STRING;
        case PUSH_INCLUDED_CONDITIONAL:
            return PUSH_INCLUDED_CONDITIONAL_STRING;
        default:
            return null;
        }
    }

    /**
     * @return Returns the String representation of the ObjectGrid type.
     */
    private String getObjectGridTypeString(int objectgridTypeInt) {
        switch (objectgridTypeInt) {
        case OBJECTGRID_STANDALONE:
            return OBJECTGRID_STANDALONE_STRING;
        case OBJECTGRID_CLIENT:
            return OBJECTGRID_CLIENT_STRING;
        case OBJECTGRID_SERVER:
            return OBJECTGRID_SERVER_STRING;
        default:
            return null;
        }
    }

    /**
     * @param userid
     *            The userid to set.
     */
    public void setCredentials(String userid, String pass) {
        this.userid = userid;
        password = pass;
    }

    private boolean containsElement(List list, Object element){
        boolean result = false;
        for(int i=0; i<list.size(); i++){
            if(list.get(i).equals(element)){
                result = true;
            }
        }
        return result;
    }

    /*
     * This method will filter out invalild map name from a list and return a new list.
     */
    private List getValidMapNames(List mapNames, boolean isProcessingConfiguredMapsToPublish){
        List validMapNames = new ArrayList();
        //need to verify if map names are valid.
        //The complete map names can be obtained from myGrid.
        // a valid map name does not means there will always be a BackingMap instance created in server side ObjectGrid,
        // becasue a MapSet may only contains partial list of maps.
        List completeMapNames = myGrid.getListOfMapNames();
        for(int i=0; i<mapNames.size(); i++){
            Object element = mapNames.get(i);
            if(containsElement(completeMapNames, element)){
                validMapNames.add(element);
            } else {
            	if(isProcessingConfiguredMapsToPublish) {
            		invalidConfiguredMapsToPublish.add(element);
            	}
            }
        }

        return validMapNames;
    }

    /*
     * This method will filter out unnecessary maps from a list and return a new list.
     * Unnecessary maps includes PESSIMISTIC, REPLICATED maps, and mapName has no corresponding BackingMap.
     * Note: in multiple MapSets per ObjectGrid case, myGrid.getMap(mapName); may return null.
     * becasue a MapSet is corresponding to an ObjectGrid instance, when a BackingMap is not deployed into a MapSet,
     * the ObjectGrid instance will not initiallize it, even the mapName is inlcuded in ObjetcGrid.getMapNames().
     */
    private List excludeUnnecessaryMaps(List mapNames){
        List excludedMapNames = new ArrayList();
        
        boolean debug = ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled();
        StringBuffer buffer =null;
        if (debug) {
			buffer = new StringBuffer();			
			if(mapNames.size()==0)
            	buffer.append("\nmapNames to process = []\n");
	        else 
	        	buffer.append("\nmapNames to process = [");
		}
                
        for(int i=0; i< mapNames.size(); i++){
            String mapName = (String)mapNames.get(i);            
            if (debug) {
				if (i == mapNames.size() - 1)
					buffer.append(mapName + "]\n");
				else
					buffer.append(mapName + ", ");
			}
        }
            
        for(int i=0; i< mapNames.size(); i++){
        	try {
        		String mapName = (String)mapNames.get(i);
        		BackingMap bMap = myGrid.getMap(mapName);
        		if(myGrid.getMap(mapName) == null) {
        			excludedMapNames.add(mapName);
        			mapNamesWithNullBackingMap.add(mapName);
        		} else {
        			LockStrategy lockStrategy = bMap.getLockStrategy();
        			if(lockStrategy != null) {
        				if(lockStrategy.equals(LockStrategy.PESSIMISTIC)) {
        					excludedMapNames.add(mapName);        					
        					pessimisticMaps.add(mapName);
        				}
        			}
        		}
        	} catch (Throwable t) {
        		t.printStackTrace();
        	}
            
        }
        
        if (debug){
        	 if(excludedMapNames.size()==0)
        		 buffer.append("mapNames to exclude = []"); 	
             else 
            	 buffer.append("mapNames to exclude = [");             
        }
        
        for(int i=0; i< excludedMapNames.size(); i++){
        	Object mapName = excludedMapNames.get(i);
        	mapNames.remove(mapName);
        	if (debug) {
				if (i == excludedMapNames.size() - 1)
					buffer.append(mapName + "]");
				else
					buffer.append(mapName + ", ");
			}                       
        }
                       
        if (debug)
        	Tr.debug(tc, "excludeUnnecessaryMaps", buffer.toString());
        
        return mapNames;
    }


    /**
     * If this is specified then only the maps in the list will be published. This allows a subset of the maps in a
     * transaction to be published.
     *
     * @param mapNames
     */
    public void setMapsToPublish(List mapNames) {
        if(mapNames==null || mapNames.size()==0){
            return;
        }

        //filter out invalid map names
        mapNames = getValidMapNames(mapNames, false);

        if(invalidationModel != NONE_INVALIDATION_MODEL){
            //this is used in client invalidation mechanism, exclude pessimistic maps,
            //even they are configured to have invalidation mechanism,
            //becasue pessimistic maps do not have near cache.
            mapNames = excludeUnnecessaryMaps(mapNames);
        }

        Iterator iter = mapNames.iterator();
        while (iter.hasNext()) {
            String s = (String) iter.next();
            if(!publishMaps.contains(s)){
                publishMaps.add(s);
            }
        }

    }

    /**
     * Process configured maps to publish.
     *
     *
     */
    private void processConfiguredMapsToPublish() {
        if(configuredMapsToPublish==null || configuredMapsToPublish.size()==0){
            if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled()) {
                Tr.debug(tc, "processConfiguredMapsToPublish -> configuredMapsToPublish is empty");
            }
            return;
        }

        //need to verify if map names are valid.
        //The complete map names can be obtained from myGrid.
        validConfiguredMapsToPublish = getValidMapNames(configuredMapsToPublish, true);
        
        if(invalidationModel != NONE_INVALIDATION_MODEL){
            //this is used in client invalidation mechanism, exclude pessimistic maps,
            //even they are configured to have invalidation mechanism,
            //becasue pessimistic maps do not have near cache.
            validConfiguredMapsToPublish = excludeUnnecessaryMaps(validConfiguredMapsToPublish);
        }

        Iterator iter = validConfiguredMapsToPublish.iterator();
        while (iter.hasNext()) {
            String s = (String) iter.next();
            if(!publishMaps.contains(s)){
                publishMaps.add(s);
            }
        }
    }


    /*
     * (non-Javadoc)
     *
     * @see javax.jms.ExceptionListener#onException(javax.jms.JMSException)
     */
    public void onException(JMSException arg0) {
        synchronized (this) {
            listenerRunning = false;
        }
    }

    /**
     * This is called when the application tells the ObjectGrid to stop. We will signal our listener thread to stop. The
     * listener thread will also close the JMS connection.
     *
     * @see com.ibm.websphere.objectgrid.plugins.ObjectGridEventListener#destroy()
     */
    public synchronized void destroy() {
        listenerRunning = false;
        try {
            if (publishConnection != null) {
                publishConnection.close();
            }
        } catch (Throwable t) {
            t.printStackTrace();
        }
    }

    public void release() {
        destroy();
    }


    /**
     * Setter for the jms_topicConnectionFactoryJndiName property
     */
    public void setJms_topicConnectionFactoryJndiName(String s) {
        if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled()) {
            Tr.debug(tc, "setJms_topicConnectionFactoryJndiName -> " + s);
        }
        topicConnectionFactoryJndiName = s;
    }

    /**
     * Setter for the jms_topicJndiName property
     */
    public void setJms_topicJndiName(String s) {
        if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled()) {
            Tr.debug(tc, "setJms_topicJndiName -> " + s);
        }
        topicJndiName = s;
    }

    /**
     * Setter for the topicName property
     */
    public void setJms_topicName(String s) {
        if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled()) {
            Tr.debug(tc, "setJms_topicName -> " + s);
        }
        topicName = s;
    }

    /**
     * Setter for the jms_userid property
     */
    public void setJms_userid(String s) {
        if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled()) {
            Tr.debug(tc, "setJms_userid -> " + s);
        }

        if(s.equals("")) return;
        userid = s;
    }

    /**
     * Setter for the jms_password property
     */
    public void setJms_password(String s) {
        if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled()) {
            Tr.debug(tc, "setJms_password -> " + s);
        }
        if(s.equals("")) return;
        password = s;
    }

    /**
     * Setter for the jndiProperties property
     */
    public void setJndi_properties(String s) {
        if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled()) {
            Tr.debug(tc, "setJndi_properties -> " + s);
        }

        if(s==null || s.equals("")){
            return;
        }

        jndiProperties = new Properties();
        StringTokenizer tokenizer = new StringTokenizer(s, ";");
        while(tokenizer.hasMoreTokens()){
            String element = tokenizer.nextToken();
            StringTokenizer elementTokenizer = new StringTokenizer(element, "=");
            while(elementTokenizer.hasMoreTokens()){
                String key = elementTokenizer.nextToken();
                String value = null;
                if(elementTokenizer.hasMoreTokens()){
                    value = elementTokenizer.nextToken();
                }
                jndiProperties.setProperty(key, value);
            }
        }
    }

    /**
     * Setter for the invalidationModel property
     */
    public void setInvalidationModel(String s) {
        if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled()) {
            Tr.debug(tc, "setInvalidationModel -> " + s);
        }

        if(s==null){
            throw new IllegalArgumentException("The parameter is not valid.  The passed parameter is " + s);
        }

        if(s.equals(CLIENT_SERVER_MODEL_STRING)){
            invalidationModel = CLIENT_SERVER_MODEL;
        } else if(s.equals(CLIENT_AS_DUAL_ROLES_MODEL_STRING)){
            invalidationModel = CLIENT_AS_DUAL_ROLES_MODEL;
        } else if(s.equals(NONE_INVALIDATION_MODEL_STRING)){
            invalidationModel = NONE_INVALIDATION_MODEL;
        } else {
            throw new IllegalArgumentException("The parameter is not valid.  The passed parameter is " + s);
        }
    }

    /**
     * Setter for the invalidationStrategy property
     */
    public void setInvalidationStrategy(String s) {
        if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled()) {
            Tr.debug(tc, "setInvalidationStrategy -> " + s);
        }

        if(s==null){
            throw new IllegalArgumentException("The parameter is not valid.  The passed parameter is " + s);
        }

        if(s.equals(INVALIDATE_STRING)){
            invalidationStrategy = INVALIDATE;
            setMode(DistributionMode.INVALIDATE);
        } else if(s.equals(INVALIDATE_CONDITIONAL_STRING)){
            invalidationStrategy = INVALIDATE_CONDITIONAL;
            setMode(DistributionMode.INVALIDATE_CONDITIONAL);
        } else if(s.equals(PUSH_STRING)){
            invalidationStrategy = PUSH;
            setMode(DistributionMode.PUSH);
        } else if(s.equals(PUSH_CONDITIONAL_STRING)){
            invalidationStrategy = PUSH_CONDITIONAL;
            setMode(DistributionMode.PUSH_CONDITIONAL);
        } else if(s.equals(PUSH_INCLUDED_STRING)){
            invalidationStrategy = PUSH_INCLUDED;
            setMode(DistributionMode.PUSH);
            includedMode = true;
        } else if(s.equals(PUSH_INCLUDED_CONDITIONAL_STRING)){
            invalidationStrategy = PUSH_INCLUDED_CONDITIONAL;
            setMode(DistributionMode.PUSH_CONDITIONAL);
            includedMode = true;
        } else {
            throw new IllegalArgumentException("The parameter is not valid.  The passed parameter is " + s);
        }

    }

    /**
     * Setter for the replicationStrategy property
     */
    public void setReplicationStrategy(String s) {
        if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled()) {
            Tr.debug(tc, "setReplicationStrategy -> " + s);
        }

        //The ReplicationStrategy term is used in peer-2-peer replication scenario
        //and is same as InvalidationStrategy term in client invalidation scenario.
        setInvalidationStrategy(s);
    }

    /**
     * Setter for the replicationRole property
     */
    public void setReplicationRole(String s){
        if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled()) {
            Tr.debug(tc, "setReplicationRole -> " + s);
        }

        if(s==null){
            throw new IllegalArgumentException("The parameter is not valid.  The passed parameter is " + s);
        }

        if(s.equals(DUAL_ROLES_STRING)){
            replicationRole = DUAL_ROLES;
        } else if(s.equals(PUBLISHER_ROLE_STRING)){
            replicationRole = PUBLISHER_ROLE;
        } else if(s.equals(SUBSCRIBER_ROLE_STRING)){
            replicationRole = SUBSCRIBER_ROLE;
        } else {
            throw new IllegalArgumentException("The parameter is not valid.  The passed parameter is " + s);
        }
    }

    private void setReplicationRole(int role){
        if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled()) {
            Tr.debug(tc, "setReplicationRole -> " + getReplicationRoleString(role));
        }
        replicationRole = role;
    }
    /**
     * Setter for the mapsToPublish property
     */
    public void setMapsToPublish(String s) {
        if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled()) {
            Tr.debug(tc, "setMapsToPublish -> " + s);
        }

        if(s==null || s.equals("")){
            return;
        }

        StringTokenizer tokenizer = new StringTokenizer(s, ";");
        while(tokenizer.hasMoreTokens()){
            String element = tokenizer.nextToken();
            if(!element.equals("")) {
                configuredMapsToPublish.add(element);
            }
        }

    }
    /**
     * Setter for the enableOnClientObjectGrid property
     */
    public void setEnableOnClientObjectGrid(boolean s) {
        if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled()) {
            Tr.debug(tc, "setEnableOnClientObjectGrid -> " + s);
        }

        enableOnClientObjectGrid = s;
    }
    /**
     * Setter for the enableOnServerObjectGrid property
     */
    public void setEnableOnServerObjectGrid(boolean s) {
        if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled()) {
            Tr.debug(tc, "setEnableOnServerObjectGrid -> " + s);
        }

        enableOnServerObjectGrid = s;
    }

    /*
     * future configuration approach
    */
    private void setJndiPropertiesConfig(Object configObject) {
    }
    /*
     * future configuration approach
    */
    private void setJmsPropertiesConfig(Object configObject) {
    }

    String getPublishMapsString() {
        StringBuffer buffer = new StringBuffer();
        if (publishMaps != null && publishMaps.size() > 0) {
            Iterator i = this.publishMaps.iterator();
            int count = 1;
            buffer.append("PublishMaps:");
            while (i.hasNext()) {
                buffer.append(" " + (count++) + ":" + i.next() + "  ");
            }

        } else {
            buffer.append("PublishMaps is empty.");            
        }

        return buffer.toString();
    }
    
    String getStringForList(List list) {
		StringBuffer buffer = new StringBuffer();
		if (list == null) {
			buffer.append("null");
		} else {
			int size = list.size();
			if (list.size() == 0) {
				buffer.append("[]");
			} else {
				buffer.append("[");
				for (int i = 0; i < size; i++) {
					if (i == size - 1)
						buffer.append(list.get(i) + "]");
					else
						buffer.append(list.get(i) + ", ");
				}
			}
		}
		return buffer.toString();
	}

    void displayConfigurationInfo() {
    	if (ObjectGridManagerImpl.isTraceEnabled && tc.isDebugEnabled()) {
            StringBuffer buffer = new StringBuffer();
            buffer.append("\n------------------------\n");
            buffer.append("Replication Properties: \n");
            buffer.append("------------------------\n");
            buffer.append("    ObjectGridType = " + getObjectGridTypeString(objectgridType) + "\n");
            buffer.append("    invalidationModel = " + getInvalidationModelString(invalidationModel) + "\n");
            buffer.append("    invalidationStrategy = " + getInvalidationStrategyString(invalidationStrategy) + "\n");
            buffer.append("    replicationStrategy = " + getInvalidationStrategyString(invalidationStrategy) + "\n");
            buffer.append("    replicationRole = " + getReplicationRoleString(replicationRole) + "\n");
            buffer.append("    enableOnClientObjectGrid = " + enableOnClientObjectGrid + "\n");
            buffer.append("    enableOnServerObjectGrid = " + enableOnServerObjectGrid + "\n");
            buffer.append("    distributionMode = " + mode + "\n");
            buffer.append("    includedMode = " + includedMode + "\n");
            buffer.append("    ObjectGridName = " + myGrid.getName() + "\n");
            buffer.append("    myGrid.getListOfMapNames() = " + getStringForList(myGrid.getListOfMapNames()) + "\n");
            buffer.append("    mapNamesWithBackingMap     = " + getStringForList(actualCompleteMapNames) + "\n");
            buffer.append("    configuredMapsToPublish = " + configuredMapsToPublish + "\n");
            if(configuredMapsToPublish!=null && configuredMapsToPublish.size() > 0) {
                buffer.append("    validConfiguredMapsToPublish = " + validConfiguredMapsToPublish + "\n");
                buffer.append("    invalidConfiguredMapsToPublish = " + invalidConfiguredMapsToPublish + "\n");
            } 
            buffer.append("    mapNamesWithNullBackingMap = " + mapNamesWithNullBackingMap + "\n");               
            buffer.append("    pessimisticMaps = " + pessimisticMaps + "\n");
            buffer.append("    " + getPublishMapsString() + "\n");
            buffer.append("    publishAllMaps = " + publishAllMaps+ "\n");
            
            buffer.append("------------------------\n");
            buffer.append("JMS Properties: \n");
            buffer.append("------------------------\n");
            buffer.append("    topicConnectionFactoryJndiName = " + topicConnectionFactoryJndiName + "\n");
            buffer.append("    topicJndiName = " + topicJndiName + "\n");
            buffer.append("    topicName = " + topicName + "\n");
            buffer.append("    userid = " + userid + "\n");
            buffer.append("    password = " + password + "\n");

            if(jndiProperties==null){
                buffer.append("------------------------\n");
                buffer.append("jndiProperties: null\n");
                buffer.append("------------------------\n");
            } else {
                buffer.append("------------------------\n");
                buffer.append("jndiProperties: \n");
                buffer.append("------------------------\n");
                Enumeration e = jndiProperties.propertyNames();
                while(e.hasMoreElements()) {
                    String key = (String) e.nextElement();
                    Object value = jndiProperties.getProperty(key);
                    buffer.append("    " + key + " = " + value + "\n");
                }
            }
                                   
            Tr.debug(tc, "displayConfigurationInfo", buffer.toString());
    	}
    }

    /**
	 * This method is default to do nothing, but it allows child classes want to
	 * process LogSequence collection embedded on received JMS message by
	 * overriding this method.
	 * 
	 * @param logSequences the LogSequence collection embedded in received JMS message.
	 * 
     * @since 7.0.0.0 FIX1
	 */
    protected void onMessage(Collection logSequences) {
    	// default to do nothing, child classes can process Logsequences by overrideing this method.
    }

}
