// IBM Confidential OCO Source Material
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2005,2006
// The source code for this program is not published or otherwise divested
// of its trade secrets, irrespective of what has been deposited with the
// U.S. Copyright Office.
//
// $Id$
//
// Module  :  MapIndexPlugin.java
//
// Source File Description:
//
//    Need a file description.
//
// Change Activity:
//
// Reason       Version   Date         Userid    Change Description
// ------------ --------- ------------ --------- -----------------------------------------
// XD60_M2_B.2  WASX.XD   03/01/2005   newport   new code drop
// XD60_M3_B.10 WASX.XD   04/07/2005   sutter    Re-factor datagrid/datagrid.impl
// 271122       WASX.XD   04/28/2005   kjlaw     add throws to batchUpdate method.
// 258492       WASX.XD   05/06/2005   kjlaw     changes for rollback of index changes.
// 285430       WASX.XD   06/21/2005   pdykes    Pre-GM changes to remove non-supported public interfaces
// b1153        XD6.1     06/12/2006   cheng1    B25: add setEntityMetadata() method
// ------------ --------- ------------ --------- -----------------------------------------
package com.ibm.websphere.objectgrid.plugins.index;

import java.util.Properties;

import com.ibm.websphere.objectgrid.BackingMap;
import com.ibm.websphere.objectgrid.ObjectGridException;
import com.ibm.websphere.objectgrid.ObjectGridRuntimeException;
import com.ibm.websphere.objectgrid.ObjectMap;
import com.ibm.websphere.objectgrid.OutputFormat;
import com.ibm.websphere.objectgrid.TxID;
import com.ibm.websphere.objectgrid.plugins.BackingMapLifecycleListener;
import com.ibm.websphere.objectgrid.plugins.BackingMapPlugin;
import com.ibm.websphere.objectgrid.plugins.EventListener;
import com.ibm.websphere.objectgrid.plugins.LogElement;
import com.ibm.websphere.objectgrid.plugins.LogSequence;
import com.ibm.websphere.objectgrid.plugins.io.dataobject.SerializedEntry;
import com.ibm.websphere.objectgrid.plugins.io.dataobject.SerializedKey;
import com.ibm.websphere.objectgrid.plugins.io.dataobject.SerializedValue;
import com.ibm.websphere.projector.md.EntityMetadata;

/**
 * An index implementation must implement this interface to get modifications to the
 * Map so it can maintain the index as transactions are committed.  Only attributes
 * that implement the <code>Comparable</code> interface are eligible to be indexed.
 * <p>
 * A <code>MapIndexPlugin</code> implementation that also implements the {@link BackingMapLifecycleListener}
 * interface will be automatically added as an {@link EventListener} on the {@link BackingMap} when the
 * index set on the backing map.
 * <p>
 * A <code>MapIndexPlugin</code> may also implement the {@link BackingMapPlugin} interface in order
 * to receive enhanced BackingMap plug-in lifecycle method calls. The plug-in is then
 * also required to correctly implement each of the bean methods related to
 * introspection of its state (for example <code>isInitialized()</code>,
 * <code>isDestroyed()</code>, etc).
 *
 * @since WAS XD 6.0.1
 *
 * @see BackingMap#addMapEventListener(EventListener)
 * @see BackingMap#addMapIndexPlugin(MapIndexPlugin)
 * @see BackingMap#createDynamicIndex(MapIndexPlugin, DynamicIndexCallback)
 * @see MapIndex
 * @see MapRangeIndex
 */
public interface MapIndexPlugin
{	
    /**
     * A built-in index of the the local ObjectGrid instance that allows
     * iterating over all of the keys in a map.  This index does not require a
     * transaction and does not hold any locks.  The keys that are returned are
     * a snapshot and are not guaranteed to be consistent.<p>
     *
     * This index does not support client/server indexes.  Invoking this index on
     * a client will allow iterating over the keys in the client cache (if enabled) and
     * will not iterate over all keys in the grid.<p>
     * @since 7.1.0.3
     */
    public String SYSTEM_KEY_INDEX_NAME = "com.ibm.ws.objectgrid.builtin.map.KeyIndex";

    /**
     * Set properties of a MapIndexPlugin instance.  The MapIndexPlugin instance can initialize
     * itself with the passed properties.  In distributed ObjectGrid scenario, both setProperties
     * and getProperties methods are required to support dynamic indexing.
     *
     * @param properties
     *            The Properties object that can be used to initialize a MapIndexPlugin instance.
     *
     * @see #getProperties()
     * @since WAS XD 6.1
     */
    public void setProperties(Properties properties);

    /**
     * Extract properties from a MapIndexPlugin instance. The extracted properties can be used to
     * initialize another MapIndexPlugin instance to have the same internal states. In distributed
     * ObjectGrid scenario, both setProperties and getProperties methods are required to support
     * dynamic indexing.
     *
     * @see #setProperties(Properties)
     * @since WAS XD 6.1
     */
    public Properties getProperties();

    /**
     * Called by the ObjectGrid runtime to set the EntityMetadata of the BackingMap.
     * <p>
     * The EntityMetadata is required for supporting Tuple indexing.
     *
     * @param entityMetadata the EntityMetadata of the BackingMap.
     * @since WAS XD 6.1
     */
    public void setEntityMetadata(EntityMetadata entityMetadata);

    /**
     * Indicates if this index is the type of MapRangeIndex interface.
     *
     * @see MapRangeIndex
     * @since WAS XD 6.1
     */
    public boolean isRangeIndex();

    /**
     * Extracts the attribute from a specified value Object.
     * For example, if an Employee object has an attribute called employeeName that is indexed,
     * the getAttribute method can be used to extract the employeeName attribute value from
     * the specified Employee object.  This method is required in distributed ObjectGrid
     * environment.
     *
     * @param value
     *            The value Object.
     *
     * @return attribute from the value Object, which may be a null reference.
     * @since WAS XD 6.1
     */
    public Object getAttribute(Object value);

    /**
     * Gets the name of the indexed attribute.
     * @since WAS XD 6.1
     */
    public String getAttributeName();


    /**
     * Sets the name of the attribute to be indexed.
     * <p>
     * If the object has an attribute called employeeName or EmployeeName, the
     * index will call the "getEmployeeName" method. The attribute name must be
     * the same as that in the get method and the attribute must implement the
     * <code>Comparable</code> interface. If the attribute type is boolean,
     * the "isAttributeName" method pattern is supported too.
     *
     * @param attributeName the name of the attribute to set.
     */
    public void setAttributeName(String attributeName);

    /**
     * Gets the name of the index.
     *
     * @return the name of the index.
     *
     * @see ObjectMap#getIndex(String)
     */
    String getName();

    /**
     * Gets an index proxy object for performing index lookup operations.
     * <p>
     * The caller must cast the object returned to either a <code>MapIndex</code> or
     * <code>MapRangeIndex</code> to perform the lookup operations.
     *
     * @param map the MapIndexInfo object required for maintaining the index.
     * .
     * @return a proxy to either an object that implements MapIndex or MapRangeIndex.
     *
     * @see MapIndex
     * @see MapRangeIndex
     * @see ObjectMap#getIndex(String)
     */
    Object getIndexProxy(MapIndexInfo map);

    /**
     * Called by ObjectGrid runtime to allow the index to be updated as the
     * result of changes applied to a map during the commit cycle of a
     * transaction.
     * <p>
     * Use  the <code>LogElement.getType()</code> method to determine what
     * operation is required for updating the index. Use the
     * <code>LogElement.getBeforeImage()</code> method to get the value object
     * that existed prior to the committing transaction applying a change to the
     * map.  Use the <code>LogElement.getAfterImage()</code> method to get the
     * value object after the committing transaction applied the change to the
     * map entry.
     * <p>
     * For a MapIndexPlugin in an {@link ObjectMap} that is configured to use {@link OutputFormat#RAW}
     * for the keys or values, the keys and values objects in the LogSequence
     * will be {@link SerializedKey} or {@link SerializedValue} objects respectively. If required, you
     * can use the {@link SerializedEntry#getObject()} method to retrieve (possibly inflating
     * the serialized object) the original key or value object.
     * <p>
     * To override the map's output format configuration, use the {@link PluginOutputFormat} 
     * annotation in the implementation class.
     * <p>
     * Note, the <code>undoBatchUpdate(TxID, LogSequence)</code> method may be
     * called later to undo these changes if an exception occurs that causes
     * the committing transaction to be rolled back instead.
     *
     * @param txid the transaction context (TxID) for the changes.
     * @param sequence the log sequence that contains the changes from transaction.
     *
     * @throws ObjectGridRuntimeException is a failure occurs that requires transaction
     *         to be rolled back.
     *
     * @see #undoBatchUpdate(TxID, LogSequence)
     * @see LogElement#getType()
     * @see LogElement#getBeforeImage()
     * @see LogElement#getAfterImage()
     * @see LogSequence
     * @see TxID
     */
    void doBatchUpdate(TxID txid, LogSequence sequence) throws ObjectGridRuntimeException;

    /**
     * Called by the ObjectGrid runtime to undo any changes made to the index as
     * a result of a prior call to the <code>doBatchUpdate(TxID, LogSequence)</code>
     * method.
     * <p>
     * This method is called when an exception or error condition that requires
     * all changes made by a transaction to be rolled back.  For this reason,
     * the implementation of this method should catch all Throwables and
     * continue with the next <code>LogElement</code> in the
     * <code>LogSequence</code> until all <code>LogElement</code>s are processed
     * so that as many changes to the index are undone as possible.  An
     * ObjectGridException should only be thrown after processing the entire
     * <code>LogSequence</code> and this method was unable to successfully undo
     * one or more changes in the <code>LogSequence</code>.
     * <p>
     * Use the <code>LogElement.getUndoType()</code> method to determine what
     * operation is required to undo any change made to the index.  Use the
     * <code>LogElement.getBeforeImage()</code> to get the value object that
     * existed prior to the committing transaction applying a change to the map.
     * Use the <code>LogElement.getAfterImage()</code> to get the value
     * object after the committing transaction applied the change to the map entry.
     * <p>
     * For a MapIndexPlugin in an {@link ObjectMap} that is configured to use {@link OutputFormat#RAW}
     * for the keys or values, the keys and values objects in the LogSequence
     * will be {@link SerializedKey} or {@link SerializedValue} objects respectively. If required, you
     * can use the {@link SerializedEntry#getObject()} method to retrieve (possibly inflating
     * the serialized object) the original key or value object.
     * <p>
     * To override the map's output format configuration, use the {@link PluginOutputFormat} 
     * annotation in the implementation class.
     * 
     * @param txid the transaction context (TxID) for the changes.
     * @param sequence the log sequence that contains changes from transaction.
     *
     * @throws ObjectGridException is an error occurred during processing
     *
     * @see LogElement#getUndoType()
     * @see LogElement#getBeforeImage()
     * @see LogElement#getAfterImage()
     * @see LogSequence
     * @see TxID
     */
    void undoBatchUpdate( TxID txid, LogSequence sequence) throws ObjectGridException;
}
