// IBM Confidential OCO Source Material
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2007, 2009
// The source code for this program is not published or otherwise divested
// of its trade secrets, irrespective of what has been deposited with the
// U.S. Copyright Office.
//
// $Id$
//
// Module  :  AgentPermission.java
//
// Source File Description:
//
// Need a file description.
//
// Change Activity:
//
// Reason       Version   Date         Userid    Change Description
// ------------ --------- ------------ --------- -----------------------------------------
// b4719        XD 6.1    02/06/2007   jian.tang Support agent manager authorizations
// b14971       6.1.0.5   02/09/2009   jian.tang Make AgentPermission accept OG name or map name with "." in it
// ------------ --------- ------------ --------- -----------------------------------------
package com.ibm.websphere.objectgrid.security;

import java.security.Permission;
import java.util.StringTokenizer;

import com.ibm.ejs.ras.Tr;
import com.ibm.ejs.ras.TraceComponent;
import com.ibm.ws.objectgrid.Constants;
import com.ibm.ws.objectgrid.security.PermissionHelper;
import com.ibm.ws.objectgrid.util.ObjectGridUtil;

/**
 * This class represents permissions to the datagrid agents.
 * <p>
 * The name of the agent permission is the full map name. A full Map name is
 * a concatenation of the ObjectGrid name, a period, and then the Map name. For example,
 * if the object grid name is "myobjectgrid" and the map name is "mymap", then the full map
 * name used in the permission is "myobjectgrid.mymap".
 * <p>
 * The action of the permission is a "," delimited string of agent implementation class names or
 * package names. For example, if your agent implementation class name is "com.acme.agent.EntryAgentImpl", then
 * the agent with implementation class com.acme.agent.EntryAgentImpl is authorized. If the permission
 * action is "com.acme.agent.*", then all agents with implemention class belonging to the
 * com.acme.agent package are authorized. You can use "," to separate classes or packages in the
 * action string, for example, "com.acme.agent.EntryAgentImpl,com.acme.agent2.*".
 *
 * <p>
 * Wildcards can be used in map name with some restrictions. A wild card "*" can be used
 * to replace the map name or the object grid name, but not partially. For example,
 * "myObjectGrid.*", "*.myMap", and "*.*" are valid names, but "myObject*.*" is not valid.
 *
 * @since WAS XD 6.1
 *
 * @see Permission
 */
public final class AgentPermission extends Permission {
    private static final long serialVersionUID = 1L;

    private final static String CLASS_NAME = AgentPermission.class.getName();

    private final static TraceComponent TC = Tr.register(CLASS_NAME, Constants.TR_SECURITY_GROUP_NAME,
            Constants.TR_RESOURCE_BUNDLE_NAME);

    private final static boolean DEBUG = false;

    /**
     * parsed names
     * */
    private String[] names = new String[2];

    /**
     * Whether the name has been parsed into ObjectGrid and map name yet.
     */
    private boolean isNameParsed = false;

    private String[] classes = null;

    private String classesString = null;

    /**
     * Constructs a new AgentPermission for the named map with the specified
     * actions.
     * <p>
     * The map name for this permission should be in the format of
     * &lt;ObjectGrid name&gt;.&lt;map name&gt;, for example "og1.map1".  A special value
     * of "*" is used to represent all objectgrids or all maps, for example
     * "*.map1", "og1.*", or "*.*". A "*" is used to substitue a whole
     * ObjectGrid name or map name. A partial substitution like "og*.map" is not
     * supported.
     *
     * @param name the full name of the map. It is also the permission name.
     * @param actions
     *            the comma separated list of agent classes or packages, such as
     *            "com.acme.EntryAgentImpl,com.acme.ReduceAgentImpl,com.acme.actions.*".
     *
     * @throws IllegalArgumentException if name is <code>null</code> or malformed
     *         or if actions is <code>null</code>.
     */
    public AgentPermission(String name, String actions) {
        super(name == null ? "" : name.trim());

        if (DEBUG && TC.isEntryEnabled())
            Tr.entry(TC, "<init>", new Object[] { name, actions });

        if (name == null) {
            throw new IllegalArgumentException("Name cannot be null.");
        }
        name = name.trim();

        if (actions == null) {
            throw new IllegalArgumentException("Actions cannot be null.");
        }
        parseAction(classesString = actions);
        isNameParsed = PermissionHelper.parseName(getName(), names);

        if (DEBUG && TC.isEntryEnabled())
            Tr.exit(TC, "<init>", this);
    }

    /**
     * Constructs a new AgentPermission for the named map with the specified
     * actions.
     * <p>
     * This method is used for the ObjectGrid map of which the ObjectGrid name
     * or map name contains a ".". In this case, we pass the ObjectGrid name and
     * map name separately as the parameters.
     * <p>
     * ObjectGrid uses &lt;ObjectGrid name&gt;.&lt;map name&gt; as the permission name. Therefore,
     * your application still needs to make sure there is no name conflicts for the
     * &lt;ObjectGrid name&gt;.&lt;map name&gt;. For example, you should not have an ObjectGrid
     * with name "og" and one of its maps with name "session.data", and another
     * ObjectGrid with name "og.session" and one of its maps with name "data".
     *
     * @param ogName the ObjectGrid name.
     * @param mapName the map name.
     * @param actions
     *            the comma separated list of agent classes or packages, such as
     *            "com.acme.EntryAgentImpl,com.acme.ReduceAgentImpl,com.acme.actions.*".
     *
     * @throws IllegalArgumentException if ObjectGrid name or map name is <code>null</code>
     *         or if actions is <code>null</code> or contains an invalidate
     *         action name
     * @since 6.1.0.5 FIX2
     */
    public AgentPermission(String ogName, String mapName, String actions) {
        super((ogName == null ? "" : ogName.trim()) + "." + (mapName == null ? "" : mapName.trim()));

        if (DEBUG && TC.isEntryEnabled())
            Tr.entry(TC, "<init>", new Object[] { ogName, mapName, actions });

        if (ogName == null) {
            throw new IllegalArgumentException("ObjectGrid name cannot be null.");
        }
        if (mapName == null) {
            throw new IllegalArgumentException("ObjectGrid Map name cannot be null.");
        }
        if (actions == null) {
            throw new IllegalArgumentException("Actions cannot be null.");
        }
        parseAction(actions);
        names[0] = ogName;
        names[1] = mapName;
        isNameParsed = true;

        if (DEBUG && TC.isEntryEnabled())
            Tr.exit(TC, "<init>", this);
    }


    /**
     * Parses the action string
     *
     * @param action
     *            the action string
     */
    private void parseAction(String action) {
        StringTokenizer st = new StringTokenizer(action, ",");

        classes = new String[st.countTokens()];
        int index = 0;
        while (st.hasMoreTokens()) {
            classes[index++] = st.nextToken().trim();
        }
    }

    /**
     * Checks if the specified permission is "implied" by
     * this object.
     * <P>
     * More specifically, this method returns true if:<p>
     * <ul>
     * <li> <i>permission</i>'s class is the same as this object's class, and<p>
     * <li> <i>permission</i>'s name is equal to or (in the case of wildcards)
     *      is implied by this object's
     *      name. For example, "a.*" implies "a.b"., and<p>
     * <li> <i>permission</i>'s actions are equal to or are a subset of
     *         this object's actions
     * </ul>
     *
     * @param permission the permission to check against.
     *
     * @return true if the passed permission is equal to or
     * implied by this permission, false otherwise.
     */
    public boolean implies(Permission permission) {
        if (!(permission.getClass().getName().equals(CLASS_NAME))) {
            return false;
        }

        AgentPermission p = (AgentPermission) permission;

        if (!getName().equals(p.getName())) {

            if (isNameParsed) {
                if (p.isNameParsed) {
                    // this  has og name and map name
                    // other has og name and map name
                    String[] otherNames = p.getParsedNames();
                    for (int i = 1; i >= 0; i--) {
                        if (!names[i].equals("*") && !names[i].equals(otherNames[i])) {
                            return false;
                        }
                    }
                } else {
                    // this  has og name and map name
                    // other doesn't have og name and map name, only has full name.

                    if (names[0].equals("*")) {
                        if (!names[1].equals("*") && !p.getName().endsWith("."+names[1])){
                            return false;
                        }
                    } else {
                        if (names[1].equals("*")) {
                            if (!p.getName().startsWith(names[0] + ".")) {
                                return false;
                            }
                        } else {
                            // since their full name is not equal, must be false.
                            return false;
                        }
                    }
                }
            } else {
                // this doesn't have og name and map name, only has full name, so it doesn't contain *
                // so implies only return true if full name are equal.
                // However, they are not equal, so it is always false.
                return false;
            }
        }

        for (int i = 0; i < p.classes.length; i++) {

            boolean found = false;
            for (int j = 0; j < classes.length; j++) {

                if (isPermissionImplied(classes[j], p.classes[i])) {
                    found = true;
                    break;
                }
            }

            if (!found) {
                return false;
            }
        }

        return true;
    }

    /**
     * What is considered str1 implies str2
     * 1. str1.equals(str2)
     * 2. str1 is a package (end with .*), and str2.startsWith(str1-(.*));
     * 3. str1 and str2 are all packages, and str2-(.*).startswith(st1-(.*));
     */
    private boolean isPermissionImplied(String str1, String str2) {
        if (str2.length() == 0 || str1.equals(str2)) {
            return true;
        }

        if (str2.endsWith(".*") && !str1.endsWith(".*")) {
            return false;
        }

        if (str1.endsWith(".*")) {
            String package1 = str1.substring(0, str1.length()-1);
            String classOrPackage2 = str2.endsWith(".*")? str2.substring(0, str2.length()-1): str2;
            return classOrPackage2.startsWith(package1);
        }

        return false;
    }

    /**
     * Checks two AgentPermission objects for equality. Two AgementPermission
     * objects are equal if and only if their names and actions are all equal.
     * <P>
     * Do not use the <code>equals</code> method for making access control
     * decisions; use the <code>implies</code> method.
     *
     * @param o the object we are testing for equality with this object.
     *
     * @return <code>true</code> if both AgentPermission objects are equivalent.
     */
    public boolean equals(Object o) {
        if (!(o.getClass().getName().equals(CLASS_NAME)))
            return false;

        // check the name and action mask for deep equality
        AgentPermission p = (AgentPermission) o;
        return ((p.getName().trim().equals(getName().trim())) && ObjectGridUtil.arrayEqualsIgnoreOrder(classes, p.classes));
    }

    /**
     * Returns the hash code value for this AgentPermission object.
     *
     * @return a hash code value for this object.
     */
    public int hashCode() {
        // a hash code for permissions,
        return getName().hashCode() + classes.hashCode();
    }

    /**
     * Returns the actions as a String.
     *
     * @return the actions of this Permission.
     */
    public String getActions() {
        return classesString;
    }

    /**
     * Returns an array which contains the ObjectGrid name and map name.
     *
     * @return the parsed name array
     */
    public String[] getParsedNames() {
        return names;
    }

}