// IBM Confidential OCO Source Material
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2006
// The source code for this program is not published or otherwise divested
// of its trade secrets, irrespective of what has been deposited with the
// U.S. Copyright Office.
//
// $Id$
//
// Module  :  ObjectGridPermission.java
//
// Source File Description:
//
// Need a file description.
//
// Change Activity:
//
// Reason       Version   Date         Userid    Change Description
// ------------ --------- ------------ --------- -----------------------------------------
// b4719        WASX.XD   12/21/2006   jian.tang Add security to ObjectGrid
// ------------ --------- ------------ --------- -----------------------------------------
package com.ibm.websphere.objectgrid.security;

import java.security.Permission;
import java.util.StringTokenizer;

import com.ibm.websphere.objectgrid.ObjectGrid;

/**
 * This class represents permissions to the ObjectGrid. It has two different
 * actions:
 * </p>
 * <ul>
 * <li>QUERY: create an object query or entity query</li>
 * <li>DYNAMIC_MAP: create a remote dynamic map</li>
 * </ul>
 *
 * @since WAS XD 6.1
 *
 * @see Permission
 * @see ObjectGrid
 */

// TODO Quit using BIT MASKS!  This will only scales to 32 permissions.  This is fine in reality, but is
// the minimized overhead that much more beneficial than a more scalable structure.


public final class ObjectGridPermission extends Permission {
    private static final long serialVersionUID = 1L;

    private final static String CLASS_NAME = ObjectGridPermission.class.getName();

//    private final static TraceComponent TC = Tr.register(CLASS_NAME, Constants.TR_GROUP_NAME,
//            Constants.TR_RESOURCE_BUNDLE_NAME);

    /**
     * mask indicating the permissions
     */
    private int mask = 0;

    /**
     * Query permission constant
     */
    public static final int QUERY = 0x01;

    // make private to avoid 0x02 being used again by a new permission that could happen if it was deleted. 
    /**
     * Stream query permission constant
     */
    private static final int STREAM_QUERY = 0x02;

    /**
     * Dynamic map permission constant
     */
    public static final int DYNAMIC_MAP = 0x04;

    /**
     * An array with the string representation of each permission.
     * <p>
     * The indexes to the String representation are the <code>int</code>
     * constants for each permission (i.e. QUERY, DYNAMIC_MAP)
     */
    public static final String[] PERMISSIONS = new String[DYNAMIC_MAP + 1];

    /**
     * All permissions String
     */
    public static final String ALL_PERMISSION = "all";

    /**
     * All permission constant
     */
    public static final int ALL = QUERY | DYNAMIC_MAP;

    /**
     * The total number of possible Permissions
     */
    public static final int NUMBER_PERMISSION = 2;

    /**
     * parsed names
     * */
    private String name = null;

    static {
        PERMISSIONS[QUERY] = "query";
        PERMISSIONS[DYNAMIC_MAP] = "dynamicmap";
    }

    /**
     * Constructs a new ObjectPermission for the named ObjectGrid with the specified
     * actions.
     * <p>
     * The ObjectGrid name for this permission should be in the format of
     * <ObjectGrid name>, for example "og1".  A special value
     * of "*" is used to represent all objectgrids, for example
     * "*".
     *
     * @param name the ObjectGrid name.
     * @param actions
     *            the comma separated list of actions, such as "query", and "dynamicmap".
     *
     * @throws IllegalArgumentException if name is <code>null</code> or malformed
     *         or if actions is <code>null</code> or contains an invalidate
     *         action name
     */
    public ObjectGridPermission(String name, String actions) {
        super(name);

//        if (TC.isEntryEnabled())
//            Tr.entry(TC, "<init>", new Object[] { name, actions });

        if (name == null) {
            throw new IllegalArgumentException("ObjectGrid name cannot be null.");
        }
        if (actions == null) {
            throw new IllegalArgumentException("Actions cannot be null.");
        }
        parseAction(actions);
        this.name = name;

//        if (TC.isEntryEnabled())
//            Tr.exit(TC, "<init>", this);
    }

    /**
     * Constructs a new ObjectPermission for the named ObjectGrid with the specified
     * actions.
     * <p>
     * The ObjectGrid name for this permission should be in the format of
     * <ObjectGrid name>, for example "og1".  A special value
     * of "*" is used to represent all objectgrids, for example
     * "*".
     *
     * @param name the ObjectGrid name
     * @param actions
     *            the actions mask. For example,
     *            ObjectGridPermission.QUERY|ObjectGridPermission.DYNAMIC_MAP
     *
     * @throws IllegalArgumentException if name is <code>null</code> or malformed
     *         or if actions is less than 0 or greater than <code>ALL</code>
     */
    public ObjectGridPermission(String name, int actions) {
        super(name);
        //        if (TC.isEntryEnabled())
        //            Tr.entry(TC, "<init>", new Object[]{name, new Integer(actions)});

        if (name == null) {
            throw new IllegalArgumentException("Name cannot be null.");
        }
        parseIntAction(actions);
        this.name = name;

        //        if (TC.isEntryEnabled())
        //            Tr.exit(TC, "<init>", this);
    }

    /**
     * Parses the action string
     *
     * @param action
     *            the action string
     */
    private void parseAction(String action) {
        StringTokenizer st = new StringTokenizer(action, ",");

        mask = 0;
        while (st.hasMoreTokens()) {
            String tok = st.nextToken().trim();
            if (tok.equals(PERMISSIONS[QUERY])) {
                mask |= QUERY;
            } else if (tok.equals(PERMISSIONS[DYNAMIC_MAP])) {
                mask |= DYNAMIC_MAP;
            } else if (tok.equals(ALL_PERMISSION)) {
                mask |= ALL;
            } else if (!tok.equals("")) {
                throw new IllegalArgumentException("Unknown action " + tok);
            }
        }
    }

    /**
     * Parses the action integer
     *
     * @param action
     *            the action integer
     */
    private void parseIntAction(int action) {
        if (action > ALL || action < 0) {
            throw new IllegalArgumentException("Action mask value should be between 0 and ObjectGridPermission.ALL.");
        }

        if (action == 0) {
            return;
        } else if (action == ALL) {
            mask |= ALL;
            return;
        } else {
            if ((action & QUERY) == QUERY) {
                mask |= QUERY;
            }
            if ((action & DYNAMIC_MAP) == DYNAMIC_MAP) {
                mask |= DYNAMIC_MAP;
            }
        }
    }

    /**
     * Checks if the specified permission is "implied" by this object.
     * <P>
     * More specifically, this method returns true if:<p>
     * <ul>
     * <li> <i>permission</i>'s class is the same as this object's class, and<p>
     * <li> <i>permission</i>'s name is equal to or (in the case of wildcards)
     *      is implied by this object's
     *      name. For example, "*" implies "a", and<p>
     * <li> <i>permission</i>'s actions are equal to or are a subset of
     *         this object's actions
     * </ul>
     *
     * @param permission the permission to check against.
     *
     * @return true if the passed permission is equal to or
     * implied by this permission, false otherwise.
     */
    public boolean implies(Permission permission) {
//        if (TC.isEntryEnabled())
//            Tr.entry(TC, "implies", new Object[] { this, permission });
        if (!(permission.getClass().getName().equals(CLASS_NAME))) {
//            if (TC.isEntryEnabled())
//                Tr.exit(TC, "implies", "Not a ObjectGridPermission object, return false");
            return false;
        }

        ObjectGridPermission p = (ObjectGridPermission) permission;

        // The name part must be either the wildcard *, or the name must match
        // our name

        if (!getName().trim().equals(p.getName().trim())) {
            if (!name.equals("*")) {
//                if (TC.isEntryEnabled())
//                    Tr.exit(TC, "implies", "Name not match, return false");
                return false;
            }
        }

        if ((mask & p.mask) != p.mask) {
//            if (TC.isEntryEnabled())
//                Tr.exit(TC, "implies", "permission mask not match, return false");
            return false;
        }

        // Only if both the action and name match do we return true.
//        if (TC.isEntryEnabled())
//            Tr.exit(TC, "implies", "true");
        return true;
    }

    /**
     * Checks two ObjectGridPermission objects for equality. Two ObjectGridPermission
     * objects are equal if and only if their names and actions are equal.
     * <P>
     * Do not use the <code>equals</code> method for making access control
     * decisions; use the <code>implies</code> method.
     *
     * @param o the object we are testing for equality with this object.
     *
     * @return <code>true</code> if both ObjectGridPermission objects are equivalent.
     */
    public boolean equals(Object o) {
        if (!(o.getClass().getName().equals(CLASS_NAME)))
            return false;

        // check the name and action mask for deep equality
        ObjectGridPermission p = (ObjectGridPermission) o;
        return ((p.getName().trim().equals(getName().trim())) && (p.mask == mask));
    }

    /**
     * Returns the hash code value for this ObjectGridPermission object.
     *
     * @return a hash code value for this object.
     */
    public int hashCode() {
        // a hash code for permissions,
        return getName().hashCode() ^ mask;
    }

    /**
     * Returns the actions as a String.
     *
     * @return the actions of this Permission.
     */
    public String getActions() {
        return getActions(mask);
    }

    /**
     * Returns the actions as a String for the specified mask.
     *
     * @param m the mask
     *
     * @return the action string representation for the mask
     */
    public static String getActions(int m) {
        if (m > ALL || m < 0) {
            throw new IllegalArgumentException("Action mask should be between 0 and ObjectGridPermission.ALL.");
        }
        if (m == 0) {
            return "";
        }

        int index = 1;
        boolean isCommaNeeded = false;
        StringBuffer sb = new StringBuffer(32);

        while (index <= DYNAMIC_MAP) {
            if ((index & m) == index) {
                if (isCommaNeeded) {
                    sb.append(", ");
                } else {
                    isCommaNeeded = true;
                }
                sb.append(PERMISSIONS[index]);
            }
            index *= 2;
            if (index == STREAM_QUERY) {
            	index *= 2;
            }
        }

        return new String(sb);
    }

    /**
     * Returns an int mask value which represents the permission's actions.
     * <p>
     * For example, for a ObjectGridPermission which has "query" and "dynamicmap" permissions,
     * the returned value will be 3 (ObjectGridPermission.QUERY|ObjectGridPermission.DYNAMIC_MAP).
     *
     * @return an int value which represents the permission.
     */
    public int getActionsInInt() {
        return mask;
    }
}