// IBM Confidential OCO Source Material
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2005,2006
// The source code for this program is not published or otherwise divested
// of its trade secrets, irrespective of what has been deposited with the
// U.S. Copyright Office.
//
// 1.3 XD/ws/code/objectgrid.client/src/com/ibm/websphere/objectgrid/security/plugins/builtins/WSTokenCredentialGenerator.java, WAS.objectgrid, WASX.XD, b0614.107 2/14/06 18:22:10
//
// Module  :  WSTokenCredentialGenerator.java
//
// Source File Description: 
//
//  Need a file description.
//
// Change Activity:
//
// Reason       Version   Date         Userid    Change Description
// ------------ --------- ------------ --------- -----------------------------------------
// XD601_M3_Bx  WASX.XD   11/15/2005   jitang    ObjectGrid security support
// d335512      WASX.XD   01/04/2006   jitang    push both AT and OAT in the WS credential generator
// PK17543      WASX.XD   01/11/2006   jitang    Update JavaDoc
// d344266      WXD601.XD 02/08/2006   jitang    Add setProperties method
// ------------ --------- ------------ --------- -----------------------------------------
package com.ibm.websphere.objectgrid.security.plugins.builtins;

import java.security.AccessController;
import java.security.PrivilegedActionException;
import java.security.PrivilegedExceptionAction;

import javax.security.auth.Subject;

import com.ibm.ejs.ras.Tr;
import com.ibm.ejs.ras.TraceComponent;
import com.ibm.websphere.objectgrid.security.plugins.CannotGenerateCredentialException;
import com.ibm.websphere.objectgrid.security.plugins.Credential;
import com.ibm.websphere.objectgrid.security.plugins.CredentialGenerator;
import com.ibm.websphere.security.WSSecurityException;
import com.ibm.websphere.security.auth.WSLoginFailedException;
import com.ibm.websphere.security.auth.WSSecurityContext;
import com.ibm.websphere.security.auth.WSSubject;
import com.ibm.ws.ffdc.FFDCFilter;
import com.ibm.ws.management.util.SecurityHelper;
import com.ibm.ws.objectgrid.Constants;
import com.ibm.ws.objectgrid.ObjectGridManagerImpl;
import com.ibm.wsspi.security.token.WSOpaqueTokenHelper;

/**
 * <p>This class represents a credential generator when running in 
 * WebSphere Application Server. 
 * 
 * <p>When the <code>getCredential()</code> method is called, the Subject 
 * associated with the current thread is retrieved. The security information in 
 * this Subject object is converted into a <code>WSTokenCredential</code>. This 
 * credential object has enough information for the receiving side to rebuild 
 * the security context.
 *   
 * <p>This scenario takes advantage of the fact that the ObjectGrid client has 
 * already been authenticated. Since application servers housing ObjectGrid 
 * servers are in the same security domain as the application servers housing 
 * the ObjectGrid clients, the security tokens can be propagated from the 
 * ObjectGrid client to the ObjectGrid server so there is no need to 
 * re-authenticate to the same user registry.
 *
 * <p>Users can specify whether to retrieve a runAs subject or a caller subject 
 * from the thread by using the <code>RUN_AS_SUBJECT</code> or 
 * <code>CALLER_SUBJECT</code> constant.
 * 
 * <p>WSTokenCredentialGenerator has a one to many relationship with WSTokenCredential 
 * because it can generate different WSTokenCredential objects based on what Subject is 
 * associated with the current thread.
 * 
 * @since WAS XD 6.0.1, XC10
 * 
 * @see WSTokenCredential
 */
public class WSTokenCredentialGenerator implements CredentialGenerator {

    private static final TraceComponent TC = Tr.register(
            WSTokenCredentialGenerator.class,
            Constants.TR_SECURITY_GROUP_NAME,
            Constants.TR_RESOURCE_BUNDLE_NAME);

    /** 
     * A constant representing the runAs Subject type 
     */
    public static final int RUN_AS_SUBJECT = 1;

    /**
     * A constant representing the caller Subject type 
     */
    public static final int CALLER_SUBJECT = 2;

    /** 
     * A constant representing the runAs Subject type 
     */
    public static final String RUN_AS_SUBJECT_STRING = "runAs";

    /** 
     * A constant representing the caller Subject type 
     */
    public static final String CALLER_SUBJECT_STRING = "caller";

    private int type = RUN_AS_SUBJECT;

    /**
     * Creates a new WSTokenCredentialGenerator with a default
     * runAs subject type.
     * 
     * @see #RUN_AS_SUBJECT
     * @see #setProperties(String)
     * @see #setType(int) 
     */
    public WSTokenCredentialGenerator() {
        super();
    }

    /**
     * Creates a new WSTokenCredentialGenerator with the specified subject type. 
     * <p>
     * A valid subject type is either <code>RUN_AS_SUBJECT</code> or 
     * <code>CALLER_SUBJECT</code>.
     * 
     * @param aType either <code>RUN_AS_SUBJECT</code> or 
     *        <code>CALLER_SUBJECT</code>
     * 
     * @throws IllegalArgumentException if the specified type is invalid
     * 
     * @see #CALLER_SUBJECT
     * @see #RUN_AS_SUBJECT
     */
    public WSTokenCredentialGenerator(int aType) {
        if (aType == RUN_AS_SUBJECT || aType == CALLER_SUBJECT) {
            type = aType;
        }
        else {
            throw new IllegalArgumentException("Unsupported type : " + aType);
        }
    }

    /**
     * Creates a new <code>WSTokenCredential</code> object using this
     * object's using the security information from the Subject 
     * associated with the current thread.
     * 
     * @return a new <code>WSTokenCredential</code> instance
     * 
     * @throws CannotGenerateCredentialException if an error occurs during
     *         retrieval of the Subject's security information, see the cause
     *         by exception for more information
     * 
     * @see CredentialGenerator#getCredential()
     * @see WSTokenCredential
     */
    public Credential getCredential() throws CannotGenerateCredentialException {

        Credential cred = null;

        try {
            // get the Subject from the thread
            final int finalType = type;
            cred = (Credential) AccessController.doPrivileged(new PrivilegedExceptionAction() {
                public Object run() throws WSSecurityException, WSLoginFailedException {
                    byte[] token = null;
                    byte[] OAT = null;
                    Subject subject = null;

                    if (finalType == RUN_AS_SUBJECT) {
                        subject = WSSubject.getRunAsSubject();
                    }
                    else if (finalType == CALLER_SUBJECT) {
                        subject = WSSubject.getCallerSubject();
                    }

                    WSOpaqueTokenHelper helper = WSOpaqueTokenHelper.getInstance();

                    // call the helper method discussed in 2615 to collect all information and put into a single opaque authorization token (OAT)
                    OAT = helper.createOpaqueTokenFromSubject(subject);

                    WSSecurityContext ctx = SecurityHelper.getHelper().getWSSecurityContext();
                    String realm = SecurityHelper.getHelper().getRealm();
                    String targetServerName = null;

                    if (ctx != null) {
                        // create the authentication mechanism specific authorization token (AT) in mechanism neutral way
                        token = ctx.initSecContext(subject, targetServerName, realm);
                    }

                    return new WSTokenCredential(token, OAT);
                }
            });

        }
        catch (PrivilegedActionException pae) {
            FFDCFilter.processException(
                    pae,
                    "com.ibm.websphere.objectgrid.security.plugins.builtins.WSTokenCredentialGenerator.getCredential",
                    "74",
                    this);

            if (ObjectGridManagerImpl.isTraceEnabled && TC.isDebugEnabled())
                Tr.debug(TC, "PrivilegedActionException: " + pae);

            Exception rootE = pae.getException();

            throw new CannotGenerateCredentialException(rootE);
        }

        return cred;
    }

    /**
     * Gets the subject type.
     * 
     * @return   the argument that was passed to the constructor or the 
     *           <code>setType(int)</code> method, the value from the
     *           argument passed to the <code>setProperties</code> 
     *           method, or the default value of <code>RUN_AS_SUBJECT</code>
     *           if the default constructor is used
     * 
     * @see #RUN_AS_SUBJECT
     * @see #setProperties(String)
     * @see #setType(int) 
     */
    public int getType() {
        return type;
    }

    /**
     * Sets the subject type, either <code>RUN_AS_SUBJECT</code> or 
     * <code>CALLER_SUBJECT</code>.
     * 
     * @param aType the subject type
     * 
     * @throws IllegalArgumentException if the specified type is invalid
     */
    public void setType(int aType) {
        if (aType == RUN_AS_SUBJECT || aType == CALLER_SUBJECT) {
            type = aType;
        }
        else {
            throw new IllegalArgumentException("Unsupported type : " + aType);
        }
    }

    /**
     * Sets additional properties namely the subject type. 
     * 
     * @param properties the property should be either 
     * <code>RUN_AS_SUBJECT_STRING</code> or <code>CALLER_SUBJECT_STRING</code>
     * 
     * @throws IllegalArgumentException if properties is not one of the expected
     *         values
     * 
     * @see #CALLER_SUBJECT_STRING
     * @see #RUN_AS_SUBJECT_STRING
     */
    public void setProperties(String properties) {
        properties = properties.trim();

        if (properties.equals(RUN_AS_SUBJECT_STRING)){
            type = RUN_AS_SUBJECT;
        }
        else if (properties.equals(CALLER_SUBJECT_STRING)){
            type = CALLER_SUBJECT;
        }
        else {
            throw new IllegalArgumentException(
            "The value of the properties should be either "+RUN_AS_SUBJECT_STRING+" or "+CALLER_SUBJECT_STRING);
        }
    }
}