/*
 * IBM Confidential OCO Source Material
 * 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2007, 2010
 * The source code for this program is not published or otherwise divested
 * of its trade secrets, irrespective of what has been deposited with the
 * U.S. Copyright Office.
 *
 * $HeadURL$
 * $Id$
 *
 * Change Activity:
 *
 * Reason       Version   Date         Userid    Change Description
 * ------------ --------- ------------ --------- ---------------------------------------
 * bug 100037   6.1 ifix3 10/10/2007   jhanders  Use context classloader to load the class
 * b12020       6.1.0.3   05/07/2008   jhanders  Use a different thread pool than the IBM ORB default
 * b12271       7.0       04/09/2009   schmittm  Move reference to Messages class, add translated message
 * b16176       7.0       06/02/2009   jian.tang Clean up the property file usage
 * b19358       7.1       04/19/2010   jhanders  Update so that serverStarting trace comes out in trace.log
 * b21775       7.1FIX1   11/14/2010   jhanders  Update shutdown hook to not call an API that starts another thread.
 * bug 22089    7.1.x     12/09/2010   jian.tang RBC: Order shutdown hook
*/

package com.ibm.websphere.objectgrid.server;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.lang.reflect.Constructor;
import java.net.URL;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.util.ArrayList;
import java.util.Enumeration;
import java.util.List;

import com.ibm.ejs.ras.Tr;
import com.ibm.ejs.ras.TraceComponent;
import com.ibm.websphere.objectgrid.ObjectGridRuntimeException;
import com.ibm.ws.ffdc.FFDCFilter;
import com.ibm.ws.objectgrid.Constants;
import com.ibm.ws.objectgrid.WXSProperties;
import com.ibm.ws.objectgrid.plugins.ServerLifecycleListener;
import com.ibm.ws.objectgrid.runtime.RuntimeInfo;
import com.ibm.ws.objectgrid.server.CatalogServerPropertiesImpl;
import com.ibm.ws.objectgrid.server.ServerPropertiesImpl;
import com.ibm.ws.objectgrid.server.ShutdownHookManagerFactory;
import com.ibm.ws.objectgrid.util.ObjectGridRASUtil;
import com.ibm.ws.xs.NLSConstants;
import com.ibm.ws.xs.util.Messages;
import com.ibm.ws.xs.util.dopriv.DoPrivUtil;

/**
 * A singleton factory for creating and retrieving the ObjectGrid Server singleton.
 *
 * @since WAS XD 6.1 FIX3
 */
public final class ServerFactory {
    private static TraceComponent tc = null;

    static Server instance;

    static final String CLASS_NAME = ServerFactory.class.getName();

    private static final ServerPropertiesImpl serverProperties = new ServerPropertiesImpl();

    private static final CatalogServerPropertiesImpl catalogServerProperties = new CatalogServerPropertiesImpl();

    private static List messages = new ArrayList();

    private static List traces = new ArrayList();

    final static private boolean isPropFileFound;

    static {
        final ServerPropertiesImpl serverProps = serverProperties;
        final CatalogServerPropertiesImpl catalogProps = catalogServerProperties;
        final List finalMessages = messages;
        final List finalTraces = traces;

        PrivilegedAction action = new PrivilegedAction() {

            /**
             * bug 16176 - re-work on the logic of loading server property files.
             * 1) load objectGridServer.properties from classpath.
             * 2) Check -Dobjectgrid.server.props and
             *    -Dcom.ibm.websphere.objectgrid.CatalogServerProperties. If
             *    -Dobjectgrid.server.props exists, then load it using file path and go to next
             *    step. Otherwise if -Dcom.ibm.websphere.objectgrid.CatalogServerProperties
             *    exists, load it using file path and go to the next step. The load here will
             *    actually merge the properties found from this step with the last step.
             * @return
             * @see java.security.PrivilegedAction#run()
             */
            public Object run() {

                // First load objectGridServer.properties from classpath.
                boolean isFileFound = getServerPropsFromClassPath();

                // Check for the -D for CatalogServerProperties and
                String catalogServerProps = System.getProperty(Constants.CATALOGSERVERPROPERTIES_PROPERTY);
                String ogServerProps = System.getProperty(Constants.OGSERVER_PROPS_SYSTEM_PROPERTY);

                String fileToLoad = null;

                if (ogServerProps != null && ogServerProps.trim().length() > 0) {
                    fileToLoad = ogServerProps.trim();
                    finalTraces.add("System property " + Constants.OGSERVER_PROPS_SYSTEM_PROPERTY + " is specified: "
                            + fileToLoad);
                } else if (catalogServerProps != null && catalogServerProps.trim().length() > 0) {
                    fileToLoad = catalogServerProps.trim();
                    finalTraces.add("System property " + Constants.CATALOGSERVERPROPERTIES_PROPERTY + " is specified: "
                            + fileToLoad);
                }

                if (fileToLoad != null) {
                    boolean temp = getServerPropsFromFile(fileToLoad);
                    isFileFound = isFileFound || temp;
                    // We don't want to log a deprecated message here if catalog server property
                    // is used because we hope nobody knows it.
                }

                return isFileFound ? Boolean.TRUE : Boolean.FALSE;

            }

            private final boolean getServerPropsFromClassPath() {
                ClassLoader classLoader = Thread.currentThread().getContextClassLoader();
                Enumeration enumeration = null;

                finalTraces.add("Check default server Property file=" + Constants.SERVERPROPERTIES_DEFAULT_FILE);

                try {
                    enumeration = classLoader.getResources(Constants.SERVERPROPERTIES_DEFAULT_FILE);
                } catch (IOException ioe) {
                    finalTraces.add("getServerPropsFromClassPath - IOException=" + ioe);
                }

                if (enumeration == null) {
                    return false;
                }

                boolean isFileFound = false;
                while (enumeration.hasMoreElements()) {
                    URL element = (URL) enumeration.nextElement();
                    InputStream input = null;

                    boolean isExceptionOccurred = false;

                    try {
                        input = element.openStream();
                        if (input != null) {
                            serverProps.load(input);
                            isFileFound = true;
                        }
                    } catch (IOException e) {
                        FFDCFilter.processException(e, CLASS_NAME + ".getServerPropsFromClassPath", "141");
                        finalTraces.add("getServerPropsFromClassPath - load serverProps - IOException=" + e);
                        isExceptionOccurred = true;
                    } finally {
                        if (input != null) {
                            try {
                                input.close();
                            } catch (IOException closeE) {
                            } finally {
                                input = null;
                            }
                        }
                    }

                    if (isExceptionOccurred) {
                        continue;
                    }

                    try {
                        input = element.openStream();
                        if (input != null) {
                            catalogProps.load(input);
                        }
                    } catch (IOException e) {
                        FFDCFilter.processException(e, CLASS_NAME + ".getServerPropsFromClassPath", "163");
                        finalTraces.add("getServerPropsFromClassPath - load catalogProps - IOException=" + e);
                        isExceptionOccurred = true;
                    } finally {
                        if (input != null) {
                            try {
                                input.close();
                            } catch (IOException closeE) {
                            } finally {
                                input = null;
                            }
                        }
                    }

                    finalMessages.add(new Message(Message.INFO, NLSConstants.LOADED_PROPERTY_FILES_CWOBJ0913,
                            new Object[] { element }));

                }

                return isFileFound;
            }

            private boolean getServerPropsFromFile(String fileToLoad) {
                FileInputStream fi = null;

                boolean isExceptionOccurred = false;

                try {
                    fi = new FileInputStream(fileToLoad);
                    catalogProps.load(fi);

                } catch (FileNotFoundException e) {
                    finalMessages.add(new Message(Message.WARNING, NLSConstants.SERVER_PROPERTY_NOT_FOUND_CWOBJ0919,
                            new Object[] { fileToLoad }));
                    isExceptionOccurred = true;
                } catch (IOException e) {
                    FFDCFilter.processException(e, CLASS_NAME + ".getServerPropsFromFile", "187");
                    finalTraces.add("getServerPropsFromFile - load serverProps -IOException=" + e);
                    isExceptionOccurred = true;
                } finally {
                    if (fi != null) {
                        try {
                            fi.close();
                        } catch (IOException closeE) {
                        } finally {
                            fi = null;
                        }
                    }
                }

                if (isExceptionOccurred) {
                    return false;
                }

                try {
                    fi = new FileInputStream(fileToLoad);
                    serverProps.load(fi);
                } catch (FileNotFoundException e) {
                    isExceptionOccurred = true;
                } catch (IOException e) {
                    FFDCFilter.processException(e, CLASS_NAME + ".getServerPropsFromFile", "202");
                    finalTraces.add("getServerPropsFromFile - load catalogProps -IOException=" + e);
                    isExceptionOccurred = true;
                } finally {
                    if (fi != null) {
                        try {
                            fi.close();
                        } catch (IOException closeE) {
                        } finally {
                            fi = null;
                        }
                    }
                }

                if (!isExceptionOccurred) {
                    finalMessages.add(new Message(Message.INFO, NLSConstants.LOADED_PROPERTY_FILES_CWOBJ0913,
                            new Object[] { fileToLoad }));
                }

                return !isExceptionOccurred;
            }
        };

        isPropFileFound = ((Boolean) AccessController.doPrivileged(action)).booleanValue();
    }

    private ServerFactory() {
        // private constructor for singleton
    }

    /**
     * Answers true if the ServerFactory was initialized with a server properties file.
     * @since 7.1.1
     */
    static public boolean isPropFileFound() {
        return isPropFileFound;
    }

    /**
     * Retrieves the Server singleton for the process, where the first call to this method will initialize the
     * singleton.
     *
     * Note, if user wants this server to include the Catalog Service in addition to any Containers they create,
     * then prior to calling getInstance() for the first time on this process,
     * get the CatalogServerProperties object via the getCatalogProperties() method on this class, and the call
     * setCatalogServer(true) on that object.
     *
     * Containers can be created and started after the getInstance method is called.  First create a
     * DeploymentPolicy, and then pass that into the Server.createContainer method.
     *
     * @see com.ibm.websphere.objectgrid.server.CatalogServerProperties
     * @see com.ibm.websphere.objectgrid.server.Container
     * @see com.ibm.websphere.objectgrid.server.Server
     * @see com.ibm.websphere.objectgrid.deployment.DeploymentPolicyFactory
     * @see com.ibm.websphere.objectgrid.deployment.DeploymentPolicy
     *
     * @return The currently running server instance.
     */
    public static synchronized Server getInstance() {
        if (instance != null) {
            return instance;
        }

        return (Server) AccessController.doPrivileged(new PrivilegedAction() {
            public Object run() {
                return initialize();
            }
        });
    }

    static Server initialize() {
        if (instance != null) {
            throw new IllegalStateException("An instance of server is already running.");
        }

        // To avoid issues with the ImplFactory in Liberty, the forcedist property must be set when running on a non-ZOS platform:
        if (!RuntimeInfo.instance().isWASProcess()) {
            AccessController.doPrivileged(new PrivilegedAction<Object>() {
                public Object run() {
                    System.setProperty("com.ibm.websphere.zos.forcedist", "true");
                    return null;
                }
            });
        }

        serverProperties.setServer(true);

        String serverName = serverProperties.getServerName();
        if (serverName == null) {
            String FS = System.getProperty("file.separator");
            String libertyDirName = System.getProperty("server.config.dir");
            
            if (libertyDirName != null) {// use liberty server name if running in liberty
                libertyDirName = libertyDirName.replace(FS, " ").trim();
                serverName = libertyDirName.substring(libertyDirName.lastIndexOf(" ") + 1);
                serverProperties.setServerName(serverName);
            }
        }
        if (serverName == null && catalogServerProperties.isCatalogServer()) { // not in liberty
            serverName = ServerPropertiesImpl.DEFAULT_SERVER;
            serverProperties.setServerName(serverName);
        }

        try {
            if (tc == null) {
                if (!RuntimeInfo.instance().isLoggingProvidedByTheFramework()) {
                    ObjectGridRASUtil.beginRasInitialization();
                }
                tc = Tr.register(ServerFactory.class, Constants.TR_CATALOGSERVER_GROUP_NAME,
                        Constants.TR_RESOURCE_BUNDLE_NAME);
            }
            
            /* Print out WXS properties to the logs. */
//            WXSProperties.logProperties();

            if (tc.isDebugEnabled()) {
                int size = traces.size();
                for (int i = 0; i < size; i++) {
                    Tr.debug(tc, (String) traces.get(i));
                }
            }

            if (tc.isEntryEnabled()) {
                // set server before trace just to not confuse as client.
                Tr.entry(tc, "initialize", new Object[] { serverProperties, catalogServerProperties });
            }

            if (messages != null && messages.size() != 0) {
                dumpMessages(tc, messages);
                messages.clear();
            }

            Class serverLifecycleListenerDelegateListener = DoPrivUtil.forName("com.ibm.ws.objectgrid.server.impl.ServerLifecycleListenerDelegateImpl");
            ServerLifecycleListener serverLifecycleListener = (ServerLifecycleListener) serverLifecycleListenerDelegateListener.newInstance();
            Class serverClass = DoPrivUtil.forName("com.ibm.ws.objectgrid.server.impl.ServerImpl");
            Constructor serverCtor = serverClass.getConstructor(new Class[] {ServerLifecycleListener.class});

            final Server server = (Server)serverCtor.newInstance(new Object[] {serverLifecycleListener});
            if (((com.ibm.ws.objectgrid.Server) server).isStopping()) {
                throw new ObjectGridRuntimeException("Server stopping before finished starting");
            }
            instance = server;

            // need to put this logic here instead of in the ServerImpl constructor so that the ServerLifeCycleListener
            // can get a reference to the Server instance using ServerFactory.getInstance()
            if(null!=serverLifecycleListener)serverLifecycleListener.serverStarted();

            //if this is not a WAS process, register the shutdown hook so that the ServerImpl
            //will shutdown cleanly if the process is stopped by an OS signal
            // We don't want to add the shutdown hook in an OSGi process either.
            if (!RuntimeInfo.instance().isWASServerProcess() && !RuntimeInfo.instance().isOSGiClientProcess()){
                Runtime.getRuntime().addShutdownHook(ShutdownHookManagerFactory.getShutdownHook(server));
            }

            if (tc.isEntryEnabled()) {
                Tr.exit(tc, "initialize");
            }
            
            return instance;
        } catch (Exception e) {
            IllegalStateException ise = new IllegalStateException(Messages.getMsg(NLSConstants.CWOBJ0902, serverName));
            ise.initCause(e);
            throw ise;
        }
    }

    /**
     * Returns the properties object instance used in configuring the server.
     *
     * @return The properties object instance.
     */
    public static ServerProperties getServerProperties() {
        return serverProperties;
    }

    /**
     * Returns the properties object instance used in configuring the catalog service aspects of the server.
     *
     * @return The properties object instance.
     */
    public static CatalogServerProperties getCatalogProperties() {
        return catalogServerProperties;
    }

    /**
     * Stops the server if it was previously started
     *
     * @return <code>true</code> if server was stopped, <code>false</code> if not
     */
    public static synchronized boolean stopServer() {
        boolean result = false;
        if (instance != null) {
            result = instance.stopServer();
        }
        return result;
    }

    /**
     * Checks whether the server was started or not
     *
     * @return <code>true</code> if server was started, <code>false</code> if not
     */
    public synchronized static boolean isServerStarted() {
        return instance == null ? false : true;
    }

    private static void dumpMessages(TraceComponent trace, List msgList) {
        int size = msgList.size();

        for (int i = 0; i < size; i++) {
            Message msg = (Message) msgList.get(i);
            if (msg.type == Message.INFO) {
                Tr.info(trace, msg.msGKey, msg.params);
            } else if (msg.type == Message.WARNING) {
                Tr.warning(trace, msg.msGKey, msg.params);
            } else if (msg.type == Message.ERROR) {
                Tr.error(trace, msg.msGKey, msg.params);
            }
        }
    }

    static class Message {
        public static final int INFO = 0;

        public static final int WARNING = 1;

        public static final int ERROR = 2;

        public int type;

        public String msGKey;

        public Object[] params;

        public Message(int type, String msGKey, Object[] params) {
            super();
            this.type = type;
            this.msGKey = msGKey;
            this.params = params;
        }

    }
}
