/*
 * IBM Confidential OCO Source Material
 * 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2012
 * The source code for this program is not published or otherwise divested
 * of its trade secrets, irrespective of what has been deposited with the
 * U.S. Copyright Office.
*/
package com.ibm.websphere.objectgrid.spring;

import java.io.IOException;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.HashSet;
import java.util.Set;
import java.util.concurrent.CountDownLatch;

import org.springframework.beans.factory.InitializingBean;
import org.springframework.cache.Cache;
import org.springframework.cache.support.SimpleValueWrapper;
import org.springframework.util.Assert;
import org.springframework.util.StringUtils;

import com.ibm.ejs.ras.Tr;
import com.ibm.ejs.ras.TraceComponent;
import com.ibm.websphere.objectgrid.BackingMap;
import com.ibm.websphere.objectgrid.ObjectGrid;
import com.ibm.websphere.objectgrid.ObjectGridException;
import com.ibm.websphere.objectgrid.ObjectMap;
import com.ibm.websphere.objectgrid.PartitionManager;
import com.ibm.websphere.objectgrid.Session;
import com.ibm.websphere.objectgrid.datagrid.AgentManager;
import com.ibm.websphere.objectgrid.datagrid.MapGridAgent;
import com.ibm.ws.ffdc.FFDCFilter;
import com.ibm.ws.objectgrid.Constants;
import com.ibm.ws.objectgrid.datagrid.AgentManagerImpl;
import com.ibm.ws.objectgrid.spring.CacheKeyWrapper;
import com.ibm.ws.objectgrid.spring.CacheValueWrapper;
import com.ibm.ws.objectgrid.spring.PutAgent;
import com.ibm.ws.objectgrid.spring.SpringCachePartitonPingAgent;
import com.ibm.ws.objectgrid.util.security.SecurityContextRunnableFactory;
import com.ibm.ws.xs.NLSConstants;
import com.ibm.ws.xs.io.ObjectStreamPool;
import com.ibm.ws.xs.io.ObjectStreamPool.ReusableInputStream;
import com.ibm.ws.xs.io.ObjectStreamPool.ReusableOutputStream;
import com.ibm.ws.xs.util.Messages;

/**
 * <p>This class is a WebSphere eXtreme Scale implementation of the Spring Framework's {@link org.springframework.cache.Cache}
 * interface.</p>
 * 
 * <p>Users must provide a name and a {@link ObjectGridClientBean} to configure this cache. The ObjectMap name
 * is optional when using the provided XML configuration files.</p>
 * 
 * <p>This implementation allows for the storage of <code>null</code> values and does not support <code>null</code> keys.</p>
 * 
 * <p>The following Spring Inversion of Control (IoC) container configuration snippet creates two caches, named default and books
 * hosted by the catalog service domain with connection end-points of <tt>host1:2809,host2:2809</tt>.</p>
 * 
 * <pre>
 * &lt;bean id="wxsCSDomain" class="com.ibm.websphere.objectgrid.spring.ObjectGridCatalogServiceDomainBean"
 *     p:catalog-service-endpoints="host1:2809,host2:2809" /&gt;
 *       
 * &lt;bean id="wxsGridClient" class="com.ibm.websphere.objectgrid.spring.ObjectGridClientBean"  
 *      p:catalog-service-domain-ref="wxsCSDomain" /&gt;
 *        
 * &lt;bean id="cacheManager" class="org.springframework.cache.support.SimpleCacheManager"&gt;
 *     &lt;property name="caches"&gt;
 *         &lt;set&gt;
 *             &lt;bean class="com.ibm.websphere.objectgrid.spring.ObjectGridCache" 
 *                 p:name="default" 
 *                 p:object-grid-client-ref="wxsGridClient" /&gt;
 *             &lt;bean class="com.ibm.websphere.objectgrid.spring.ObjectGridCache" 
 *                 p:name="books" 
 *                 p:object-grid-client-ref="wxsGridClient" /&gt;
 *         &lt;/set&gt;
 *     &lt;/property&gt;
 * &lt;/bean&gt;
 * </pre>
 * 
 * @since 8.5, XC10
 */
public final class ObjectGridCache implements Cache, InitializingBean {
    
    final static String CLASS_NAME = ObjectGridCache.class.getName();
    final static TraceComponent tc = Tr.register(CLASS_NAME, Constants.TR_SPRING_GROUP_NAME,
            Constants.TR_RESOURCE_BUNDLE_NAME);
    
    /** Prefix for the template map used in the provided objectgrid and deployment xml files */
    private static final String TEMPLATE_MAP_PREFIX = "IBM_SPRING_PARTITIONED_";
    
    /** Known key types */
    private static final Set<Class<?>> KEY_TYPES = new HashSet<Class<?>>(); 
    
    static {
        KEY_TYPES.add(Character.class);
        KEY_TYPES.add(String.class);
        KEY_TYPES.add(Byte.class);
        
        // numbers
        KEY_TYPES.add(Short.class);
        KEY_TYPES.add(Integer.class);
        KEY_TYPES.add(Long.class);
        KEY_TYPES.add(Float.class);
        KEY_TYPES.add(Double.class);
        KEY_TYPES.add(BigDecimal.class);
        KEY_TYPES.add(BigInteger.class);
        
        KEY_TYPES.add(Date.class);
        
        // SQL types
        KEY_TYPES.add(java.sql.Date.class);
        KEY_TYPES.add(java.sql.Time.class);
        KEY_TYPES.add(java.sql.Timestamp.class);
    }
    
    private static final ObjectStreamPool objectStreamPool = ObjectStreamPool.getInstance();
    
    /**
     * The object grid client configuration to use.
     */
    private ObjectGridClientBean objectGridClient;
    
    // Fast-fail related
    private static final String DISABLE_SPRING_FAST_FAIL_KEY = "com.ibm.websphere.objectgrid.spring.disable.fastfail";
    // true disables fast fail
    private static final boolean fastFailDisabled = Boolean.parseBoolean(System.getProperty(DISABLE_SPRING_FAST_FAIL_KEY));
    private CountDownLatch gridDownLatch;
    private boolean connectedToGrid = true; // never false if fast fail is disabled
    private Thread gridDownProcThread;
    private ArrayList<Integer> keysForFastFailAgent;
    
    
    /**
     * Cache name
     */
    private String name;
    
    /**
     * ObjectMap name.
     */
    private String mapName;
    
    /**
     * Initializes this cache. This method is to be called after all setter methods have been called and will throw
     * an {@link IllegalArgumentException} if {@link #setObjectGridClient(ObjectGridClientBean)} or 
     * {@link #setName(String)} have not been called prior.
     * @throws Exception if {@link #setObjectGridClient(ObjectGridClientBean)} or {@link #setName(String)} 
     * have not been called prior or an error occurs initializing the cache.
     * @see org.springframework.beans.factory.InitializingBean#afterPropertiesSet()
     */
    public void afterPropertiesSet() throws Exception {
        if (tc.isEntryEnabled()) {
            Tr.entry(tc, "afterPropertiesSet");
        }
        Assert.hasLength(name, Messages.getMsg(NLSConstants.INVALID_VALUE_ERROR_CWOBJ0008, new Object[]{name, "name"}));
        Assert.notNull(objectGridClient, 
                Messages.getMsg(NLSConstants.INVALID_VALUE_ERROR_CWOBJ0008, new Object[]{name, "objectGridClient"}));
        if (!StringUtils.hasLength(this.mapName)) {
            setMapName(TEMPLATE_MAP_PREFIX + name);
            if (tc.isDebugEnabled()) {
                Tr.debug(tc, "No value specified for mapName so using " + mapName + " for the map name");
            }
        }
        
        if (tc.isDebugEnabled()) {
            Tr.debug(tc, "name: " + name);
            Tr.debug(tc, "mapName: " + mapName);
        }
        
        // get the map or initialize the template map
        Session session = null;
        try {
            session = getSession();
            ObjectMap map = session.getMap(mapName);
        } catch (Exception e) {
            FFDCFilter.processException(e, CLASS_NAME + ".afterPropertiesSet", "142", this, new Object[] {});
            if (tc.isEventEnabled()) {
                Tr.event(tc, "Error retrieving ObjectMap", e);
            }
            throw e; //rethrow
        } finally {
            if (session != null) {
                session.close();
            }
        }
        
        // setup fast-fail
        if (fastFailDisabled) {
            Tr.info(tc, NLSConstants.SPRING_FAST_FAIL_DISABLED_CWOBJ7604);
        } else {
            BackingMap backingMap = objectGridClient.getObjectGrid().getMap(mapName);
            if (backingMap != null) { // should never be null
                // get a key for each partition
                PartitionManager partitionManager = backingMap.getPartitionManager();
                int numPartitions = partitionManager.getNumOfPartitions();
                
                if (tc.isDebugEnabled()) {
                    Tr.debug(tc, "Number of partitions for map " + mapName + ":" + numPartitions);
                }
                
                keysForFastFailAgent = new ArrayList<Integer>(numPartitions);
                for (int i = 0; i < numPartitions; i++){
                    Integer tmp = Integer.valueOf(i);
                    keysForFastFailAgent.add(tmp);
                }
            }
            
            // start the fast-fail thread if fast fail is not disabled
            gridDownProcThread = new Thread(SecurityContextRunnableFactory.getRunnable(new SpringGridDownProcessor()));
            gridDownProcThread.setDaemon(true);
            gridDownProcThread.start();
            gridDownProcThread.setName("WXSSpringFastFail[" + name + ":" + mapName + "]");
        }
    }
        
    /**
     * Returns the name of the cache.
     * @return The name of the cache.
     * @see org.springframework.cache.Cache#getName()
     */
    public String getName() {
        return name;
    }
    
    
    /**
     * Sets the cache name.
     * @param name The cache name.
     * @see #getName()
     */
    public void setName(String name) {
        this.name = name;
    }
    
    /**
     * Sets the ObjectMap name. This is optional when using the provided XML configuration files.
     * @param mapName The name of the ObjectMap
     */
    public void setMapName(String mapName) {
        this.mapName = mapName;
    }
    
    /**
     * Sets the {@link ObjectGridClientBean} to use.
     * @param objectGridClient The {@link ObjectGridClientBean} to use
     */
    public void setObjectGridClient(ObjectGridClientBean objectGridClient) {
        this.objectGridClient = objectGridClient;
    }
    
    /**
     * This implementation returns <code>null</code>.
     * @return This implementation returns <code>null</code>.
     * @see org.springframework.cache.Cache#getNativeCache()
     */
    public Object getNativeCache() {
        return null;
    }

    /**
     * Retrieves the object from the cache at the given key. Returns <code>null</code> if there is no mapping for the 
     * key or if the key is <code>null</code>
     * @return the object from the cache at the given key or <code>null</code> if there is no mapping or the key
     * is <code>null</code>
     * @see org.springframework.cache.Cache#get(java.lang.Object)
     */
    public ValueWrapper get(Object key) {
        if (tc.isEntryEnabled()) {
            Tr.entry(tc, "get", key);
        }
        
        if (!connectedToGrid || key == null) {
            return null;
        }
        
        Session session = null;
        ValueWrapper returnValue = null;
        
        try {
            // process the key first so that if it fails we fail out early
            Object theKey = processKey(key);
            
            session = getSession();
            if (session != null) {
                ObjectMap map = session.getMap(mapName);
                CacheValueWrapper wrappedValue = (CacheValueWrapper)map.get(theKey);
                if (wrappedValue != null) {
                    Object obj = byteArrayToObject(wrappedValue.getValue());
                    
                    if (tc.isDebugEnabled()) {
                        Tr.debug(tc, "Retrieved entry ["+ key + "," + obj + "]");
                    }
                    
                    returnValue = new SimpleValueWrapper(obj);
                }
            }
        } catch (Exception e) {
            FFDCFilter.processException(e, getClass().getName() + ".get", "188" , this, new Object[] {key});
            if (tc.isEventEnabled()) {
                Tr.debug(tc, "Error occured retrieving value for key '" + key + "'", e);  // use debug so the exception will show only once
            }
            // fail fast logic
            checkGridDown(e);
        } finally {
            if (session != null) {
                session.close();
            }
        }
        // null if there was no entry, a SimpleValueWrapper(null) if the value is null
        return returnValue; 
    }
    
    
    /**
     * Creates an entry in the cache. Overwrites the value if an
     * entry for the given key exists. This method does not create an entry if the given key is <code>null</code>.
     * @see org.springframework.cache.Cache#put(java.lang.Object, java.lang.Object)
     */
    public void put(Object key, Object value) {
        if (tc.isEntryEnabled()) {
            Tr.entry(tc, "put", new Object[]{key, value});
        }
        if (!connectedToGrid || key == null) {
            return;
        }
        
        Session session = null;
        try {
            // process key and wrap value first so that if it fails we fail out early
            Object theKey = processKey(key);
            CacheValueWrapper valueWrapper = wrapValue(value);
            
            session = getSession();
            if (session != null) {
                ObjectMap map = session.getMap(mapName);
                // need putagent pooling
                MapGridAgent putAgent = new PutAgent(valueWrapper);
                AgentManager agentManager = map.getAgentManager();
                ((AgentManagerImpl)agentManager).setSynchronousMode(true);
                agentManager.callMapAgent(putAgent, Collections.singletonList(theKey));
            }
        } catch (Exception e) {
            FFDCFilter.processException(e, getClass().getName() + ".put", "224", this, new Object[] {key, value});
            if (tc.isEventEnabled()) {
                Tr.debug(tc, "Error occured while executing put for key '" + key + "'", e);  // use debug so the exception will show only once
            }
            // fail fast logic
            checkGridDown(e);
        } finally {
            if (session != null) {
                session.close();
            }
        }
    }
    
    /**
     * Evicts the entry at the given key. If the provided key is <code>null</code> this method does not evict any entries.
     * @see org.springframework.cache.Cache#evict(java.lang.Object)
     */
    public void evict(Object key) {
        if (tc.isEntryEnabled()) {
            Tr.entry(tc, "evict", key);
        }
        if (!connectedToGrid || key == null) {
            return;
        }
        
        Session session = null;
        try {
            // process the key first so that if it fails we fail out early
            Object theKey = processKey(key);
            
            session = getSession();
            if (session != null) {
                ObjectMap map = session.getMap(mapName);
                map.remove(theKey);
            }
        } catch (Exception e) {
            FFDCFilter.processException(e, getClass().getName() + ".evict", "254" , this, new Object[] {key});
            if (tc.isEventEnabled()) {
                Tr.debug(tc, "Error occured evicting entry with key '" + key + "'", e); // use debug so the exception will show only once
            }
            // fail fast logic
            checkGridDown(e);
        } finally {
            if (session != null) {
                session.close();
            }
        }
    }
    
    /**
     * Clears all entries from the cache.
     * @see org.springframework.cache.Cache#clear()
     */
    public void clear() {
        if (tc.isEntryEnabled()) {
            Tr.entry(tc, "clear");
        }
        if (!connectedToGrid) {
            return;
        }
        
        Session session = null;
        try {
            session = getSession();
            if (session != null) {
                ObjectMap map = session.getMap(mapName);
                map.clear();
            }
        } catch (Exception e) {
            FFDCFilter.processException(e, getClass().getName() + ".clear", "278" , this);
            if (tc.isEventEnabled()) {
                Tr.debug(tc, "Error occured while clearing the cache", e); // use debug so the exception will show only once
            }
            // fail fast logic
            checkGridDown(e);
        } finally {
            if (session != null) {
                session.close();
            }
        }
    }
    
    /**
     * Retrieves a ObjectGrid session from the object grid client configuration.
     * @return an ObjectGrid session or null if one cannot be retrieved.
     * @throws ObjectGridException if an error occurs retrieving a session
     */
    private Session getSession() throws ObjectGridException {
        if (tc.isEntryEnabled()) {
            Tr.entry(tc, "getSession");
        }
        if (objectGridClient != null) {
            ObjectGrid grid = objectGridClient.getObjectGrid();
            if (grid != null) {  // shouldn't be possible 
                return grid.getSession();
            }
        }
        // should be impossible, a proper exception would be thrown from getSession()
        Tr.debug(tc, "Unable to retrieve session from the object grid client configuration");
        return null;
    }
    
    /**
     * If key is not a 'known' key type then wrap it and return it. If the key is known,
     * then return it. 
     * @throws ObjectGridException 
     */
    private static Object processKey(Object key) throws ObjectGridException {
        if (tc.isEntryEnabled()) {
            Tr.entry(tc, "processKey", key);
        }
        if (KEY_TYPES.contains(key.getClass())) {
            return key;
        }
        
        if (tc.isDebugEnabled()) {
            Tr.debug(tc, "Wrapping key: " + key);
        }
        
        byte[] keyAsBytes = null;
        try {
            keyAsBytes = objectToByteArray(key);
        } catch (Exception e) {
            FFDCFilter.processException(e, CLASS_NAME + ".processKey", "330", new Object[] {key});
            if (tc.isEventEnabled()) {
                Tr.event(tc, "Error occured while serializing a cache entry key", e);
            }
            Tr.error(tc, Messages.getMsg(NLSConstants.CANNOT_SERIALIZE_KEY_CWOBJ7601, key), CLASS_NAME);
            throw new ObjectGridException(e);
        }
        
        // not a known key type so convert it
        return new CacheKeyWrapper(keyAsBytes);
    }
    
    /**
     * Converts an object to a <code>byte[]</code>
     * @param object the object to convert
     * @return the <code>byte[]</code> representing the serialized form of the object to convert
     * @throws IOException if an error occurs converting the object to a <code>byte[]</code>
     */
    private static byte[] objectToByteArray(Object object) throws IOException {
        if (tc.isEntryEnabled()) {
            Tr.entry(tc, "objectToByteArray");
        }
        
        if (object == null) {
            return null;
        }
        
        ReusableOutputStream outputStream = null;
        byte[] result = null;
        try {
            outputStream = objectStreamPool.getOutputStream();
            outputStream.writeObject(object);
            result = outputStream.toByteArray();
        } catch (IOException e) {
            FFDCFilter.processException(e, CLASS_NAME + ".objectToByteArray", "364");
            if (tc.isEventEnabled()) {
                Tr.event(tc, "Error occured while converting object to byte array", e);
            }
            throw e;
        } finally {
            if (outputStream != null) {
                objectStreamPool.returnOutputStream(outputStream);
            }
        }
        return result;
    }
    
    /**
     * Converts a <code>byte[]</code> to a object
     * @param byteArray the bytes to convert
     * @return the object represented by <code>byteArray</code>
     * @throws ClassNotFoundException if an the class of the resulting object is not found
     * @throws IOException if an error occurs converting the <code>byte[]</code> to a object
     */
    private static Object byteArrayToObject(byte[] byteArray) throws IOException, ClassNotFoundException {
        if (tc.isEntryEnabled()) {
            Tr.entry(tc, "byteArrayToObject");
        }
        if (byteArray == null) {
            return null;
        }
        
        ReusableInputStream inputStream = null;
        Object result = null;
        try {
            inputStream = objectStreamPool.getInputStream(byteArray);
            result = inputStream.readObject();
        } catch (IOException e) {
            FFDCFilter.processException(e, CLASS_NAME + ".byteArrayToObject", "398");
            if (tc.isEventEnabled()) {
                Tr.event(tc, "Error occured while converting byte array to object", e);
            }
            throw e;
        } catch (ClassNotFoundException e) {
            FFDCFilter.processException(e, CLASS_NAME + ".byteArrayToObject", "404");
            if (tc.isEventEnabled()) {
                Tr.event(tc, "Error occured while converting byte array to object", e);
            }
            throw e;
        } finally {
            if (inputStream != null) {
                objectStreamPool.returnInputStream(inputStream);
            }
        }
        return result;
    }
    
    private static CacheValueWrapper wrapValue(Object value) throws ObjectGridException {
        if (tc.isEntryEnabled()) {
            Tr.entry(tc, "wrapValue", value);
        }
        
        byte[] valueAsBytes = null;
        try {
            valueAsBytes = objectToByteArray(value);
        } catch (Exception e) {
            FFDCFilter.processException(e, CLASS_NAME + ".wrapValue", "426", new Object[] {value});
            if (tc.isEventEnabled()) {
                Tr.event(tc, "Error occured while serializing a cache entry value", e);
            }
            
            Tr.error(tc, Messages.getMsg(NLSConstants.CANNOT_SERIALIZE_VALUE_CWOBJ7600, value), CLASS_NAME);
            throw new ObjectGridException(e);
        }
        
        return new CacheValueWrapper(valueAsBytes);
    }
    
    private void checkGridDown(Exception ex) {
        // never set connectedToGrid = false if fast fail is disabled
        if (fastFailDisabled) { 
            return;
        }
        if (connectedToGrid && isTargetNotAvailableInStack(ex)) {
            connectedToGrid = false;
            // log a message saying the grid is down
            Tr.error(tc, NLSConstants.SPRING_CACHE_GRID_DISCONNECTED_CWOBJ7605, new Object[]{name, objectGridClient.getObjectGrid(), mapName});
            gridDownLatch.countDown();// start monitoring thread for grid reconnect
        }
    }
    
    /**
     * Returns true if TargetNotAvailableException is in the exception stack and false otherwise.
     * This should probably be placed in a common area and shared with dynacache
     * @param ex exception to examine
     * @return whether TargetNotAvailableException is in the exception stack
     */
    private static boolean isTargetNotAvailableInStack(Exception ex) {
        if (ex instanceof com.ibm.websphere.objectgrid.TargetNotAvailableException) {
            return true;
        }
        Throwable cause = ex.getCause();
        while (cause != null){
            if (cause instanceof com.ibm.websphere.objectgrid.TargetNotAvailableException){
                return true;
            }
            cause = cause.getCause();
        }
        
        return false;
    }
    
    private boolean checkGridConnection() {
        if (tc.isDebugEnabled()) {
            Tr.entry(tc,  "checkGridConnection", mapName);
        }
        boolean isConnected = false;
        
        Session session = null;
        try {
            session = getSession();
            ObjectMap map = session.getMap(mapName);
            MapGridAgent gridHealthAgent = new SpringCachePartitonPingAgent();
            AgentManager agentManager = map.getAgentManager();
            agentManager.callMapAgent(gridHealthAgent, keysForFastFailAgent);
            // if it returns without error then it should mean we're connected
            isConnected = true;
        } catch (Exception e) {
            // exception while checking health
            FFDCFilter.processException(e, CLASS_NAME + ".checkGridConnection", "523" , this);
            if (tc.isEventEnabled()) {
                Tr.event(tc, "Error occured while checking the grid connection", e);
            }
            if (!isTargetNotAvailableInStack(e)) {
                if (tc.isDebugEnabled()) {
                    Tr.debug(tc, "Encountered an exception that was not a TargetNotAvailableException while checking grid connection", e);
                }
                isConnected =  true;
            }
        } finally {
            if (session != null) {
                session.close();
            }
        }
        if (tc.isEntryEnabled()) {
            Tr.exit(tc, "checkGridConnection", isConnected);
        }
        return isConnected;
    }
    
    /**
     * Thread for fast fail
     */
    class SpringGridDownProcessor implements Runnable {
        
        private final String CLASS_NAME = SpringGridDownProcessor.class.getName();
        private final TraceComponent tc = Tr.register(CLASS_NAME, Constants.TR_SPRING_GROUP_NAME,
                Constants.TR_RESOURCE_BUNDLE_NAME);
        
        private final int RETRY_TIME = 5000;    // 5 seconds
        private boolean stopRequested = false;
        
        public SpringGridDownProcessor() {
            // empty
        }

        public void run() {
            if (tc.isEntryEnabled()) {
                Tr.entry(tc, "run");
            }
            try {
                while (!stopRequested) {
                    gridDownLatch = new CountDownLatch(1);
                    gridDownLatch.await();
                    
                    // if the connection is down then keep checking periodically until we are reconnected
                    while (!connectedToGrid) {
                        connectedToGrid = checkGridConnection();
                        if (connectedToGrid) {
                            Tr.info(tc, NLSConstants.SPRING_CACHE_GRID_RECONNECTED_CWOBJ7606, new Object[]{name, objectGridClient.getObjectGrid(), mapName});
                        } else {
                            // still not connected
                            if (tc.isDebugEnabled()) {
                                Tr.debug(tc, "Not connected to cache " + name + " and map " + mapName);
                            }
                        }
                        if (!connectedToGrid) {
                            //wait before checking connection again
                            Thread.sleep(RETRY_TIME);
                        }
                    }
                    
                } // while(!stopRequested)
            } catch (Throwable t) {
                if (tc.isDebugEnabled()) {
                    Tr.debug(tc, "GridDownProcessor thread failed: ", t);
                }
                FFDCFilter.processException(t, CLASS_NAME + ".run", "520", this);
                if (tc.isEventEnabled()) {
                    Tr.event(tc, "unexpected", t);
                }
            }
            if (tc.isEntryEnabled()) {
                Tr.exit(tc, "run");
            }
            
        }

        public void stop() {
            if (tc.isEntryEnabled()) {
                Tr.entry(tc, "Stopping GridDownProcessor");
            }
            stopRequested = true;
            if (tc.isDebugEnabled()) {
                Tr.debug(tc, "Set stopRequested");
            }
        }
    }
}
