// IBM Confidential OCO Source Material
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2005,2012
// The source code for this program is not published or otherwise divested
// of its trade secrets, irrespective of what has been deposited with the
// U.S. Copyright Office.
//
// $HeadURL$
// $Id$
//
// Module  :  StatsAccessor.java
//
// Source File Description:
//
//  Need a file description.
//
// Change Activity:
//
// Reason       Version   Date         Userid    Change Description
// ------------ --------- ------------ --------- -----------------------------------------
// XD60_M3_B.9  WASX.XD   04/15/2005   jitang    ObjectGrid - Metric enablement - Drop 1
// d346485      WXD601.XD 02/10/2006   jitang    House cleaning for stats
// b10090       XD 6.1    10/10/2007   cdjohnson Fix javadoc to incorporate new dynamic environment.
// ------------ --------- ------------ --------- -----------------------------------------
package com.ibm.websphere.objectgrid.stats;

import com.ibm.websphere.objectgrid.management.AgentManagerMBean;
import com.ibm.websphere.objectgrid.management.HashIndexMBean;
import com.ibm.websphere.objectgrid.management.MapMBean;
import com.ibm.websphere.objectgrid.management.ObjectGridMBean;
import com.ibm.websphere.objectgrid.management.QueryManagerMBean;
import com.ibm.websphere.objectgrid.management.SessionMBean;

/**
 * This interface is used to access <tt>StatsFact</tt> objects from the statistic
 * repository and set the statistic specification for all in-process ObjectGrid instances.<p>
 *
 * Retrieve a StatsAccessor instance using the <code>StatsAccessorFactory.getStatsAccessor()</code> method<p>
 *
 * Setting the StatsSpec using the StatsAccessor only applies the stats specification to all ObjectGrid instances running in
 * the current process (for example: currently allocated shards).<p>
 * 
 * If the ObjectGrid is a distributed ObjectGrid, statistics specification can be set as follows:
 * <ul>
 * <li> Dynamic deployment topology: Set the <tt>statsSpec</tt> property in the
 * server.properties file.
 * <li> The <code>xscmd</code> command-line utility.
 * </ul><p>
 *
 * StatsFact objects should only be used for local ObjectGrid instances or
 * the client side (near cache) ObjectGrid instances. In a distributed environment, use the following
 * mechanisms:
 * <ul>
 * <li>Dynamic deployment topology:  Use the following JMX MBeans:
 *  <ul>
 *  <li>{@link ObjectGridMBean#retrieveStatsModule()}
 *  <li>{@link MapMBean#retrieveStatsModule()}
 *  <li>{@link SessionMBean#retrieveStatsModule()}
 *  <li>{@link HashIndexMBean#retrieveStatsModule()}
 *  <li>{@link QueryManagerMBean#retrieveStatsModule(String)}
 *  <li>{@link AgentManagerMBean#retrieveStatsModule(String)}
 *  </ul>
 * </ul>
 *
 * @since WAS XD 6.0.1
 * @see StatsFact
 */
public interface StatsAccessor {

    /**
     * Sets the StatsSpec for all the local ObjectGrid instances in this process.  The statistics
     * specification defines the amount of statistics gathering the ObjectGrid is to perform.<p>
     *
     * The statistic specification is applied globally to the entire process and
     * applies to all ObjectGrid and BackingMap instances when initialized.
     * Instances that are already initialized will ignore any StatsSpec changes.
     *
     * @param spec the StatsSpec to be set
     *
     * @see StatsSpec
     */
    void setStatsSpec(StatsSpec spec);

    /**
     * Sets the StatsSpec for all the local ObjectGrid instances in this process.  The statistics
     * specification defines the amount of statistics gathering the ObjectGrid is to perform.<p>
     *
     * The statistic specification is applied globally to the entire process and
     * applies to all ObjectGrid and BackingMap instances when initialized.
     * Instances that are already initialized will ignore any StatsSpec changes.
     *
     * @param specString the StatsSpec string to be set.
     *
     * @see StatsSpec
     */
    void setStatsSpec(String specString);

    /**
     * Retrieves the StatsSpec defined for the process.
     *
     * @return the current StatsSpec setting for this process set using one of the
     *         <code>setStatsSpec</code> methods, or <code>null</code> if
     *         a <code>setStatsSpec</code> method has not been called
     */
    StatsSpec getStatsSpec();

    /**
     * Retrieves the StatsFact object based on the path. The path is specified as
     * an array.  The statistics module is then retrieved from the StatsFact.<p>
     *
     * The path array contains the path to the module. Each path array index is defined follows:<p>
     *
     * <strong>Dynamic Deployment Topology</strong>
     * <ul>
     * <li>OGStatsModule: [0]: Partition ID, [1]: "root", [2]: ObjectGrid name
     * <li>MapStatsModule: [0]: Partition ID, [1]: ObjectGrid name, [2]: Map name
     * <li>SessionStatsModule: [0]: Server ID
     * <li>HashIndexStatsModule: [0]: Partition ID, [1]: ObjectGrid name, [2]: Map name, [3]: Index name
     * <li>QueryStatsModule: [0]: Partition ID, [1]: ObjectGrid name, [2]: Query string
     * <li>AgentStatsModule: [0]: Partition ID, [1]: ObjectGrid name, [2]: Agent class name
     * </ul><p>
     *
     * <strong>Local, in-memory ObjectGrid or client</strong>
     * <ul>
     * <li>OGStatsModule: [0]: "root", [1]: ObjectGrid name
     * <li>MapStatsModule: [0]: ObjectGrid name, [1]: Map name
     * <li>HashIndexStatsModule: [0]: ObjectGrid name, [1]: Map name, [2]: Index name
     * <li>QueryStatsModule: [0]: ObjectGrid name, [1]: Query string
     * </ul><p>
     *
     * If the path array contains only a partial path, it will return an aggregated StatsModule
     * for that level. For example, in a dynamic deployment topology:
     * <pre>
     * StatsFact sf = statsAccessor.getStatsFact(new String[]{"0", "root"}, StatsModule.MODULE_TYPE_OBJECT_GRID);
     * OGStatsModule ogStats = (OGStatsModule) sf.getStatsModule();
     * System.out.println(ogStats.toString());
     * </pre>
     * will retrieve an OGStatsModule with statistics for all ObjectGrids for partition 0.<p>
     *
     * To limit the statistics scope to a single ObjectGrid, set additional arguments in the array.  For example, in a
     * dynamic deployment topology:
     * <pre>
     * StatsFact sf = statsAccessor.getStatsFact(new String[]{"0", "root", "BookstoreGrid"}, StatsModule.MODULE_TYPE_OBJECT_GRID);
     * OGStatsModule ogStats = (OGStatsModule) sf.getStatsModule();
     * System.out.println(ogStats.toString());
     * </pre>
     * will retrieve an OGStatsModule with statistics for the "BookStore" ObjectGrid for partition 0.
     *
     * @param paths the array of paths to the module
     * @param moduleType the module type defined in {@link StatsModule}
     *
     * @return the StatsFact object
     */
    StatsFact getStatsFact(String[] paths, int moduleType);

    /**
     * Retrieves the StatsFact object based on the path. The path is defined using a forward slash ("/") delimited
     * string.  The statistics module is then retrieved from the StatsFact.<p>
     *
     * The path contains the path to the module. The path is defined as follows:<p>
     *
     * <strong>Dynamic Deployment Topology</strong>
     * <ul>
     * <li>OGStatsModule: &LT;Partition ID&GT;/root/&LT;ObjectGrid name&GT;
     * <li>MapStatsModule: &LT;Partition ID&GT;/&LT;ObjectGrid name&GT;/&LT;Map name&GT;
     * <li>SessionStatsModule: &LT;Server ID&GT;
     * <li>HashIndexStatsModule: &LT;Partition ID&GT;/&LT;ObjectGrid name&GT;/&LT;Map name&GT;/&LT;Index name&GT;
     * <li>QueryStatsModule: &LT;Partition ID&GT;/&LT;ObjectGrid name&GT;/&LT;Query string&GT;
     * <li>AgentStatsModule: &LT;Partition ID&GT;/&LT;ObjectGrid name&GT;/&LT;Agent class name&GT;
     * </ul><p>
     *
     * <strong>Local, in-memory ObjectGrid or client</strong>
     * <ul>
     * <li>OGStatsModule: root/&LT;ObjectGrid name&GT;
     * <li>MapStatsModule: &LT;ObjectGrid name&GT;/&LT;Map name&GT;
     * <li>HashIndexStatsModule: &LT;ObjectGrid name&GT;/&LT;Map name&GT;/&LT;Index name&GT;
     * <li>QueryStatsModule: &LT;ObjectGrid name&GT;/&LT;Query string&GT;
     * </ul><p>
     *
     * If the path array contains only a partial path, it will return an aggregated StatsModule
     * for that level. For example, in a dynamic deployment topology:
     * <pre>
     * StatsFact sf = statsAccessor.getStatsFact("0/root", StatsModule.MODULE_TYPE_OBJECT_GRID);
     * OGStatsModule ogStats = (OGStatsModule) sf.getStatsModule();
     * System.out.println(ogStats.toString());
     * </pre>
     * will retrieve an OGStatsModule with statistics for all ObjectGrids for partition 0.<p>
     *
     * To limit the statistics scope to a single ObjectGrid, set additional arguments in the array.  For example, in a
     * dynamic deployment topology:
     * <pre>
     * StatsFact sf = statsAccessor.getStatsFact("0/root/BookstoreGrid", StatsModule.MODULE_TYPE_OBJECT_GRID);
     * OGStatsModule ogStats = (OGStatsModule) sf.getStatsModule();
     * System.out.println(ogStats.toString());
     * </pre>
     * will retrieve an OGStatsModule with statistics for the "BookStore" ObjectGrid for partition 0.
     *
     * @param path the path to the module
     * @param moduleType the module type defined in {@link StatsModule}
     *
     * @return the StatsFact object
     */
    StatsFact getStatsFact(String path, int moduleType);
}


