// IBM Confidential OCO Source Material
// 5724-J34 (C) COPYRIGHT International Business Machines Corp. 2005,2012
// The source code for this program is not published or otherwise divested
// of its trade secrets, irrespective of what has been deposited with the
// U.S. Copyright Office.
//
// Module  :  StatsSpec.java
//
// Source File Description:
//
//
// Change Activity:
//
// Reason       Version   Date         Userid    Change Description
// ------------ --------- ------------ --------- -----------------------------------------
// XD60_M3_B.9  WASX.XD   04/16/2005   jitang    ObjectGrid - Metric enablement - Drop 1
// XD601_M2_B.x WASX.XD   09/07/2005   jitang    ObjectGrid - Statistics enablement for 6.0.1
// d338878      WASX.XD   01/31/2006   jitang    provide request processing time stats
// d332028      WXD601.XD 02/10/2006   sutter    Replication Stats Module updates
// 13570		WASX.XD	  12/04/2008   bsandman	 Spec additions for query, agent, and index
// b18334		XS7.1	  03/22/2010   bsandman	 methods for whether any query, index, or agent statistics are enabled
// wi8555       8.5       04/09/2012   cdjohnson Cleaned-up the javadoc to make it clear how to use the stats strings.
// ------------ --------- ------------ --------- -----------------------------------------
package com.ibm.websphere.objectgrid.stats;

import java.io.Serializable;
import java.util.HashMap;
import java.util.StringTokenizer;

import com.ibm.ejs.ras.Tr;
import com.ibm.ejs.ras.TraceComponent;
import com.ibm.websphere.objectgrid.datagrid.AgentManager;
import com.ibm.ws.objectgrid.Constants;
import com.ibm.ws.objectgrid.stats.VMStatsModule;

/**
 * This class defines the statistics specification strings and integer masks to use for enabling and disabling
 * statistics modules.<p>
 * 
 * StatsSpec object can be constructed in the following three ways:</p>
 * <OL>
 * <LI>Use new StatsSpec() and then call setStatSpec methods;</LI>
 * <LI>Use new StatsSpec(int); </LI>
 * <LI>Use new StatsSpec(String). </LI>
 * </OL><p>
 *
 * ObjectGrid supports the following statistic modules:
 * <ul>
 * <li>{@link MapStatsModule} - Map statistics
 * <li>{@link OGStatsModule} - ObjectGrid statistics
 * <li>{@link SessionStatsModule} - HTTP session statistics
 * <li>{@link HashIndexStatsModule} - MapIndexPlugin statistics.
 * <li>{@link QueryStatsModule} - ObjectQuery and entity query statistics.
 * <li>{@link AgentStatsModule} - AgentManager agent statistics.
 * </ul><p>
 *
 * <p>The statistic specification string is a ";" delimited string.
 * The format of each sub string is "[STATISTIC]=[enabled|disabled]",where
 * the value of [STATISTIC] can be one of the following:</p>
 *
 * <table>
 * <tr><td><b>Statistics Module</b></td><td><b>Statistics String</b></td><td><b>Description</b></td></tr>
 * <tr><td>{@link AgentStatsModule}</td><td>agent.all</td><td>All data grid agent statistics.</td></tr>
 * <tr><td>{@link MapStatsModule}</td><td>map.all</td><td>All map statistics.</td></tr>
 * <tr><td>{@link MapStatsModule}</td><td>map.batchupdate</td><td>Map loader batch update time statistic.</td></tr>
 * <tr><td>{@link MapStatsModule}</td><td>map.hitrate</td><td>Map hit rate statistic.</td></tr>
 * <tr><td>{@link MapStatsModule}</td><td>map.numentries</td><td>Map entry count statistic.</td></tr>
 * <tr><td>{@link MapStatsModule}</td><td>map.usedbytes</td><td>Map entry size statistic.</td></tr>
 * <tr><td>{@link HashIndexStatsModule}</td><td>index.all</td><td>All HashIndex statistics.</td></tr>
 * <tr><td>{@link OGStatsModule}</td><td>og.all</td><td>All ObjectGrid statistics.</td></tr>
 * <tr><td>{@link OGStatsModule}</td><td>og.transactiontime</td><td>ObjectGrid transaction time statistic.</td></tr>
 * <tr><td>{@link QueryStatsModule}</td><td>query.all</td><td>All ObjectQuery and EntityManager Query statistics.</td></tr>
 * <tr><td>{@link SessionStatsModule}</td><td>session.all</td><td>All HTTP session statistics.</td></tr>
 * </table>
 *
 * <p>For example, "map.hitrate=enabled;map.numentries=enabled" will enable both
 * the map hit rate statistic and the map number of entries statistic.
 * "map.all=enabled" will enable all the map statistics.</p>
 *
 * <p>The content of a Object Grid statistic specification
 * is parsed from the left to the right. The statistic
 * specification string scanned later will override the previously scanned
 * string if they represent the same statistic. For example,
 * "map.hitrate=enabled;map.all=disabled" will disable all the statistics, and
 * "map.hitrate=disabled;map.batchupdate=enabled;map.hitrate=enabled" will enable
 * both the map hit rate and map batch update time statistics.</p>
 *
 * <p>the statistic specification string is case-insensitive. For example,
 * "map.hitRATE=enabled" and "MAP.hitrate=ENABLED" are all valid statistic
 * specification strings. However, users are encouraged to use the constant
 * strings defined in this class.
 *
 * <p>Users can also use an integer to set the statistic specification. For example:
 * <pre>
 * setSpec(StatsSpec.MAP_HIT_RATE|StatsSpec.MAP_NUM_ENTRIES);
 * </pre>
 * </p>
 *
 * @since WAS XD 6.0.1
 *
 * @see StatsModule#getStatsSpec()
 */
public final class StatsSpec implements Serializable {

    private static final long serialVersionUID = -4976077607259531386L;

    private static final TraceComponent tc = Tr.register(
            StatsSpec.class,
            Constants.TR_STATS_GROUP_NAME,
            Constants.TR_RESOURCE_BUNDLE_NAME);

    /**
     * The delimiter for separating the statistic specification string.
     */
    public final static String SPECIFICATION_DELIMIT = ";";

    /**
     * A string constant for the "enabled" string
     */
    public final static String ENABLED_STRING = "enabled";

    /**
     * A string constant for the the "disabled" string
     */
    public final static String DISABLED_STRING = "disabled";

    /**
     * A string constant for the map hit rate statistic
     */
    public final static String MAP_HIT_RATE_STRING = "map.hitrate";

    /**
     * A string constant for the map number of entries statistic
     */
    public final static String MAP_NUM_ENTRIES_STRING = "map.numentries";

    /**
     * A string constant for the map loader batch update statistic
     */
    public final static String MAP_BATCH_UPDATE_TIME_STRING = "map.batchupdate";

    /**
     * A string constant for the map used bytes statistic
     */
    public final static String MAP_USED_BYTES_STRING = "map.usedbytes";

    /**
     * A string constant for the all of the map statistics
     */
    public final static String MAP_ALL_STRING = "map.all";

    /**
     * An integer constant for the map hit count statistic
     */
    public final static int MAP_HIT_RATE = 1 << 0;

    /**
     * An integer constant for the map number of entries statistic
     */
    public final static int MAP_NUM_ENTRIES = 1 << 1;

    /**
     * An integer constant for the map loader batch update time statistics
     */
    public static final int MAP_BATCH_UPDATE_TIME = 1 << 2;

    /**
     * An integer constant for the map number of bytes used statistic
     */
    public static final int MAP_USED_BYTES = 1 << 3;

    /**
     * An integer constant for all of the map statistics
     */
    public final static int MAP_ALL = MAP_HIT_RATE | MAP_NUM_ENTRIES | MAP_BATCH_UPDATE_TIME |
                                      MAP_USED_BYTES;

    /**
     * An integer constant for no statistics
     */
    public final static int NONE = 0x00;

    /**
     * A string constant for the ObjectGrid transaction time statistic
     */
    public final static String OG_TRANSACTION_TIME_STRING = "og.transactiontime";

    /**
     *  A string constant for all of the ObjectGrid statistics
     */
    public final static String OG_ALL_STRING = "og.all";

    /**
     * An integer constant for the ObjectGrid transaction time statistic
     */
    public final static int OG_TRANSACTION_TIME = 0x010;

    /**
     * An integer constant for all of the ObjectGrid statistics
     */
    public final static int OG_ALL = 0x010;

    /**
     * A string constant for all of the replication statistics
     */
    public final static String REPLICATION_ALL_STRING = "replication.all";

    /**
     * An integer constant for all of the replication statistics
     */
    public final static int REPLICATION_ALL = 0x0100;

    /**
     * A string constant for the server client request process time statistic
     */
    public final static String SERVER_PROCESS_TIME_STRING = "server.processtime";

    /**
     * A string constant for all of the server statistics
     */
    public final static String SERVER_ALL_STRING = "server.all";

    /**
     * An integer constant for the server client request process time statistic
     */
    public final static int SERVER_PROCESS_TIME = 0x1000;

    /**
     * An integer constant for all of the server statistics
     */
    public final static int SERVER_ALL = 0x1000;

    /**
     * A string constant for all of the session statistics
     */
    public final static String SESSION_ALL_STRING = "session.all";

    /**
     * An integer constant for all of the session statistics
     */
    public final static int SESSION_ALL = 0x10000;

    /**
     * A string constant for all of the statistics
     */
    public final static String ALL_STRING = "all";

    /**
     * An integer constant for all of the index statistics
     * @since WAS XD 6.1.0.5
     */
    public final static int INDEX_ALL = 0x40000; // 262144 -> 2^18

    /**
     * A string constant for all of the index statistics
     * @since WAS XD 6.1.0.5
     */
    public final static String INDEX_ALL_STRING = "index.all";

    /**
     * An integer constant for all of the ObjectQuery and entity query statistics
     * @since WAS XD 6.1.0.5
     */
    public final static int QUERY_ALL = 0x80000; // 524288 -> 2^19

    /**
     * A string constant for all of the ObjectQuery and entity query statistics
     * @since WAS XD 6.1.0.5
     */
    public final static String QUERY_ALL_STRING = "query.all";


    /**
     * An integer constant for all of the AgentManager agent  and entity query statistics
     * @see AgentManager
     * @since WAS XD 6.1.0.5
     */
    public final static int AGENT_ALL = 0x100000; // 1048576 -> 2^20

    /**
     * An integer constant for all of the virtual memory (diskoverflow) statistics.
     * There's no point in turning off specific sub-statistics because all of them are tracked
     * by the VM layer regardless, so there's no performance penalty associated.
     * @since WXS 7.1.0.3
     */
    /* these are private for now and are defined in VMStatsModule.
     * Flag 0x400000 should still be reserved for use by VM in the future.  
    public final static int VM_ALL=0x400000;
    public final static int VM_HIT_COUNT_ID=1111;
    public final static int VM_ACCESS_COUNT_ID=1112;
    public final static int VM_DISK_USAGE_ID=1113;
    public final static int VM_MEMORY_USAGE_ID=1114;
    public final static int VM_EVICT_NOOP_ID=1115;
    public final static int VM_EVICT_COUNT_ID=1116;
    */
    
    /**
     * A string constant for all of the AgentManager agent  and entity query statistics
     * @see AgentManager
     * @since WAS XD 6.1.0.5
     */
    public final static String AGENT_ALL_STRING = "agent.all";

    /**
     * An integer constant for all of the statistics
     */
    public final static int ALL = OG_ALL | MAP_ALL | REPLICATION_ALL | SERVER_ALL | SESSION_ALL | INDEX_ALL | QUERY_ALL | AGENT_ALL | VMStatsModule.VM_ALL;

    private int specValue = NONE;

    private final static HashMap INT_TO_STR_MAP = new HashMap();

    private final static HashMap STR_TO_INT_MAP = new HashMap();

    static {
        INT_TO_STR_MAP.put(Integer.valueOf(MAP_HIT_RATE), MAP_HIT_RATE_STRING);
        INT_TO_STR_MAP.put(Integer.valueOf(MAP_NUM_ENTRIES), MAP_NUM_ENTRIES_STRING);
        INT_TO_STR_MAP.put(Integer.valueOf(MAP_USED_BYTES), MAP_USED_BYTES_STRING);
        INT_TO_STR_MAP.put(Integer.valueOf(MAP_BATCH_UPDATE_TIME), MAP_BATCH_UPDATE_TIME_STRING);
        INT_TO_STR_MAP.put(Integer.valueOf(MAP_ALL), MAP_ALL_STRING);

        INT_TO_STR_MAP.put(Integer.valueOf(OG_TRANSACTION_TIME), OG_TRANSACTION_TIME_STRING);
        INT_TO_STR_MAP.put(Integer.valueOf(OG_ALL), OG_ALL_STRING);

        INT_TO_STR_MAP.put(Integer.valueOf(REPLICATION_ALL), REPLICATION_ALL_STRING);

        INT_TO_STR_MAP.put(Integer.valueOf(SERVER_PROCESS_TIME), SERVER_PROCESS_TIME_STRING);
        INT_TO_STR_MAP.put(Integer.valueOf(SERVER_ALL), SERVER_ALL_STRING);

        INT_TO_STR_MAP.put(Integer.valueOf(SESSION_ALL), SESSION_ALL_STRING);

        INT_TO_STR_MAP.put(Integer.valueOf(INDEX_ALL), INDEX_ALL_STRING);
        INT_TO_STR_MAP.put(Integer.valueOf(QUERY_ALL), QUERY_ALL_STRING);
        INT_TO_STR_MAP.put(Integer.valueOf(AGENT_ALL), AGENT_ALL_STRING);

        INT_TO_STR_MAP.put(Integer.valueOf(ALL), ALL_STRING);



        STR_TO_INT_MAP.put(MAP_HIT_RATE_STRING, Integer.valueOf(MAP_HIT_RATE));
        STR_TO_INT_MAP.put(MAP_NUM_ENTRIES_STRING, Integer.valueOf(MAP_NUM_ENTRIES));
        STR_TO_INT_MAP.put(MAP_USED_BYTES_STRING, Integer.valueOf(MAP_USED_BYTES));
        STR_TO_INT_MAP.put(MAP_BATCH_UPDATE_TIME_STRING, Integer.valueOf(MAP_BATCH_UPDATE_TIME));
        STR_TO_INT_MAP.put(MAP_ALL_STRING, Integer.valueOf(MAP_ALL));

        STR_TO_INT_MAP.put(OG_TRANSACTION_TIME_STRING, Integer.valueOf(OG_TRANSACTION_TIME));
        STR_TO_INT_MAP.put(OG_ALL_STRING, Integer.valueOf(OG_ALL));

        STR_TO_INT_MAP.put(REPLICATION_ALL_STRING, Integer.valueOf(REPLICATION_ALL));

        STR_TO_INT_MAP.put(SERVER_PROCESS_TIME_STRING, Integer.valueOf(SERVER_PROCESS_TIME));
        STR_TO_INT_MAP.put(SERVER_ALL_STRING, Integer.valueOf(SERVER_ALL));

        STR_TO_INT_MAP.put(SESSION_ALL_STRING, Integer.valueOf(SESSION_ALL));

        STR_TO_INT_MAP.put(QUERY_ALL_STRING, Integer.valueOf(QUERY_ALL));
        STR_TO_INT_MAP.put(INDEX_ALL_STRING, Integer.valueOf(INDEX_ALL));
        STR_TO_INT_MAP.put(AGENT_ALL_STRING, Integer.valueOf(AGENT_ALL));

        STR_TO_INT_MAP.put(ALL_STRING, Integer.valueOf(ALL));


    }

    /**
     * Creates a new StatsSpec object with a <code>NONE</code> statistic
     * specification
     *
     * @see #NONE
     */
    public StatsSpec() {
        // no-op
    }

    /**
     * Creates a new StatsSpec object with a string representation of the
     * statistic specification.
     *
     * @param spec the statistic specification
     */
    public StatsSpec(String spec) {
        if (tc.isEntryEnabled())
            Tr.entry(tc, "StatsSpec", spec);
        setSpec(spec);
        if (tc.isEntryEnabled())
            Tr.exit(tc, "StatsSpec", this);
    }

    /**
     * Creates a new StatsSpec with an integer representation of the statistic
     * specification.
     *
     * @param aSpecValue the statistic specification int value
     */
    public StatsSpec(int aSpecValue) {
        setSpec(aSpecValue);
    }

    /**
     * Sets the statistic specification
     *
     * @param spec the statistic specification
     *
     * @throws IllegalArgumentException if the specified statistic specification
     *         is not valid
     */
    public void setSpec(String spec) {
        if (spec == null) {
            throw new IllegalArgumentException("null is not a valid statistic specification");
        }

        if (spec.trim().length() == 0) {
            specValue = 0;
            return;
        }

        StringTokenizer tokenizer = new StringTokenizer(spec, SPECIFICATION_DELIMIT);

        int newSpecValue = 0;

        while (tokenizer.hasMoreElements()) {
            String token = (String) tokenizer.nextElement();
            String[] valuePair = checkValuePair(token);

            Object intStat = STR_TO_INT_MAP.get(valuePair[0].toLowerCase());

            // the stat specification string cannot be found
            if (intStat == null) {
                throw new IllegalArgumentException("The statistic specification string "
                        + valuePair[0] + " is not valid.");
            }

            if (valuePair[1].equalsIgnoreCase(ENABLED_STRING)) {
                newSpecValue |= ((Integer) intStat).intValue();
            }
            else if (valuePair[1].equalsIgnoreCase(DISABLED_STRING)) {
                newSpecValue &= (ALL - ((Integer) intStat).intValue());
            }
            else {
                if (tc.isEntryEnabled())
                    Tr.exit(tc, "setSpec", "IllegalArgumentException");
                throw new IllegalArgumentException(
                    "The statistic specification enabled/disabled string " + valuePair[1]
                        + " is not valid.");
            }
        }

        specValue = newSpecValue;
    }

    /**
     * Sets the statistic specification
     *
     * @param value the statistic specification
     *
     * @throws IllegalArgumentException if the specified statistic specification
     *         is not valid
     */
    public void setSpec(int value) {
        if (value > ALL || value < NONE) {
            throw new IllegalArgumentException("The statistic specification value " + value
                    + " is not valid.");
        }
        this.specValue = value;
    }

    private String[] checkValuePair(String specString) {
        int index = specString.indexOf("=");
        if(index < 0) {
            throw new IllegalArgumentException("The statistic specification string " + specString
                    + " is not valid.");
        }

        String key = specString.substring(0, index);
        String value = specString.substring(index + 1);

        if (value == null || value.indexOf("=") >= 0) {
            throw new IllegalArgumentException("The statistic specification string " + specString
                    + " is not valid.");
        }

        return new String[] { key, value };
    }

    /**
     * Gets the integer representation of the statistics specification
     *
     * @return the statistic specification as an integer value
     */
    public int getSpecValue() {
        return specValue;
    }

    /**
     * Determines whether the specified statistics are enabled
     *
     * @param stat the integer representation of the statistics to check
     *
     * @return <code>true</code> if this StatsSpec enables the specified
     *         statistics.
     *
     * @throws IllegalArgumentException if the specified statistic is not valid
     */
    public boolean isStatisticEnabled(int stat) {
        if ((stat & (stat - 1)) == 0) {
            // i represents a specific statistic
            return (specValue & stat) == 0 ? false : true;
        }

        throw new IllegalArgumentException("The int value " + stat
                + " does not represent a specific statistic.");
    }

    /**
     * Enables the specified statistics
     *
     * @param stat the integer representation of a statistic
     *
     * @throws IllegalArgumentException if the specified statistic is not valid
     */
    public void enableStatistic(int stat) {
        if ((stat & (stat - 1)) == 0) {
            // i represents a specific statistic
            specValue |= stat;
        }
        else {
            throw new IllegalArgumentException("The int value " + stat
                    + " does not represent a specific statistic.");
        }
    }

    /**
     * Disables the specified statistics
     *
     * @param stat the integer representation of the statistics
     *
     * @throws IllegalArgumentException if the specified statistic is not valid
     */
    public void disableStatistic(int stat) {
        if ((stat & (stat - 1)) == 0) {
            // i represents a specific statistic
            specValue = (specValue | stat) - stat;
        }
        else {
            throw new IllegalArgumentException("The int value " + stat
                    + " does not represent a specific statistic.");
        }
    }

    /**
     * Checks whether any map statistic is enabled
     *
     * @return <code>true</code> if any map statistic is enabled
     */
    public boolean isAnyMapStatEnabled() {
        return (specValue & MAP_ALL) != 0;
    }

    /**
     * Checks whether any query statistic is enabled
     *
	 * @since 7.1
     * @return <code>true</code> if any query statistic is enabled
     */
    public boolean isAnyQueryStatEnabled() {
        return (specValue & QUERY_ALL) != 0;
    }

    /**
     * Checks whether any index statistic is enabled
     *
     * @since 7.1
     * @return <code>true</code> if any index statistic is enabled
     */
    public boolean isAnyIndexStatEnabled() {
        return (specValue & INDEX_ALL) != 0;
    }

    /**
     * Checks whether any agent statistic is enabled
     *
	 * @since 7.1
     * @return <code>true</code> if any agent statistic is enabled
     */
    public boolean isAnyAgentStatEnabled() {
        return (specValue & AGENT_ALL) != 0;
    }

    /**
     * Checks whether any replication statistic is enabled
     *
     * @return <code>true</code> if any replication statistic is enabled
     */
    public boolean isAnyReplicationStatEnabled() {
        return (specValue & REPLICATION_ALL) != 0;
    }

    /**
     * Checks whether any ObjectGrid statistic is enabled
     *
     * @return <code>true</code> if any ObjectGrid statistic is enabled
     */
    public boolean isAnyOGStatEnabled() {
        return (specValue & OG_ALL) != 0;
    }

    /**
     * Checks whether any server statistic is enabled
     *
     * @return <code>true</code> if any server statistic is enabled
     */
    public boolean isAnyServerStatEnabled() {
        return (specValue & SERVER_ALL) != 0;
    }

    /**
     * Checks whether any session statistic is enabled
     *
     * @return <code>true</code> if any session statistic is enabled
     */
    public boolean isAnySessionStatEnabled() {
        return (specValue & SESSION_ALL) != 0;
    }

    /**
     * Gets the string representation of the statistics specification
     *
     * @return the statistic specification as a string value
     */
    public String getSpecStringValue() {
        if (specValue == NONE) {
            return "all=disabled";
        }
        if(specValue == ALL) {
            return ALL_STRING.concat("=").concat(ENABLED_STRING);
        }
        int i=1;
        boolean isColonAdded = false;
        StringBuffer s = new StringBuffer();
        if((MAP_ALL & specValue) == MAP_ALL) {
            s.append(MAP_ALL_STRING).append("=").append(ENABLED_STRING);
            i=MAP_USED_BYTES*2;
            isColonAdded = true;
        }

        for (; i <= ALL; i *= 2) {
            if ((specValue & i) == i) {
                if (isColonAdded) {
                    s.append(";");
                }
                s.append(INT_TO_STR_MAP.get(Integer.valueOf(i))).append("=").append(ENABLED_STRING);
                isColonAdded = true;
            }
        }
        return new String(s);
    }

    /**
     * Gets the String representation of this StatsSpec
     *
     * @return the string representation of this object
     */
    public String toString() {
        StringBuffer sb = new StringBuffer();
        sb
        .append("StatsSpec")
        .append("@")
        .append(hashCode())
        .append("[")
        .append(specValue)
        .append(",")
        .append(getSpecStringValue())
        .append("]");
        return new String(sb);
    }
}
